#------------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: _headerGenerator.py
#   - Version: 1.2
#   - Desc: COMMS (Arduino/Python) header generation utility.
#
#   [Revision History]
#   2019/09/21: v1.2 - Further integration with bash script.
#   2019/03/22: v1.1 - General clean-up and code simplification.
#                      Code adapted from prototype VKOS software.
#   2016/??:    v1.0 - Initial creation.
#------------------------------------------------------------------



# Imports
#---------
import sys, os, platform, datetime
from collections import OrderedDict



# File object
#-------------
class headerFile:
        # Properties
        #------------
        sections = OrderedDict()	# Sections contain a dictionary of commands/values
        numSections = 0			# Number of sections
        numCmds = 0			# Number of constants
        numLines = 0			# Number of lines in file

        

        # Functions
        #-------------------------------------------------------
        # banner(sCaption)
        # Desc: Used to section/separate terminal output.
        #-------------------------------------------------------
        def banner(self, sCaption):
                print(("="*50))
                print(">>> "+sCaption)
                print(("="*50))
                return



        #-------------------------------------------------------
        # display()
        # Desc: Debug-> Show contents of internal arrays.
        #-------------------------------------------------------
        def display(self):
                self.banner("Enumerate collected data")
                print("   * Total sections:  " + str(self.numSections))
                print("   * Total constants: " + str(self.numCmds)+"\n")

                # Display table of sections
                #---------------------------
                print("   [All sections]")
                lSections = self.sections.keys()
                for s in range(0,len(self.sections)):
                        print("   " + str(s) + ".) "+lSections[s])

                print("\n")	# Separator

                # Enumerate each section's contents.
                #------------------------------------
                curC = 0				# Initialize constant counter for each section
                lSections = self.sections.keys()	# Get list of all section titles
                self.banner("Enumerating contents")
                for s in range(0,len(lSections)):
                        curC = 0					# Reset for each section
                        print("   [" + lSections[s] + "]")		# Display section title
                        numC = int(len(self.sections[lSections[s]]))	# Get number of constants in section
                        print("   * Number of constants: "+str(numC))	# Display  ^     ^     ^     ^     ^

                        # Enumerate section constants
                        lCmds = self.sections[lSections[s]].keys()	# Get list of constants in this section
                        for c in range(0,int(len(lCmds))):
                                print("     " + str(curC) + ".) " + str(lCmds[c]).rstrip() + " = " + str(self.sections[lSections[s]][lCmds[c]]))	# Get stored value
                                curC += 1	# Inc. section constant counter

                        # Don't put a line after the last section... It looks odd.
                        if (s < len(lSections)): print("\n")	# Separator
                return



        #-------------------------------------------------------
        # formatData(sType, sFile)
        # Desc: Prepare a long-ass formatted string and toss it
        #       into a file.
        #-------------------------------------------------------
        # - sType: Format of output ("ard" or "py", for Arduino/Python).
        # - sFile: Output file.
        #-------------------------------------------------------
        def formatData(self, sType, sFile):
                # Base setup
                #------------
                sSeparator = "\\" if (platform.system == "Windows") else "/"
                sDecl = ""	# Declaration prefix (language dependant)
                sOutput = ""	# Output buffer

                # Determine necessary syntax prefixes/suffixes.
                #-----------------------------------------------
                if (sType == "py"):			# Python output
                        bFuncOutput = True
                        sType = "Python"
                        sComment = "###"
                        sSuffix = ""
                elif (sType == "ard"):			# Arduino/C++ output
                        bFuncOutput = False
                        sType = "Arduino/C++"
                        sDecl = "const char "
                        sComment = "//"
                        sSuffix = ";"

                # Prefix standard header.
                #-------------------------
                sOutput += sComment+" VisualK-OS: Command Constants\n"
                sOutput += sComment+" Generated on: " + str("{:%B %d, %Y [%I:%M]}".format(datetime.datetime.now())) + "\n\n"

                # Iterate through data.
                #-----------------------
                if (bFuncOutput):   # Python function data.
                    sFuncOutput = ("\n\n#" + ("-"*55) + "\n")
                    sFuncOutput += "# getConstStr(iConst)\n"
                    sFuncOutput += "# Desc: Returns a string identifying a constant.\n"
                    sFuncOutput += ("#" + ("-"*55) + "\n")
                    sFuncOutput += "def getConstStr(iConst):\n"

                lSections = self.sections.keys()	     # Get list of sections
                for s in range(0,len(self.sections)):
                        if (bFuncOutput):
                            sFuncOutput += chr(9) + "# " + lSections[s] + "\n"
                        sOutput += sComment+" "+lSections[s]+"\n"       # Section titles become comments in the output file(s).

                        # Enumerate constants into one string in memory.
                        #------------------------------------------------
                        lCmds = self.sections[lSections[s]].keys()	                # Get list of constants
                        for c in range(0,len(self.sections[lSections[s]])):
                                sName = lCmds[c].rstrip()	                        # Constant's name
                                sVal = str(self.sections[lSections[s]][lCmds[c]])	# Constant's value.
                                sOutput += (sDecl + sName + " = " + sVal + sSuffix + "\n")
                                
                                # Add onto Python function, if needed.
                                if (bFuncOutput):
                                    sFuncOutput += (chr(9) + "if (iConst == " + sName + "): return \"" + sName + "\"" + "\n")
                        sOutput += "\n"		# Line separator
                        if (bFuncOutput): sFuncOutput += "\n"

                # Python function catch-all
                if (bFuncOutput):
                    sFuncOutput += (chr(9) + "# Unknown constant\n" + chr(9) + "#------------------\n")
                    sFuncOutput += (chr(9) + "return str(\"<Unknown constant: '\" + str(iConst) + \"'>\")\n")


                # Check if the file exists (prompt to overwrite, if so).
                #--------------------------------------------------------
                sFullPath = os.getcwd()+sSeparator+sFile
                if (os.path.isfile(sFullPath)):
                    sOverwrite = str(raw_input("[!] " + sFullPath + " already exists.\n   - Overwrite? (y/n): "))
                    if (sOverwrite == "y"):
                            print("[*] Deleting \"" + sFullPath + "\"")
                            os.remove(os.getcwd()+"/"+sFile)	# Delete file.
                    else:
                            self.banner("\n[!] Did not receive confirmation... Exiting.")
                            sys.exit()

                # Output to file.
                #-----------------
                oFile = open(sFile,"w")		# Get a handle to our file
                oFile.write(sOutput)		# Actual file writing
                
                # Custom constant->string Python function.
                #------------------------------------------
                if (bFuncOutput): oFile.write(sFuncOutput)

                oFile.close()			# We're done!
                print("[*] Success! " + sType + " output to \"" + sFile + "\" complete.")
                return



        #-------------------------------------------------------
        # loadDefinitions(inF)
        # Desc: Load and parse header definition file.
        #-------------------------------------------------------
        # - inF: Input file.
        #-------------------------------------------------------
        def loadDefinitions(self, inF = "", bVerbose = False):
                # Load definitions file
                with open(inF, "r") as iFile:
                        # Setup
                        #-------
                        sLastSection = ""	# Last known section header
                        curC = 0		# Count of constants in current section
			if (bVerbose): self.banner("Processing input file")

                        # Read in file.
                        #---------------
                        for line in iFile:	# Bypass empty lines and comments
                                line.rstrip()	# Strip line-terminating characters

                                # Check line contents.
                                # Skip blank lines, space-prefixed lines and comments.
                                #------------------------------------------------------
                                if ((len(line) > 1) and (not line.startswith(" ")) and (not line.startswith("#"))):
                                        if (line.startswith("[")):
                                                # Encountered section header
                                                #----------------------------
                                                sTitle = line[1:(len(line)-2)]		# Extract section name
                                                if (sTitle == ""):			# Ensure valid section name
                                                        # Invalid section header. Keep using the previous one.
                                                        print("[!] Encountered empty section header at line #" + str(self.curLine)+"\n")
                                                else:
                                                        # Validated, move on...
                                                        self.numLines += 1		# Inc. working line counter
                                                        self.numSections += 1		# Inc. section count
                                                        if (bVerbose): print("\n  [*] New section: \"" + sTitle + "\"")
                                                        sLastSection = sTitle		# Set last known section

                                                        # Create entry for new section w/cmds sub-array
                                                        self.sections.update({sTitle:OrderedDict()})

                                        elif (line.startswith("$BASE$")):
                                                # Encountered base value specification (sets base for sequential value assignment)
                                                newBase = line[7:len(line)]

                                                # Validate
                                                #----------
                                                if (newBase):
                                                        newBase = newBase.rstrip()
                                                        if (bVerbose): print("     [*] Base value change: " + newBase)
                                                        curC = int(newBase)	# Store
                                                else:
                                                        print("     [!] Encountered invalid base value change.")

                                        else:
                                                ### Encountered constant definition
                                                if (bVerbose): print("     - Found: \"" + line.rstrip() + "\", value = " + str(curC))
                                                self.sections[sLastSection].update({line:int(curC)})	# Set constant name, assign sequential value
                                                curC += 1					# Inc. local section constant counter
                                                self.numCmds += 1       			# Inc. total number of known constants
                        print("\n")     # Line break.
                return



#-------------------------------------------------------
# Main Program
#-------------------------------------------------------
if __name__== "__main__":
    os.system("clear")
    print("[VisualK-OS Arduino/Python Header Generator]")
    
    ### Get input file
    if(len(sys.argv) == 0):
        print("[!] No input file specified.")
        quit()
    else:
        # Debug
	bEnumerateArrays = (sys.argv[0] == "-verbose")
	if not bEnumerateArrays: print("   - To enable verbose output, add '-verbose' as an argument.\n     (eg: '-verbose " + sys.argv[0] + "')")

        # Process input file.
        # --------------------
        inF = headerFile()
        sInFile = sys.argv[1] if not bEnumerateArrays else sys.argv[2]		# Set input file
        print("\n>>> Loading definitions input file: \"" + sInFile + "\"\n")
	inF.loadDefinitions(sInFile)		# Load definitions file
        if (bEnumerateArrays): inF.display()	# (Debug: Show contents of arrays)
	
        # Output; Prompt for each file type.
        #------------------------------------
        inF.banner("Data processed.")

        # Arduino header.
        #-----------------
        ard = str(raw_input("Write Arduino file (COMMS_Constants.ino)? [y/n]: "))
        if ((ard == "y")):
		inF.formatData("ard", "COMMS_Constants.ino")
        else: print("[!] Skipping Arduino file.")

        # Python include file.
        #----------------------
        py = str(raw_input("Write Python include file (VKOS_Constants.py)? [y/n]: "))
        if (py == "y"):
		inF.formatData("py", "VKOS_Constants.py")    # Write Python include file
        else: print("[!] Skipping Python include file.")

        # Exit
        print("Done!")
        sys.exit()

