#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: COMMS_Packet.py
#   - Version: 1.2
#   - Author: Kevin Slaughter
#   - Desc: Object packet for COMMS events, to keep things
#           relatively neat. This is meant to be included in
#           any module that employs COMMS connectivity.
#
#
#   [Revision History]
#   2019/09/20: v1.2 - Improved module's string representation.
#   2019/03/21: v1.1 - More state/op-related constants defined.
#   2019/03/13: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import time
from collections import OrderedDict
from VKOS_Constants import *	# Kept separate for easy updating (in sync w/Arduino header).



# Packet object
#---------------
class COMMS_Packet(object):
	# Initialization
	#----------------
	def __init__(self, iCmd = None, iLamp = None, iParam = None):
		self.clearPacket()	# Set up a fresh packet.

		# If a command and/or lamp was supplied when creating
		# the packet, assign it after the initial setup.
		if (not (iCmd == None)):
			if (not (iLamp == None)): self.setCmd(iCmd, iLamp)
			else: self.setCmd(iCmd)

			# Add parameter values, if any, from the tuple.
			if ((iParam) and (len(iParam) == 2)):
				self.addParam(iParam[0], iParam[1])
		return


	# Formatted string representation
	#---------------------------------
	def __str__(self):
		# Header/Parameter count.
		#-------------------------
		strSeparator = ("-"*18) + "\n"		# For code readability.
		rtnStr = "\n" + strSeparator + ":: COMMS_Packet ::\n" + strSeparator
		rtnStr += "cmd: " + getConstStr(self.cmd) + "\n"
		rtnStr += "   - params (" + str(len(self.params)) + ")\n"

		# List parameters.
		#------------------
		for iParam in self.params.keys():
			# Prefix with parameter type.
			rtnStr += "     - ['" + getConstStr(ord(iParam)) + "']: "

			# Attempt to resolve constants to strings, if we can.
			if (ord(iParam) == CPACK_ARG_CONST):
				rtnStr += getConstStr(self.params[iParam]) + "\n"
			elif (ord(iParam) == CPACK_ARG_COLOR):
				rtnStr += str(self.color) + "\n"
			else:
				# Paramter is not a constant.
				rtnStr += str(self.params[iParam]) + "\n"

		# Misc.
		#-------
		rtnStr += "   - processed: " + str(self.processed) + "\n"
		rtnStr += "   - timestamp: " + str(self.relativeTime) + "\n" + strSeparator
		return rtnStr



	#-------------------------------------------------------
	# Functions
	#-------------------------------------------------------
	# addParam(iType, paramData):
	# Desc: Add a parameter to the packet.
	#-------------------------------------------------------
	# -      iType: Type of data (Int, const, etc).
	# -  paramData: Parameter data.
	#-------------------------------------------------------
	def addParam(self, iType = None, paramData = None):
		# Sanity-check: Make sure we're not adding empty parameters.
		if (not (iType == None) and not (paramData == None)):
			if (iType == CPACK_ARG_COLOR):
				# Color data is stored in the packet, not as a normal parameter.
				print "paramData = " + str(paramData)
				self.color = [paramData[0], paramData[1], paramData[2]]
				self.params.update({chr(iType) : "1"})
			else:
				# Store parameter with type.
				self.params.update({chr(iType) : paramData})
			self.numParams += 1
		return



	#-------------------------------------------------------
	# clearPacket()
	# Desc: Clears and resets a packet to a default state.
	#-------------------------------------------------------
	def clearPacket(self):
		self.cmd = COMMS_NOP		# Parsed commands/params.
		self.numParams = 0		# Number of parameters.
		self.color = [0, 0, 0]		# Color tuple (if applicable).
		self.params = OrderedDict()	# Parameters for command arguments.
		self.processed = False		# Flag packet as not processed, until parsed by COMMSMAN.

		# Timestamp packet (Low resolution).
		self.relativeTime = str(time.time())[7:]
		return



	#-------------------------------------------------------
	# setCmd(newCmd, newParam):
	# Desc: Set the packet's command.
	#-------------------------------------------------------
	# -   newCmd: Command to execute.
	#-------------------------------------------------------
	def setCmd(self, iCmd = None, iParam = None):
		# Sanity-check: Make sure we're not adding empty commands.
		if (not (iCmd == None)):
			self.cmd = iCmd

			# Add first parameter, if specified.
			if (not (iParam == None)):
				self.addParam(CPACK_ARG_CONST, iParam)
		return



	#-------------------------------------------------------
	# setColor(tColor):
	# Desc: Set the packet's command.
	#-------------------------------------------------------
	# - tColor: HSV Color tuple [0, 0, 0]
	#-------------------------------------------------------
	def setColor(self, tColor):
		# Sanity-checks; Ensure values are within range.
		if (tColor[0] > 360): tColor = [360, tColor[1], tColor[2]]
		if (tColor[1] > 255): tColor = [tColor[0], 255, tColor[2]]
		if (tColor[2] > 255): tColor = [tColor[0], tColor[1], 255]
		self.addParam(CPACK_ARG_COLOR, [abs(tColor[0]), abs(tColor[1]), abs(tColor[2])])
		return
