#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: UIE_Button.py
#   - Version: 1.3
#   - Author: Kevin Slaughter
#   - Desc: UI Element >> Command Button
#
#
#   [Revision History]
#   2019/09/23: v1.3 - Removed handleMouse(), as it was just
#                      an empty placeholder function, into
#                      UIE_Base.py where such a thing should
#                      have been placed initially for all UIEs.
#   2019/02/09: v1.2 - Commenting/Documentation standardization.
#   2019/02/04: v1.1 - Addition of disabled/enabled coloring,
#                      as well as general housekeeping.
#   2018/12/30: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, vkos_globals
from UIE_Base import *



# Class Object
#--------------
class UIE_Button(UIE_Base):
	# Initialization
	#----------------
	def __init__(self, strText, dispRect=None, aLoc=None, objName="UNKNOWN", numIndex=None, owner=None):
		UIE_Base.__init__(self, dispRect, aLoc)		# Call standard init for all UIE objects.
		self.active = False				# "Active"/"Focus" state (hovered over, etc).
		self.enabled = True				# Determines if the button can be activated. (Default = True)
		if (numIndex): self.name = "c" + str(numIndex) + "-btn" + objName	# Internal object name.
		else: self.name = (owner + "-" + objName)
		self.owner = owner				# UIE host for button.
		self.state = False				# Ensure default state.
		self.text = strText				# Text to draw onto the button.
		self.type = "BUTTON"				# Define UIE type.
		self.redraw = True				# Flag initial draw.

		# State-specific colors
		self.colors.update({"FG-Active"   : vkos_globals.TC_FG_BUTTON_ACTIVE})		# Active text color
		self.colors.update({"FG-Disabled" : vkos_globals.TC_FG_BUTTON_DISABLED})	# Disabled text color
		return


	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the element itself on a given surface.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		self.surfMain.fill(self.colors["BG"])   # Blank BG
		tempRect = self.getRect()		# Prepare edges

		# Set text color
		if ((self.active) or (self.state)):
			textColor = self.colors["FG-Active"]
		elif not (self.enabled):
			textColor = self.colors["FG-Disabled"]
		else:
			textColor = self.colors["FG"]

		# Render text
		surfText = self.fontMain.render(self.text, False, textColor)
		tempRect = surfText.get_rect()
		textPos = [((self.size[0]/2)-(tempRect[2]/2)), (tempRect[3]/4)]     # Approximate centering

		# If the button is clicked, make visual changes.
		if (self.state):  		# Clicked
			invertFrame = True
			textPos = [(textPos[0]+2), (textPos[1]+2)]	# Slight position offset when active.
		else: invertFrame = False	# Normal

		# Draw BG, edges, and text... Then copy back to the host.
		vkos_globals.drawContouredRect(self.surfMain, self.getRect(), vkos_globals.TC_BG_SECTION, invertFrame)
		self.surfMain.blit(surfText, textPos)
		destSurf.blit(self.surfMain, self.rLoc)
		self.redraw = False
		return



	#-------------------------------------------------------
	# setState(bState)
	# Desc: Set the state of the button.
	#-------------------------------------------------------
	# -   bState: State to be set.
	#-------------------------------------------------------
	def setState(self, bState):
		# Sanity-check; Ensure a boolean was given.
		if (isinstance(bState, (bool))):
			self.state = bState		# Set button's visual state.
			self.redraw = True		# Flag for redraw.
		return



	#-------------------------------------------------------
	# setText(strText)
	# Desc: Set the button's displayed text.
	#-------------------------------------------------------
	# - strText: New display text.
	#-------------------------------------------------------
	def setText(self, strText=""):
		self.text = strText	# Set display text.
		self.redraw = True	# Redraw UIE.
		return
