#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: UIE_ColorWheel.py
#   - Version: 1.1
#   - Author: Kevin Slaughter
#   - Desc: UI Element >> Color selection interface.
#           Used for easy/compact color selection.
#
#
#   [Revision History]
#   2019/09/23: v1.1 - General bugfixes, unification with
#                      hosting objColorChannel properties.
#   2019/07/27: v1.0 - File created with initial framework. I
#                      changed to a color wheel for simplicity,
#                      rather than the problematic bar-graph
#                      color select that was previously in use.
#---------------------------------------------------------------



# Imports
#---------
import colorsys, math, pygame, vkos_globals
from PIL import Image
from UIE_Base import *



# Class Object
#--------------
class UIE_ColorWheel(UIE_Base):
	# Initialization
	#----------------
	def __init__(self, dispRect=None, aLoc=None, numIndex=None, owner=None):
		UIE_Base.__init__(self, dispRect, aLoc)		# Call standard init for all UIE objects.
		self.aLoc = tuple(aLoc)				# Absolute location.
		self.colors["BG"] = [0, 0, 0]			# Black = Transparent for ColorWheels.
		self.enabled = True				# UIE state. ColorWheels should almost never be disabled.
		self.rLoc = tuple(dispRect)			# Relative location.
		self.curColor = [220, 255, 0]			# Default color = Blue.
		self.name = "c" + str(numIndex) + "-colorWheel"	# Default object name.
		self.owner = owner				# UIE host for colorWheel.
		self.redraw = True				# Redraw flag.

		# Set up and draw color wheel.
		#------------------------------
		self.surfWheel = pygame.Surface([dispRect[2], dispRect[3]])	# Special, so we only draw the wheel once.
		self.surfWheel.set_colorkey([0, 0, 0])				# Set color key for transparency.
		self.drawSpectrum()						# Draw color wheel.
		return



	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the element itself on a given surface.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		destSurf.blit(self.surfWheel, self.rLoc)		# Blit spectrum to main surface.
		vkos_globals.addUpdateRgn(self.getAbsRect())		# Add screen update region.
		self.redraw = False
		return



	#-------------------------------------------------------
	# drawSpectrum()
	# Desc: Draws the HSV color spectrum.
	#-------------------------------------------------------
	def drawSpectrum(self):
		# Set up PIL image
		im = Image.new("RGB", (self.size[0], self.size[1]))
		radius = (min(im.size)/2.0)
		centre = [(im.size[0]/2), (im.size[1]/2)]
		pix = im.load()

		# Draw color wheel.
		for x in range(im.width):
			for y in range(im.height):
				rx = (x - centre[0]); ry = (y - centre[1])
				s = ((((x - centre[0])**2.0 + (y - centre[1])**2.0)**0.5) / radius)
				if s <= 1.0:
					h = (((math.atan2(ry, rx) / math.pi) + 1.0) / 2.0)
					pix[x,y] = tuple([int(round(c*255.0)) for c in colorsys.hsv_to_rgb(h, 1.0, 1.0)])

		self.surfWheel = pygame.image.fromstring(im.tobytes(), im.size, im.mode)
		self.surfWheel.set_colorkey([0, 0, 0]); self.surfWheel.convert()
		return



	#-------------------------------------------------------
	# getColor()
	# Desc: Returns the color most recently selected/assigned.
	#-------------------------------------------------------
	def getColor(self): return self.curColor



	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Mouse event handler.
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		if (event.type == pygame.MOUSEBUTTONUP):
			mPos = [(event.pos[0]-self.aLoc[0]), (event.pos[1]-self.aLoc[1])]

			# Mouse button UP (Handle clicks here)
			#--------------------------------------
			if (vkos_globals.isInRect(self.aLoc, event.pos)):
				colSelectedRGB = self.surfWheel.get_at(mPos)
				colSelectedRGB = [colSelectedRGB[0], colSelectedRGB[1], colSelectedRGB[2]]
				colSelectedHSV = colorsys.rgb_to_hsv((colSelectedRGB[0]/255.0),
								     (colSelectedRGB[1]/255.0),
								     (colSelectedRGB[2]/255.0))

				# Check position; Only clicks within the circle are valid.
				pX = (mPos[0] - (self.size[0]/2.0)); pX = (pX*pX)		# X^2
				pY = (mPos[1] - (self.size[1]/2.0)); pY = (pY*pY)		# Y^2
				radius = (min(self.size)/2.0); radius = (radius*radius)		# R^2
				if ((pX + pY) < radius):
					# Change back to integers.
					colSelectedHSV = [int((colSelectedHSV[0]*360.0)),
							  int((colSelectedHSV[1]*255.0)),
							  int((colSelectedHSV[2]*255.0))]
					self.setColor(colSelectedHSV)	# Set currently-selected color hue.
		return



	#-------------------------------------------------------
	# setColor(newColor)
	# Desc: Set the last-selected color info.
	#-------------------------------------------------------
	# - newColor: HSV Tuple.
	#-------------------------------------------------------
	def setColor(self, newColor):
		# Sanity checks; Ensure a valid color is given.
		newColor = [abs(newColor[0]), abs(newColor[1]), abs(newColor[2])]
		if (newColor[0] > 360): newColor = [360, newColor[1], newColor[2]]
		if (newColor[1] > 255): newColor = [newColor[0], 255, newColor[2]]
		if (newColor[2] > 255): newColor = [newColor[0], newColor[1], 255]

		# Assign new color.
		self.curColor = newColor
		return
