#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: UIE_SideMenu.py
#   - Version: 1.3
#   - Author: Kevin Slaughter
#   - Desc: Screen object >> Side-oriented Menu
#           Used for COMMS management and secondary commands
#           for devices (macros, control adjustment, etc).
#
#
#   [Revision History]
#   2019/09/23: v1.3 - Simplfication, UI is now slightly more
#                      compact necessatated by changing the
#                      header text from "COMMS Mgmt" to "COMMS".
#   2019/02/09: v1.2 - Commenting/Documentation standardization.
#   2019/02/04: v1.1 - Added opaque layer, to create a
#                      translucent background with its
#                      foreground objects completely opaque.
#   2019/02/01: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, random, vkos_globals
from UIE_Button import UIE_Button
from UIE_Base import *


# Constants
#-----------
MENU_OPACITY = 192


# Class Object
#--------------
class UIE_SideMenu(UIE_Base):
	# Initialization
	#----------------
	def __init__(self, dispRect):
		UIE_Base.__init__(self, dispRect, dispRect)     # Call standard init for all UIE objects.
		self.aLoc = [dispRect[0], dispRect[1]]		# SideMenu uses absolute locations.
		self.size = [dispRect[2], dispRect[3]]		# Store UIE dimensions.
		self.opacity = 255				# Fade-related flag (Default to fully-opaque).
		self.surfOpaque = pygame.Surface(self.size)	# Opaque surface, so BG can be transparent.
		self.surfOpaque.set_colorkey([255, 0, 255])	# Set transparency key (Magenta).
		self.type = "MENU"				# Define UIE type.


		# Load COMMS Management SideMenu
		#--------------------------------
		# Render header text. Since the SideMenu BG is translucent,
		# allow extra rendering (no BG color specificaiton).
		self.surfHeader = {"Bottom" : vkos_globals.SM.fonts["SideMenu-Header"].render(vkos_globals.SIDEMENU_TEXT, False, [0, 0, 255]),
				   "Top"    : vkos_globals.SM.fonts["SideMenu-Header"].render(vkos_globals.SIDEMENU_TEXT, False, [0, 220, 0])}
		self.surfMain.set_alpha(MENU_OPACITY); self.surfMain.convert_alpha()	# Apply transparency to bottom layer

		# FBA buttons
		#-------------
		rectBase = [12, 80, 135, 25]		# Most SideMenu buttons are 164x25.

		# Connect
		curY = rectBase[1]
		rectFBAConnect = tuple(rectBase)
		locFBAConnect = [(self.aLoc[0] + rectFBAConnect[0]), (self.aLoc[1] + curY)]

		# Disconnect
		curY += (rectBase[3] + 5)
		rectFBADisconnect = [rectBase[0], curY, rectBase[2], rectBase[3]]
		locFBADisconnect = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + curY)]

		# Mute
		curY += (rectBase[3] + 5)
		rectFBAMute = [rectBase[0], curY, rectBase[2], rectBase[3]]
		locFBAMute = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + curY)]


		# Thor buttons
		#--------------
		rectBase = [rectBase[0], ((self.size[1] / 2) + 50), rectBase[2], rectBase[3]]

		# Connect
		curY = rectBase[1]
		rectThorConnect = tuple(rectBase)
		locThorConnect = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + rectBase[1])]

		# Disconnect
		curY += (rectBase[3] + 5)
		rectThorDisconnect = [rectBase[0], curY, rectBase[2], rectBase[3]]
		locThorDisconnect = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + curY)]

		# Mute
		curY += (rectBase[3] + 5)
		rectThorMute = [rectBase[0], curY, rectBase[2], rectBase[3]]
		locThorMute = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + curY)]


		# Main System buttons
		#---------------------
		# "Exit"
		curY = (self.size[1] - 35)
		rectExit = [rectBase[0], curY, rectBase[2], rectBase[3]]
		locExit = [(self.aLoc[0] + rectBase[0]), (self.aLoc[1] + curY)]


		# Create button UIEs.
		#---------------------
		self.buttons = {"FBA-Connect"     : UIE_Button("Connect FBA", dispRect=rectFBAConnect, aLoc=locFBAConnect, objName="FBA-Connect", owner="SM"),
		                "FBA-Disconnect"  : UIE_Button("Disconnect FBA", dispRect=rectFBADisconnect, aLoc=locFBADisconnect, objName="FBA-Disconnect", owner="SM"),
		                "FBA-Mute"        : UIE_Button("Mute", dispRect=rectFBAMute, aLoc=locFBAMute, objName="FBA-Mute", owner="SM"),
		                "Thor-Connect"    : UIE_Button("Connect Thor", dispRect=rectThorConnect, aLoc=locThorConnect, objName="Thor-Connect", owner="SM"),
		                "Thor-Disconnect" : UIE_Button("Disconnect Thor", dispRect=rectThorDisconnect, aLoc=locThorDisconnect, objName="Thor-Disconnect", owner="SM"),
		                "Thor-Mute"       : UIE_Button("Mute", dispRect=rectThorMute, aLoc=locThorMute, objName="Thor-Mute", owner="SM"),
		                "Main-Exit"       : UIE_Button("Exit", dispRect=rectExit, aLoc=locExit, objName="Main-Exit", owner="SM")}

		# Define clickable areas
		#------------------------
		self.clickable = {"SM-FBA-Connect"     : self.buttons["FBA-Connect"].getAbsRect(),
		                  "SM-FBA-Disconnect"  : self.buttons["FBA-Disconnect"].getAbsRect(),
		                  "SM-FBA-Mute"        : self.buttons["FBA-Mute"].getAbsRect(),
		                  "SM-Thor-Connect"    : self.buttons["Thor-Connect"].getAbsRect(),
		                  "SM-Thor-Disconnect" : self.buttons["Thor-Disconnect"].getAbsRect(),
		                  "SM-Thor-Mute"       : self.buttons["Thor-Mute"].getAbsRect(),
		                  "SM-Main-Exit"       : self.buttons["Main-Exit"].getAbsRect()}
		return



	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the element itself on a given surface.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Blank BG and draw frame.
		self.surfMain.fill(vkos_globals.TC_SIDEMENU_BG)
		vkos_globals.drawContouredRect(self.surfMain, self.getRect(), vkos_globals.TC_BG_SECTION, invert=True)

		# Repaint background area under SideMenu, to retain proper alpha blending.
		vkos_globals.SM.desktop.bg.draw(destSurf, rgn=self.getAbsRect())

		# Make sure we don't have odd transparency disparities when fading in.
		if (self.opacity < 255):
			newOpacity = self.opacity if (self.opacity < MENU_OPACITY) else MENU_OPACITY
			self.surfMain.set_alpha(newOpacity); self.surfMain.convert_alpha()
		destSurf.blit(self.surfMain, self.aLoc)

		# Transparent BG, so the second layer remains opaque
		# when drawn over the alphablended background surface.
		self.surfOpaque.fill((255, 0, 255))	# Fill layer with transparent color.

		# Draw COMMS header and frame.
		#------------------------------
		textLoc = [30, 12]		# Header text [x, y]
		rectHeader = self.getRect()	# Frame
		rectHeader = [(rectHeader[0] + 10), (rectHeader[1] + 6), (rectHeader[2] - 18), 46]
		vkos_globals.drawContouredRect(self.surfOpaque, rectHeader, [64, 64, 64], invert=True)	# Draw frame.
		self.surfOpaque.blit(self.surfHeader["Bottom"], textLoc)				# Draw top layer.
		self.surfOpaque.blit(self.surfHeader["Top"], [(textLoc[0]+3), (textLoc[1]+2)])		# Draw bottom layer.

		# Draw buttons
		for b in self.buttons: self.buttons[b].draw(self.surfOpaque)

		# Apply alpha layering, if active.
		self.surfOpaque.set_alpha(self.opacity)
		self.surfOpaque.convert_alpha()

		# Copy the metaphorical sandwich back to the main surface.
		destSurf.blit(self.surfOpaque, self.getAbsRect())	# Blit secondary surface directly to main surface.
		vkos_globals.addUpdateRgn(self.getAbsRect())		# This is hacky for a translucent BG, but it works.
		self.redraw = False
		return



	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Handles mouse events
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		# No input while fading, in order to avoid a weird graphical
		# glitch that stops drawing everything but the statusBar UIE.
		if not (self.enabled): return

		# Determine affected UIE.
		#-------------------------
		curObj = None
		for c in self.clickable.keys():
			if (vkos_globals.isInRect(self.clickable[c], event.pos)):
				objName = c[3:]				# Internal object name.
				curObj = self.buttons[objName]		# SideMenu only has buttons.
		if not (curObj): return		# Exit if no object is found.


		# Process event struct
		#----------------------
		# Mousebutton DOWN
		if (event.type == pygame.MOUSEBUTTONDOWN):
			if (not (curObj.enabled)): return

			# Buttons get a "pressed down" look.
			if (curObj.type == "BUTTON"): curObj.setState(True)
			vkos_globals.SM.setActiveUIE(curObj)		# Set global reference.
			return


		# Mousebutton UP [Handle clicks here]
		#-------------------------------------
		if ((event.type == pygame.MOUSEBUTTONUP) and vkos_globals.SM.mouseCurUIE):
			if (not (curObj.enabled)): return	# If the button is disabled, exit here.

			# Buttons
			#---------
			if (curObj.type == "BUTTON"):
				curObj.setState(False)		# Revert button to normal state.
				self.redraw = True		# Redraw SideMenu for any button.

				# FBA
				if (objName == "FBA-Connect"): vkos_globals.CM.connectDevice("FBA")
				elif (objName == "FBA-Disconnect"): vkos_globals.CM.disconnectDevice("FBA")
				elif (objName == "FBA-Mute"):
					mState = vkos_globals.CM.isDeviceMuted("FBA")	# Toggle mute state.
					vkos_globals.CM.setMute("FBA", (not mState))


				# Thor
				elif (objName == "Thor-Connect"): vkos_globals.CM.connectDevice("Thor")
				elif (objName == "Thor-Disconnect"): vkos_globals.CM.disconnectDevice("Thor")
				elif (objName == "Thor-Mute"):
					mState = vkos_globals.CM.isDeviceMuted("Thor")	# Toggle mute state.
					vkos_globals.CM.setMute("Thor", (not mState))


				# EXIT
				elif (objName == "Main-Exit"):
					vkos_globals.powerOffAllDevices()		# Turn off all lighting.
					vkos_globals.SM.desktop.active = False		# Initiate exit animation.
		return



	#-------------------------------------------------------
	# setOpacity(newOpacity)
	# Desc: Adjust opacity for alpha layering.
	#-------------------------------------------------------
	# - newOpacity: Transparency level (0~255).
	#-------------------------------------------------------
	def setOpacity(self, newOpacity):
		# Sanity checks; Only positive numbers (0~255).
		if (newOpacity < 0): newOpacity = 0
		if (newOpacity > 255): newOpacity = 255
		self.opacity = newOpacity
		self.redraw = True
		return



	#-------------------------------------------------------
	# update()
	# Desc: Updates any timers/deltas in the UIE. The
	#       SideMenu has very little going on, so it's
	#       not problematic to just redraw it all instead
	#       of conditionally.
	#-------------------------------------------------------
	def update(self):
		if (vkos_globals.CM):
			# FBA button states.
			#--------------------
			if (vkos_globals.CM.isDeviceConnected("FBA")):
				# Connected
				self.buttons["FBA-Connect"].enabled = False
				self.buttons["FBA-Disconnect"].enabled = True
				self.buttons["FBA-Mute"].enabled = True

				# Update "Mute/Unmute" button text.
				#-----------------------------------
				muteText = "Unmute" if vkos_globals.CM.isDeviceMuted("FBA") else "Mute"
				self.buttons["FBA-Mute"].setText(muteText)
			else:
				# Disconnected
				self.buttons["FBA-Connect"].enabled = True
				self.buttons["FBA-Disconnect"].enabled = False
				self.buttons["FBA-Mute"].enabled = False
				self.buttons["FBA-Mute"].setText("Mute")	# Default.

			# Thor button states
			#--------------------
			if (vkos_globals.CM.isDeviceConnected("Thor")):
				# Connected
				self.buttons["Thor-Connect"].enabled = False
				self.buttons["Thor-Disconnect"].enabled = True
				self.buttons["Thor-Mute"].enabled = True
			else:
				# Disconnected
				self.buttons["Thor-Connect"].enabled = True
				self.buttons["Thor-Disconnect"].enabled = False
				self.buttons["Thor-Mute"].enabled = False

		# SideMenu has very little going on, so redraw everything
		# if any buttons need to redraw, instead of conditional drawing.
		for b in self.buttons.keys():
			if self.buttons[b].redraw == True: self.redraw = True
		return
