#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: UIE_Slider.py
#   - Version: 1.3
#   - Author: Kevin Slaughter
#   - Desc: UI Element >> Slider (aka: Scrollbar)
#
#
#   [Revision History]
#   2019/02/09: v1.3 - Commenting/Documentation standardization.
#   2019/02/04: v1.2 - General housekeeping/bugfixes.
#   2019/01/27: v1.1 - Vertical bars are finally supported.
#   2018/12/25: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, vkos_globals
from UIE_Base import *



# Constants
#-----------
LINE_LENGTH = 88 #112		# The scalar value is determined via this value.
LINE_WIDTH = 4
LINE_OFFSET_H = 10 #16
LINE_OFFSET_V = 5
SLIDER_H = (0, 25, 10, 20)	# Clickable slider, at zero-position, in horizontal mode.
SLIDER_V = (6, 20, 20, 10)	# ^               ^                ^, in vertical mode.



# Class Object
#--------------
class UIE_Slider(UIE_Base):
	# Initialization
	#----------------
	def __init__(self, dispLoc=[], aLoc=[], maxVal=255, numIndex=(-1), objName="UNKNOWN", orient="H", owner="", prefix="", startVal=0):
		# Sliders have relatively-hardcoded dimensions, so we only need coordinates.
		# Pre-render text, so we know how wide the slider needs to be (In vertical mode, height in horizontal mode).
		self.textSurf = vkos_globals.SM.fonts["Main"].render(str(prefix + str(maxVal)), True, [255, 255, 255], vkos_globals.TC_BG_SECTION)
		self.rectText = self.textSurf.get_rect()

		# Determine full UIE dimensions (via orientation).
		#--------------------------------------------------
		if (orient == "H"):
			dispRect = [dispLoc[0], dispLoc[1], ((SLIDER_H[2]*2) + LINE_LENGTH + 5), (self.rectText[3] + SLIDER_H[3] + 4)]
		elif (orient == "V"):
			dispRect = [dispLoc[0], dispLoc[1], self.rectText[2], ((SLIDER_V[3] / 2) + self.rectText[3] + (LINE_LENGTH + LINE_OFFSET_V) + 10)]

		# Set all other attributes.
		#---------------------------
		UIE_Base.__init__(self, dispRect, aLoc)     		# Call standard init for all UIE objects.
		self.actualVal = 0					# Start at 0, by default.
		self.curVal = 0						# ^          ^          ^
		self.displayPrefix = prefix				# Set text display prefix (eg: "G: ").
		self.enabled = True					# Sliders are enabled by default.
		self.maxVal = abs(maxVal)				# Set maximum value (Positive ONLY).
		self.name = "c" + str(numIndex) + "-slider" + objName	# Internal object name (Object type + name).
		self.owner = owner					# UIE host for slider.
		self.scalar = float(maxVal / float(LINE_LENGTH))	# Set scalar value to determine slider step values.
		self.setOrientation(orient)	               		# Set slider orientation (Default is horizontal).
		self.type = "SLIDER"					# Define UIE type.
		if (startVal > 0): self.setValue(startVal, bActualVal=True)	# Set default value, if specified.
		return



	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draw the slider itself and the numeric display.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Blank BG
		self.surfMain.fill(self.colors["BG"])

		# Render value display
		#----------------------
		self.textSurf = self.fontMain.render(str(self.displayPrefix + str(self.getDisplayVal())), True, vkos_globals.TC_FG_SECTION, vkos_globals.TC_BG_SECTION)
		self.rectText = self.textSurf.get_rect()
		if (self.orientation == "H"): self.surfMain.blit(self.textSurf, (2, 1))
		if (self.orientation == "V"): self.surfMain.blit(self.textSurf, (0, (LINE_LENGTH + LINE_OFFSET_V + 7)))

		# Calculate and draw centered slider "track".
		#---------------------------------------------
		if (self.orientation == "H"):		# Horizontal
			lineY = (self.rectSlider[1] + (self.rectSlider[3] / 2))
			linePos = ((LINE_OFFSET_H, lineY), ((LINE_OFFSET_H + LINE_LENGTH), lineY))
		if (self.orientation == "V"):		# Vertical
			linePos = ((LINE_OFFSET_H, LINE_OFFSET_V), (LINE_OFFSET_H, LINE_LENGTH))
		pygame.draw.line(self.surfMain, vkos_globals.TC_FRAME, linePos[0], linePos[1], LINE_WIDTH)

		# Draw slider
		#-------------
		# Fill in background and draw contoured frame.
		pygame.draw.rect(self.surfMain, vkos_globals.TC_FRAME, self.rectSlider)
		vkos_globals.drawContouredRect(self.surfMain, self.rectSlider, vkos_globals.TC_FRAME, False)

		# Copy onto destination surface.
		destSurf.blit(self.surfMain, self.rLoc)
		self.redraw = False
		return



	#-------------------------------------------------------
	# getAbsRect()
	# Desc: UIE_Base override; Returns a list with absolute
	#       screen position/dimensions. This way, we can
	#       easily avoid the text rendering areas.
	#-------------------------------------------------------
	def getAbsRect(self):
		# Horizontal
		if (self.orientation == "H"):
			return [self.aLoc[0], (self.aLoc[1]+self.rectText[3]), (self.size[0] + (SLIDER_H[2]*2)), self.size[1]]

		# Vertical
		elif (self.orientation == "V"):
			return [(self.aLoc[0] + (self.rectSlider[2]/3)), self.aLoc[1],
				(self.size[0] / 2), (LINE_LENGTH + (self.rectSlider[3] / 2))]



	#-------------------------------------------------------
	# getDisplayVal()
	# Desc: Returns the calculated (integer) value the
	#       UIE normally displays in the GUI.
	#-------------------------------------------------------
	def getDisplayVal(self):
		if (self.orientation == "H"): return int((self.curVal * self.scalar))
		else: return int((self.maxVal - (self.curVal * self.scalar)))



	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Mouse event handler.
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		if (self.mouseDown):
			# Determine which axis we're processing.
			if (self.orientation == "H"):   rPos = (event.pos[0] - self.aLoc[0] - LINE_OFFSET_H)
			elif (self.orientation == "V"): rPos = (event.pos[1] - self.aLoc[1] - LINE_OFFSET_V)

			# Adjust value via mouse position and ensure we're within boundaries.
			if ((rPos >= 0) and (rPos <= LINE_LENGTH)): self.setValue(rPos)
			elif (rPos < 0): self.setValue(0)
			elif (rPos > LINE_LENGTH): self.setValue(LINE_LENGTH)
		return



	#-------------------------------------------------------
	# setOrientation(iOrientation)
	# Desc: Set the display mode of the element.
	#       (Default: Horizontal)
	#-------------------------------------------------------
	# - cOrientation: "H" = Horizontal | "V" = Vertical.
	#-------------------------------------------------------
	def setOrientation(self, cOrientation="H"):
		# Sanity check; Make sure we get a valid state, or everything goes fucky.
		if not ((cOrientation == "H") or (cOrientation == "V")):
			vkos_globals.printLog("\nWhat in holy hell does '" + str(cOrientation) + "' mean for slider orientation?!", "!")
			quit()

		# Store new orientation and adjust drawing rects accordingly.
		#-------------------------------------------------------------
		if (cOrientation == "H"):
			self.rectSlider = list(SLIDER_H)	# Horizontal (Text above the line, no centering).
		else: self.rectSlider = list(SLIDER_V)		# Vertical (Text on top of line).
		self.orientation = cOrientation			# Save new orientation.
		self.setValue(self.curVal)			# Fix slider position.
		return



	#-------------------------------------------------------
	# setValue(v)
	# Desc: Set the current value of the slider element.
	#       This also triggers a redraw.
	#-------------------------------------------------------
	# -          v: Value, in terms of scalar.
	#               (self.actualVal = [v * self.scalar])
	# - bActualVal: Option to specify a specific value.
	#-------------------------------------------------------
	def setValue(self, v, bActualVal = None):
		v = abs(v)	# Positive numbers only.

		# If we got a specific end-result number, take
		# care of it here, before sanity-checking
		# sanitizes it for use with the scalar value.
		#----------------------------------------------
		if (bActualVal):
			if (v > self.maxVal): v = self.maxVal	# Don't exceed the defined maximum value.
			if (v > 0): v = (v / self.scalar)	# If it's zero, leave it alone.

		# Sanity check; Ensure the new value is within bounds.
		#------------------------------------------------------
		if (v > LINE_LENGTH): v = LINE_LENGTH		# Ensure upper boundary.
		self.curVal = v					# Update slider placement (Keep decimals)
		v = int(v)					# Calculate position with whole, positive numbers.

		# Calculate slider position.
		#----------------------------
		# Horizontal
		if (self.orientation == "H"):
			self.actualVal = int((self.curVal * self.scalar))
			self.rectSlider = ((LINE_OFFSET_H + int((v - (self.rectSlider[2] / 2)))),
					   self.rectSlider[1],
					   self.rectSlider[2],
					   self.rectSlider[3])

		# Vertical
		elif (self.orientation == "V"):
			# Invert the result so that 0 is at the bottom.
			self.actualVal = int((self.maxVal - (self.curVal * self.scalar)))
			self.rectSlider = (SLIDER_V[0],
					   int((v - (self.rectSlider[3] / 2)) + LINE_OFFSET_V),
					   SLIDER_V[2],
					   SLIDER_V[3])

		self.redraw = True	# Flag for redraw
		return



	#-------------------------------------------------------
	# update()
	# Desc: Update and check for any necessary redraws.
	#-------------------------------------------------------
	def update(self):
		# TODO: Unless I implement smooth transitions on slider bars,
		#       which is not a high priority at the moment, this will
		#       probably remain empty.
		return
