/* VisualK-OS (v5.0): FBA Software (COMMS Packet)
 * - File: COMMS_Packet.ino
 * - Version: 1.0
 * ---------------------------------------------
 * 
 * Revision History:
 * [2019/05/19] - v1.0: Initial framework.
 * --------------------------------------------- */



// HSV color struct.
struct color_HSV {
  unsigned int h;
  unsigned int s;
  float v;
};

// COMMS Packet object
struct COMMS_Packet {
  char cmd = COMMS_NOP;         // Packet command.
  unsigned int numParams = 0;   // Number of parameters.
  unsigned int params[1];       // Max 2 parameters.
  char paramTypes[1];           // CPACK_ARG_* indicators.
  color_HSV color;              // Any applicable color values.


  // Functions
  //-----------------------------------------------
  // addParam(pType, pVal)
  // Desc: Clear packet contents.
  //-----------------------------------------------
  // - pType: Type of parameter (CPACK_ARG_*) 
  // -  pVal: (Unsigned int) Parameter value.
  //-----------------------------------------------
  inline void COMMS_Packet::addParam(char pType, unsigned int pVal) {
    this->paramTypes[numParams] = pType;
    this->params[numParams] = pVal;
    if (this->numParams < 2) {this->numParams++;}
  }
  
  

  // makePacket(cBuffer, workingPacket)
  // Desc: Processes an accumulation buffer into
  //       a valid packet object for processing.
  //----------------------------------------------
  // - cBuffer: Serial accumulation buffer.
  //----------------------------------------------
  bool COMMS_Packet::makePacket(char* cBuffer) {
    unsigned int iNumBytes = 32; //sizeof(cBuffer);
        
    if (iNumBytes >= 4) {
      // Get location of end byte and verify packet boundaries.
      int iEndByte = 0;
      for (int a = 0; a <= 32; a++) {if (cBuffer[a] == CPACK_END) {iEndByte = a;}}
      if ((cBuffer[0] == CPACK_START) and (cBuffer[iEndByte] == CPACK_END)) {
        char iCmd = cBuffer[1];     // Command is always the first byte after CPACK_START.
        char iArgType = cBuffer[2]; // 3rd byte is the type of the first parameter.
        char iParam = cBuffer[3];   // 4th byte is the first parameter (usually the target lamp/bank/etc).
        
        // Set packet info.
        //------------------
        this->reset();                    // Clear packet.
        this->cmd = iCmd;                 // Set command.
        this->params[0] = iParam;         // First parameter is usually a lamp/bank.
        this->paramTypes[0] = iArgType;   // Declare parameter type.
        
        // Check for secondary parameters.
        //---------------------------------
        if (iNumBytes >= 7) {
          iArgType = cBuffer[4];            // Get secondary parameter type.
          this->paramTypes[1] = iArgType;
          
          // CPACK_ARG_COLOR:
          // Expects: [ASCII number][DELIMETER][ASCII number][DELIMETER][ASCII number][ARG_STOP]
          if (iArgType == CPACK_ARG_COLOR) {
            
            // Get new color.
            //----------------
            unsigned int valH = 0, valS = 0, valV = 0, i;
            i = 5;    // Adjust buffer position for first value.
            i += (this->readInt(cBuffer, i, valH));        // Get Hue.
            
            if (cBuffer[i] == CPACK_ARG_DELIMETER) {
              i++;      // Advance buffer.
              i += this->readInt(cBuffer, i, valS);      // Get Saturation.
              
              if (cBuffer[i] == CPACK_ARG_DELIMETER) {
                i++;    // Advance buffer.
                i += this->readInt(cBuffer, i, valV);    // Get Luminance.
                
                // Received all data; Set up packet.
                //-----------------------------------
                this->color = {.h = valH, .s = valS, .v = valV};
                return true;
              }
            }
          }
        
        
          // 1-byte parameters (byte/constant/color constant).
          //---------------------------------------------------
          if ((iArgType == CPACK_ARG_BYTE) or
              (iArgType == CPACK_ARG_CONST) or
              (iArgType == CPACK_ARG_COLOR_CONST)) {
                this->params[1] = cBuffer[5];
          }
          
          
          // Integer.
          //----------
          if (iArgType == CPACK_ARG_INT) {
            unsigned int iVal = 0;
            this->readInt(cBuffer, 5, iVal);
            this->params[1] = iVal;
            return true;
          }
        } // End secondary parameter check.
      }
      else {return false;}  // Error: Did not get CPACK_START/END as expected.
    }
    else {return false;}    // Error: Not enough bytes to be a valid packet.
  }
  
  

  // readInt(iStart)
  // Desc: Pulls an integer from the accumulator
  //       buffer. CPACK_ARG_DELIMETER, as well
  //       as CPACK_ARG_STOP, serve as stop
  //       indicators. Function returns the number
  //       of digits for proper buffer advancement.
  //-----------------------------------------------
  // - iStart: Buffer position at which to begin.
  //-----------------------------------------------
  int COMMS_Packet::readInt(char* iBuffer, unsigned int iStart, unsigned int &iOutput) {
    // Find number of digits.
    //------------------------
    int numDigits = 0;
    for (int a = iStart; a <= 32; a++) {
      if ((iBuffer[a] == CPACK_ARG_DELIMETER) or
          (iBuffer[a] == CPACK_ARG_STOP)) {break;}
      else {numDigits++;}
    }
    if (numDigits > 0) {
      char numBuffer[numDigits];    // Allocate only what is needed.
      
      // Get digits.
      //-------------
      for (int n = 0; n < numDigits; n++) {
        numBuffer[n] = iBuffer[iStart + n];
      }
      iOutput = atoi(numBuffer);    // Actual number.
    }
    else {Serial.write(COMMS_MSG); Serial.println("<COMMS_Packet> Failed to read int.");}
    
    return numDigits;             // Number of digits.
  }
  
  
  
  // reset()
  // Desc: Resets all packet properties.
  //----------------------------------------------
  inline void COMMS_Packet::reset() {
    this->cmd = COMMS_NOP;
    this->numParams = 0;
    this->params[0] = 0; this->params[1] = 0;
    this->paramTypes[0] = 0; this->paramTypes[1] = 0;
    this->color = {.h = 116, .s = 255, .v = 128};   // Default color (Half-dimmed green).
  }
};
