/* VisualK-OS (v5.0)
 * ---------------------------------------------
 *  - File: LightManager.ino
 *  - Version: 1.2
 *  - Desc: ShiftPWM-Based RGB LED Manager.
 * 
 * Revision History:
 * [2019/07/01] - v1.2: Removed all RGB usage,
 *                      all color is now stored
 *                      in HSV format.
 * [2019/06/16] - v1.1: Code simplificaiton.
 * [2019/05/19] - v1.0: Initial framework.
 * --------------------------------------------- */



// ShiftPWM-specific Declarations
//--------------------------------
const int ShiftPWM_latchPin = 9;
const bool ShiftPWM_invertOutputs = false;
const bool ShiftPWM_balanceLoad = true;
#include <ShiftPWM.h>               // Include AFTER pin declarations.



// Class object
class LightManager {
  private:
    // Lamp object definition
    // -----------------------
    struct objLamp {
      // General
      unsigned int index = 0;     // Addressable index.
      bool bMuted = false;        // Mute-determination flag.
      char iState = LAMP_OFF;     // Initial lamp state.
      color_HSV color = {.h = 220, .s = 255, .v = 0};   // Default color is Blue.
      
      // Timer-related
      bool bIsFading = false;           // Active fade indication flag.
      unsigned int iFadeIn = 500;       // Fade-in time (MS)
      unsigned int iFadeOut = 500;      // Fade-out time (MS)
      int iFadeDirection = (-1);        // (-1 = Fade out ; 1 = Fade in)
      unsigned long lFadeStep = 0;      // Current fade position.
      unsigned int iStepAmt = 0;        // Level of brigtness change per step.
      float fStepAmt = 0;
      unsigned long lStartMillis = 0;   // Starting time of fade.
    };
    
    
    // Vars
    //------
    char iDeviceID = COMMS_NOP;       // Invalidate device ID
    unsigned int numLamps = 0;        // Actual count of lamps.
    objLamp* lamps;                   // Array of lamp objects.
    unsigned long lLastMillis = 0;    // Last time update was called.
    
    // ShiftPWM-specific
    //-------------------
    unsigned char maxBrightness = 255;
    unsigned char pwmFrequency = 75;
    int numRegisters = 2;                 // Number of 74HC595 ICs.
    int numRGBleds = (numRegisters*8/3);  // Number of lamps.
    

    
  public:
    // beginFade(iLamp, iStep)
    // Desc: Initiates a fade in/out for a
    //       lamp based on the lamp's last
    //       known fade duration.
    //--------------------------------------
    // - iLamp: Lamp constant.
    // - iStep: Positive for fade in;
    //          Directly controls brightness.
    //--------------------------------------
    void LightManager::beginFade(char iLamp, int iFadeDirection) {
      unsigned int iFirstLamp = 0;    // First lamp to change.
      unsigned int iLastLamp = 0;     // Last lamp to change.
      
      // Determine lamps to modify.
      //----------------------------
      if ((iLamp == FBA_BANK_A) or (iLamp == FBA_BANK_B)) {
        // FBA Bank A: Lamps #0~3
        // FBA Bank B: Lamps #4~7
        iFirstLamp = (iLamp == FBA_BANK_A) ? 0 : 4;
        iLastLamp = (iLamp == FBA_BANK_A) ? 4 : 7;
      }
      else {
        // Set individual lamp.*/
        iFirstLamp = this->getLampIndex(iLamp);   // Get actual lamp index.
        iLastLamp = iFirstLamp;
      }
      
      // Apply to each lamp object.
      //----------------------------
      unsigned long lStartTime = millis();    // Sync all lamps, if more than one.
      for (unsigned int lampIndex = iFirstLamp; lampIndex <= iLastLamp; lampIndex++) {
        
        unsigned int iFadeDuration = (iFadeDirection) ? this->lamps[lampIndex].iFadeIn : this->lamps[lampIndex].iFadeOut;
        this->lamps[lampIndex].lFadeStep = 0;                         // Reset delta.
        this->lamps[lampIndex].iStepAmt = (float(255.0f / iFadeDuration) * 1000);    // Get amount of change per step.
        this->lamps[lampIndex].fStepAmt = (255.0f / iFadeDuration);
        this->lamps[lampIndex].iFadeDirection = iFadeDirection;       // Set fade direction.
        this->lamps[lampIndex].lStartMillis = lStartTime;             // Set start time (Same for all lamps).
        this->lamps[lampIndex].bIsFading = true;                      // Flag as active fade.
        this->lamps[lampIndex].color.v = (iFadeDirection) ? 0 : 255;  // Set start of fade (off or full brightness).
      }
    }
    
    
    
    // getLampColor(iLamp)
    // Desc: Returns an integer from a
    //       lamp constant.
    //-------------------------------------
    // - iLamp: Lamp constant.
    //-------------------------------------
    color_HSV LightManager::getLampColor(char iLamp) {
      int iLampIndex = this->getLampIndex(iLamp);
      return {.h = this->lamps[iLampIndex].color.h,
              .s = this->lamps[iLampIndex].color.s,
              .v = this->lamps[iLampIndex].color.v};
    }
    

    
    // getLampIndex(iLamp)
    // Desc: Returns an integer from a
    //       lamp constant.
    //-------------------------------------
    // - iLamp: Lamp constant.
    //-------------------------------------
    unsigned int LightManager::getLampIndex(char iLamp) {
      switch (iLamp) {
        case LAMP_0: return 0;
        case LAMP_1: return 1;
        case LAMP_2: return 2;
        case LAMP_3: return 3;
        case LAMP_4: return 4;
        case LAMP_5: return 5;
        case LAMP_6: return 6;
        case LAMP_7: return 7;
      }
    }
    
    
    
    // getSceneIndex(iScene)
    // Desc: Returns an integer from a
    //       scene constant.
    //------------------------------------
    unsigned int LightManager::getSceneIndex(char iScene) {
      switch (iScene) {
        case SCENE_0: return 0;
        case SCENE_1: return 1;
      }
    }
    
    
    
    // setup(iDeviceType)
    // Desc: Set up driver for x-device.
    //-------------------------------------
    // - iDeviceType: Device ID constant.
    //-------------------------------------
    LightManager::setup(char iDeviceType) {
      // Determine device type/parameters.
      //-----------------------------------
      this->iDeviceID = iDeviceType;
      switch (iDeviceID) {
        // Flood Bank Array (FBA) definitions.
        case ID_FBA:
          this->numLamps = 8;
          this->numRegisters = 3;

        // Thor (Drum Lighting) definitions.
        case ID_THOR:
          this->numLamps = 8;
          this->numRegisters = 3;
      }
      
      // Set up internal index of lamps.
      //---------------------------------
      this->numRGBleds = ((this->numRegisters * 8) / 3);                    // Number of Lamp objects.
      this->lamps = (objLamp*) realloc(lamps, sizeof(objLamp)*(numLamps));  // Dimension lamp object array.
      for (int lamp=0; lamp < this->numLamps; lamp++) {                     // Set default fade times and color.
        this->lamps[lamp].iFadeIn = 120;
        this->lamps[lamp].iFadeOut = 120;
        this->lamps[lamp].color = {.h = 220, .s = 255, .v = 0};             // Default color is Blue [220, 255, 0].
      }
      
      // Set up ShiftPWM library.
      //--------------------------
      ShiftPWM.SetAmountOfRegisters(this->numRegisters);
      ShiftPWM.Start(this->pwmFrequency, this->maxBrightness);
      ShiftPWM.SetAll(0);         // Turn off all lamps.
    }
    
    
    
    //----------------------------------------------
    // setAllHSV(iH, iS, iV)
    // Desc: Set all lamps to the given HSV color.
    //       Macro: Lamps turn on afterward.
    //----------------------------------------------
    // - iH: Hue.
    // - iS: Saturation.
    // - iV: Value/Brightness.
    //----------------------------------------------
    void LightManager::setAllHSV(unsigned int iH, unsigned int iS, unsigned int iV) {
      // Sanity-checks; HSV values must be 0~360 (H), and 0~255 (S and V).
      if (iH > 360) {iH = 360;}
      if (iS > 255) {iS = 255;}
      if (iV > 255) {iV = 255;}

      // Set all lamp object colors.
      for (unsigned int s = 0; s < this->numLamps; s++) {
        this->lamps[s].color = {.h = iH, .s = iS, .v = iV};
      }
      
      // Apply color and activate.
      ShiftPWM.SetAllHSV(iH, iS, iV);
    }
    
    

    // setBrightness(iLamp, b)
    // Desc: Set the brightness of a lamp/bank
    //----------------------------------------------
    void LightManager::setBrightness(char iLamp, unsigned int b) {
      unsigned int iFirstLamp = 0;    // First lamp to change.
      unsigned int iLastLamp = 0;     // Last lamp to change.
      
      // Determine lamps to modify.
      //----------------------------
      if (iLamp == ALL_LAMPS) {
          iFirstLamp = 0;
          iLastLamp = 7;
      }
      else if ((iLamp == FBA_BANK_A) or (iLamp == FBA_BANK_B)) {
        // FBA Bank A: Lamps #0~3
        // FBA Bank B: Lamps #4~7
        iFirstLamp = (iLamp == FBA_BANK_A) ? 0 : 4;
        iLastLamp = (iLamp == FBA_BANK_A) ? 4 : 7;
      }
      else {
        // Set individual lamp.
        iFirstLamp = this->getLampIndex(iLamp);   // Get actual lamp index.
        iLastLamp = iFirstLamp;
      }

      // Apply to each lamp object.
      //----------------------------
      for (unsigned int lampIndex = iFirstLamp; lampIndex <= iLastLamp; lampIndex++) {
        this->lamps[lampIndex].color.v = b;                               // Set new brightness.
        this->lamps[lampIndex].iState = (b > 0) ? LAMP_ON : LAMP_OFF;     // Set lamp object state.
        ShiftPWM.SetHSV(lampIndex, this->lamps[lampIndex].color.h, this->lamps[lampIndex].color.s, b);
      }
    }


    
    // setColor(iLamp, col)
    // Desc: Set the color of a lamp/bank.
    //----------------------------------------------
    // - iLamp: Lamp/Bank constant.
    // -   col: color_HSV Struct.
    //-----------------------------------------------
    void LightManager::setColor(char iLamp, color_HSV col) {
      unsigned int iFirstLamp = 0;    // First lamp to change.
      unsigned int iLastLamp = 0;     // Last lamp to change.
      
      // Determine lamps to modify.
      //----------------------------
      if (iLamp == ALL_LAMPS) {
          iFirstLamp = 0;
          iLastLamp = 7;
      }
      else if ((iLamp == FBA_BANK_A) or (iLamp == FBA_BANK_B)) {
        // FBA Bank A: Lamps #0~3
        // FBA Bank B: Lamps #4~7
        iFirstLamp = (iLamp == FBA_BANK_A) ? 0 : 4;
        iLastLamp = (iLamp == FBA_BANK_A) ? 4 : 7;
      }
      else {
        // Set individual lamp.
        iFirstLamp = this->getLampIndex(iLamp);   // Get actual lamp index.
        iLastLamp = iFirstLamp;
      }

      // Apply to each lamp object.
      //----------------------------
      for (unsigned int lampIndex = iFirstLamp; lampIndex <= iLastLamp; lampIndex++) {
        this->lamps[lampIndex].color.h = col.h;
        this->lamps[lampIndex].color.s = col.s;
        this->lamps[lampIndex].color.v = col.v;
        ShiftPWM.SetHSV(lampIndex, this->lamps[lampIndex].color.h, this->lamps[lampIndex].color.s, this->lamps[lampIndex].color.v);
      }
    }
    
    
    
    // setScene(iScene)
    // Desc: Change all lamp colors to those of a
    //       scene definition.
    //----------------------------------------------
    // - iScene: Scene index.
    //----------------------------------------------
    void LightManager::setScene(char iScene) {
      switch (iScene) {
        case SCENE_0:   // Warm tones: Red, orange, yellow, cyan.
          this->lamps[0].color.h = 0; this->lamps[0].color.s = 255;
          this->lamps[1].color.h = 20; this->lamps[1].color.s = 255;
          this->lamps[2].color.h = 20; this->lamps[2].color.s = 255;
          this->lamps[3].color.h = 40; this->lamps[3].color.s = 255;
          this->lamps[4].color.h = 40; this->lamps[4].color.s = 255;
          this->lamps[5].color.h = 20; this->lamps[5].color.s = 255;
          this->lamps[6].color.h = 20; this->lamps[6].color.s = 255;
          this->lamps[7].color.h = 0; this->lamps[7].color.s = 255;
          break;
        case SCENE_1:   // Cool tones: Purple, Lavendar, Cyan, Purple
          this->lamps[0].color.h = 280; this->lamps[0].color.s = 255;
          this->lamps[1].color.h = 220; this->lamps[1].color.s = 255;
          this->lamps[2].color.h = 180; this->lamps[2].color.s = 255;
          this->lamps[3].color.h = 280; this->lamps[3].color.s = 255;
          this->lamps[4].color.h = 280; this->lamps[4].color.s = 255;
          this->lamps[5].color.h = 220; this->lamps[5].color.s = 255;
          this->lamps[6].color.h = 180; this->lamps[6].color.s = 255;
          this->lamps[7].color.h = 280; this->lamps[7].color.s = 255;
          break;
      }
      
      // Reset the color of any active lamp.
      //-------------------------------------
      for (int lamp = 0; lamp < this->numLamps; lamp++) {
        if (this->lamps[lamp].iState == LAMP_ON) {
          ShiftPWM.SetHSV(lamp, this->lamps[lamp].color.h,
                                this->lamps[lamp].color.s,
                                this->lamps[lamp].color.v);
        }
      }
    }
    
    
    
    // turnOff(iLamp)
    // Desc: Turn off a given lamp/bank.
    //----------------------------------------------
    // -         iLamp: Lamp/Bank constant.
    // - iFadeDuration: Duration (in MS) of fade out.
    //----------------------------------------------
    void LightManager::turnOff(char iLamp, unsigned int iFadeDuration = 0) {
      unsigned int iFirstLamp = 0;    // First lamp to change.
      unsigned int iLastLamp = 0;     // Last lamp to change.

      // Determine lamp(s) to modify.
      //----------------------------
      if (iLamp == ALL_LAMPS) {
          iFirstLamp = 0;
          iLastLamp = 7;
      }
      else if ((iLamp == FBA_BANK_A) or (iLamp == FBA_BANK_B)) {
        // FBA Bank A: Lamps #0~3
        // FBA Bank B: Lamps #4~7
        iFirstLamp = (iLamp == FBA_BANK_A) ? 0 : 4;
        iLastLamp = (iLamp == FBA_BANK_A) ? 4 : 7;
      }
      else {
        // Individual lamp.
        iFirstLamp = this->getLampIndex(iLamp);
        iLastLamp = iFirstLamp;
      }
      
      // Apply to each lamp object.
      //----------------------------
      for (unsigned int lampIndex = iFirstLamp; lampIndex <= iLastLamp; lampIndex++) {
          this->lamps[lampIndex].iState = LAMP_OFF;
          ShiftPWM.SetHSV(lampIndex, 0, 0, 0);
      }
    }
    
    
    
    // turnOn(iLamp)
    // Desc: Turn on a given lamp/bank.
    //----------------------------------------------
    // -         iLamp: Lamp/Bank constant.
    // - iFadeDuration: Duration (in MS) of fade in.
    //----------------------------------------------
    void LightManager::turnOn(char iLamp, unsigned int iFadeDuration = 0) {      
      unsigned int iFirstLamp = 0;    // First lamp to change.
      unsigned int iLastLamp = 0;     // Last lamp to change.
      
      // Determine lamp(s) to modify.
      //----------------------------
      if (iLamp == ALL_LAMPS) {
          iFirstLamp = 0;
          iLastLamp = 7;
      }
      if ((iLamp == FBA_BANK_A) or (iLamp == FBA_BANK_B)) {
        // FBA Bank A: Lamps #0~3
        // FBA Bank B: Lamps #4~7
        iFirstLamp = (iLamp == FBA_BANK_A) ? 0 : 4;
        iLastLamp = (iLamp == FBA_BANK_A) ? 4 : 7;
      }
      else {
        // Individual lamp.
        iFirstLamp = this->getLampIndex(iLamp);
        iLastLamp = iFirstLamp;
      }
      
      // Apply to each lamp object.
      //----------------------------
      for (unsigned int lampIndex = iFirstLamp; lampIndex <= iLastLamp; lampIndex++) {
        this->lamps[lampIndex].iState = LAMP_ON;
        ShiftPWM.SetHSV(lampIndex,
                        this->lamps[lampIndex].color.h,
                        this->lamps[lampIndex].color.s,
                        this->lamps[lampIndex].color.v);
      }
    }
    
    
    
    // updateLamps()
    // Desc: Updates timer deltas and
    //       lamp states for fades.
    //--------------------------------
    void LightManager::updateLamps() {
      this->lLastMillis = millis();   // Set time for delta calculation.
      
      // Update any active brightness fades.
      //-------------------------------------
      for (unsigned int lamp = 0; lamp < this->numLamps; lamp++) {
        if (this->lamps[lamp].bIsFading) {
          // Update delta
          this->lamps[lamp].lFadeStep = abs(this->lLastMillis - this->lamps[lamp].lStartMillis);
          this->lamps[lamp].color.v = (this->lamps[lamp].fStepAmt * this->lamps[lamp].lFadeStep);
          
          // Sanity-check; Ensure value is within 0~255.
          if (this->lamps[lamp].color.v < 0) {this->lamps[lamp].color.v = 0;}
          if (this->lamps[lamp].color.v > 255) {this->lamps[lamp].color.v = 255;}
          
          // Adjust for fade direction (Which is conveniently used as a multiplier).
          if (this->lamps[lamp].iFadeDirection == (-1)) {
            this->lamps[lamp].color.v = (255 - this->lamps[lamp].color.v);
          }
          
          // If duration is exceeded, stop fading, go to full on/off.
          int iFadeDuration = (this->lamps[lamp].iFadeDirection) ? this->lamps[lamp].iFadeIn : this->lamps[lamp].iFadeOut;
          if (this->lamps[lamp].lFadeStep > iFadeDuration) {
            this->lamps[lamp].color.v = (this->lamps[lamp].iFadeDirection == 1) ? 255 : 0;
            this->lamps[lamp].iState = (this->lamps[lamp].color.v) ? LAMP_ON: LAMP_OFF;
            this->lamps[lamp].bIsFading = false;
          }
          
          // Apply new brightness level.
          ShiftPWM.SetHSV(lamp, this->lamps[lamp].color.h, this->lamps[lamp].color.s, this->lamps[lamp].color.v);
        }
      }

      // TODO: Update any active color fades.
    }
};
