/* VisualK-OS (v5.0)
 * ---------------------------------------------
 *  - File: VKOS_COMMS.ino
 *  - Version: 1.2
 *  - Desc: Standardized functions for serial
 *          COMMS functionality with the main
 *          VKOS system.
 * 
 * Revision History:
 * [2019/07/10] - v1.2: Simplified protocol and
 *                      packet creation/handling.
 * [2019/07/01] - v1.1: Removed RGB usage... All
 *                      colors are now stored in
 *                      HSV format for smoother
 *                      fade and better color
 *                      control.
 * [2019/05/19] - v1.0: Initial framework.
 * --------------------------------------------- */



class serCOMMS {
  // Vars
  // -----
  int iCount = 0;
  bool bHandshook = false;                // Handshake status.
  char iState = STATE_DISCONNECTED;       // Connection state.
  LightManager LM;                        // Lighting driver.
  bool bLMActive = false;                 // Prevent multiple instantiations.
  
  // Incoming serial data.
  //-----------------------
  static const unsigned int LEN_BUFFER = 32;    // Maximum buffer length.
  char iAccumulator[LEN_BUFFER];                // Accumulation buffer.
  unsigned int iNumBytes = 0;                   // Number of bytes placed in the buffer.
  bool bPacketStarted = false;                  // Flag: Packet started.
  bool bPacketEnded = false;                    // Flag: Current packet ended.
  
  
  
  public:
    // Functions
    //-----------------------------------------------
    // doHandshake(iID):
    // Desc: Establishes a connection to the main VKOS
    //       system using non-standard packets.
    //-----------------------------------------------
    // - iID: Device ID constant to handshake with.
    //-----------------------------------------------
    void serCOMMS::doHandshake(char iID) {
      // Initialize LightManager.
      //--------------------------
      if (!this->bLMActive) {       // Ensure we only init once.
        // Initalize light driver.
        //-------------------------
        this->LM.setup(iID);
        this->bLMActive = true;

        // Color channel test (Red, Green, Blue).
        //-----------------------------------------
        for (unsigned int c = 0; c < 3; c++) {
          // Fade in, fade out.
          for (unsigned int f = 0; f < 255; f+=3) {this->LM.setAllHSV((c*110), 255, f); delay(1);}
          for (unsigned int f = 255; f > 0; f-=3) {this->LM.setAllHSV((c*110), 255, f); delay(1);}
        }
        this->LM.turnOff(ALL_LAMPS);
      }
      
      
      // TODO: Change this to the control-center LEDs, instead of
      //       outputting through the entire rig.
      
      // Set up serial connection.
      //---------------------------
      Serial.begin(115200);             // Initialize connection.
      this->iState = STATE_CONNECTED;   // Set "Connected" flag.
      Serial.flush();                   // Clear buffer.
      
      // Stage #1: Send ID packet to main system.
      //------------------------------------------
      int iFader = 1; bool bFadeOut = false;
      bool bGotACK = false; bool bGotResponse = false;
      unsigned long iLastTime = millis(); unsigned long iTmp = 0;
      
      // Send ID packet.
      //-----------------
      COMMS_Packet tmpPacket;
      tmpPacket.cmd = COMMS_OP_CONNECT;
      tmpPacket.addParam(COMMS_DEVICE_ID, iID);
      this->sendPacket(tmpPacket); delay(20);
      
      // Check for response, otherwise, pulse LEDs.
      //--------------------------------------------
      while (!bGotResponse) {
        if (Serial.available() >= 3) {
          LM.setAllHSV(280, 255, 255); delay(100);
          bGotACK = this->waitACK();
          if(bGotACK) {bGotResponse = true;}
          else {
            // Error while waiting for ACK. Flash LEDs red.
            for (int err = 0; err < 3; err++) {
              this->LM.setAllHSV(0, 255, 64); delay(500);
              this->LM.turnOff(ALL_LAMPS); delay(500);
              err = 1;    // Infinite loop. Reset device to try again.
            }
          }
        }
        else {
          // No response; Pulse LEDs in yellow.
          //------------------------------------
          if (iLastTime == iTmp) {
            if (bFadeOut) {   // Fading out.
              iFader--;
              if (iFader <= 0) {iFader = 0; bFadeOut = false;}
            }
            else {            // Fading in.
              iFader++;
              if (iFader >= 125) {iFader = 125; bFadeOut = true;}
            }
            this->LM.setAllHSV(45, 255, iFader);    // Set LED brightness.
          }
        }
        
        // Update LED fade timer delta.
        //------------------------------
        iTmp = millis();
        if ((iTmp - iLastTime) > 40) {iLastTime = iTmp;}
      }
      
      
      // Stage #2: Received ACK packet, respond with double ACK.
      //--------------------------------------------------------
      if (bGotACK) {
        this->LM.setAllHSV(280, 255, iFader);
        tmpPacket.reset();              // Recycle packet.
        tmpPacket.cmd = COMMS_ACK;      // Respond with double-ACK.
        tmpPacket.addParam(CPACK_ARG_CONST, COMMS_ACK);
        this->sendPacket(tmpPacket);    // Alles gut! Set main handshake flag.
        
        // Flash green 3x to indicate a successful connection.
        for(int g=0; g<3; g++) {
          this->LM.setAllHSV(110, 255, 64); delay(20);
          this->LM.turnOff(ALL_LAMPS);
        }
      }
    }
    
    
    
    // handlePacket(p)
    // Desc: Processes the next packet in the queue.
    //-----------------------------------------------
    // - p: COMMS_Packet object.
    //-----------------------------------------------
    void serCOMMS::handlePacket(COMMS_Packet p) {
      int iFadeDirection = 1;
      
      // Parse based on command.
      //-------------------------
      switch (p.cmd) {
        // COMMS_OP_FADE_IN
        // Desc: Fade a bank/lamp in.
        //----------------------------
        case COMMS_OP_FADE_IN:
        case COMMS_OP_FADE_OUT:
          iFadeDirection = (p.cmd == COMMS_OP_FADE_IN) ? 1 : (-1);
          this->LM.beginFade(p.params[0], iFadeDirection);
          break;
        
        // COMMS_OP_OFF
        // Desc: Turn a bank/lamp off.
        //-----------------------------
        case COMMS_OP_OFF:
          this->LM.turnOff(p.params[0]);
          break;

        // COMMS_OP_ON
        // Desc: Turn a bank/lamp on.
        //-----------------------------
        case COMMS_OP_ON:
          this->LM.turnOn(p.params[0]);
          break;
        

        // COMMS_OP_SET_BRIGHTNESS
        // Desc: Set lamp brightness.
        //-----------------------------
        case COMMS_OP_SET_BRIGHTNESS:
          this->LM.setBrightness(p.params[0], p.params[1]);
          break;
        
        
        // COMMS_OP_SET_COLOR
        // Desc: Set lamp color.
        //-----------------------------
        case COMMS_OP_SET_COLOR:
          this->LM.setColor(p.params[0], p.color);
          break;


        // COMMS_OP_SET_SCENE
        // Desc: Set rig-wide lamp colors via pre-defined scenes.
        //-----------------------------
        case COMMS_OP_SET_SCENE:
          this->LM.setScene(p.params[0]);
          break;
      }
    }
    
    
    
    // isConnected()
    // Desc: Returns bool for connection state.
    //-----------------------------------------------
    bool serCOMMS::isConnected() {return bool(iState == STATE_CONNECTED);}
    
    
    
    // processIO()
    // Desc: Constructs packets from incoming serial
    //       data, via non-blocking code.
    //-----------------------------------------------
    void serCOMMS::processIO() {
      // Process the next received byte.
      //---------------------------------
      if (Serial.available()) {
        char inByte = Serial.read();    // Get byte.
        
        
        switch (inByte) {
          // Packet start
          //--------------
          case CPACK_START:
            // Begin a new packet. This also takes care of restarting
            // the process if we get an errant CPACK_START in the
            // middle of receiving another packet, where we would
            // reset anyway due to not knowing whether or not we
            // got all the data necessary for the previous packet.
            this->resetAccumulator();
            this->bPacketStarted = true;
            this->iAccumulator[0] = CPACK_START;
            this->iNumBytes = 1;
            break;
          
          // Packet end
          //------------
          case CPACK_END:
            // Parse the accumulator array.
            //------------------------------
            //  Byte #0: Command
            //  Byte #1: Parameter type
            //  
            //  * Data length varies by type after this.
            //    Parameters will lead the branch logic
            //    after this byte since we have reached
            //    a defined end of (presumed) grouped data.
            
            // Make sure we're in a known packet state.
            if ((this->bPacketStarted) and (!this->bPacketEnded)) {
              this->iAccumulator[this->iNumBytes] = CPACK_END;
              this->iNumBytes++;
              this->bPacketEnded = true;
            }
            else {
              // Got CPACK_END while in an unknown state... Clear and reset.
              this->resetAccumulator();
            }
            break;
          
          
          // Store all non-CPACK_START/END data.
          //-------------------------------------
          default:
            if (this->iNumBytes >= LEN_BUFFER) {
              // Buffer exceeded... Something went wrong.
              // No packet should ever get to this length.
              // Clear out accumulator and restart.
              this->resetAccumulator();
            }
            else {
              // Add incoming data to the accumulation buffer.
              this->iAccumulator[this->iNumBytes] = inByte;
              this->iNumBytes++;
            }
            break;
          }
        }
        
        
        // If we got the end of a packet, we can
        // now attempt to process the buffer.
        if (this->bPacketEnded) {
          COMMS_Packet p;
          if (p.makePacket(this->iAccumulator)) {
            this->handlePacket(p);
            //this->sendString("Created packet successfuly!");    // TO-DO: Change this to an acknowledgment byte for system-side checking.
          }
          else {
            this->sendString("Packet creation failed.");
            this->LM.setAllHSV(0, 255, 255); delay(3000);   // Indicate error with red LEDs (3sec pause).
            this->LM.turnOff(ALL_LAMPS);
          }
          
          // Reset after processing.
          this->resetAccumulator();
        }
    }
    
    
    
    // resetAccumulator()
    // Desc: Resets the serial accumulation buffer.
    //-----------------------------------------------
    void serCOMMS::resetAccumulator() {
      this->bPacketStarted = false; this->bPacketEnded = false;
      for (int a = 0; a <= LEN_BUFFER; a++) {this->iAccumulator[a] = 0;}
      this->iNumBytes = 0;
    }
    
    
    
    // sendPacket(cmdByte, p1, p2)
    // Desc: Send a packet (with up to 2 parameters)
    //       to the main system.
    //-----------------------------------------------
    // - cmdByte: Command.
    // -      p1: Parameter #1.
    // -      p2: Parameter #2.
    //-----------------------------------------------
    void serCOMMS::sendPacket(COMMS_Packet p) {
      Serial.write(CPACK_START);      // Start of packet.
      Serial.write(p.cmd);           // Send command.
      
      // Send parameters.
      if (p.numParams > 0) {
        for (int a = 0; a < p.numParams; a++) {
          // TODO: Handle data types accordingly.
          switch (p.cmd) {
            case COMMS_OP_CONNECT:
              Serial.write(p.paramTypes[a]);    // Arg type
              Serial.write(p.params[a]);        // Parameter value.
              break;

            default:
              Serial.write(p.paramTypes[a]);
              Serial.write(p.params[a]);
          }
        }
      }
      
      Serial.write(CPACK_END);        // End of packet.
    }
    
    
    
    // sendString(strText)
    // Desc: Sends a string to the main
    //       system for console display for
    //       more verbose communications.
    void serCOMMS::sendString(char* strText) {
      Serial.write(CPACK_START);
      Serial.write(COMMS_MSG);
      
      // Newline '\n' terminator, instead of CPACK_END,
      // since CPACK_END's value is a displayable character
      // and could occur within the message.
      Serial.println(strText);
    }
    
    
    
    // updateLights()
    // Desc: Wrapper to call the private LightManager
    //       update function.
    //-----------------------------------------------
    void serCOMMS::updateLights() {this->LM.updateLamps();}
    
    
    
    // waitACK()
    // Desc: Returns true if an acknowledgment
    //       packet is received. The Arduino should
    //       be the only one sending double ACKs,
    //       and even then, only during a handshake.
    //       Unlike the Python counterpart, this
    //       side only needs to look for one ACK.
    //-----------------------------------------------
    bool serCOMMS::waitACK() {
      unsigned int iRetryCount = 0; char inBytes[3];
      for (unsigned int iRetryCount = 0; iRetryCount < 10; iRetryCount++) {
        if (Serial.available() >= 3) {
          for (int i = 0; i < 3; i++) {inBytes[i] = Serial.read();}
          return bool((inBytes[0] == CPACK_START) and
                      (inBytes[1] == COMMS_ACK) and
                      (inBytes[2] == CPACK_END));
        }
      }
      return false;
    }
};
