#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objCOMMSManager.py
#   - Version: 1.6
#   - Author: Kevin Slaughter
#   - Desc: COMmunications Management System.
#
#
#   [Revision History]
#   2019/09/23: v1.6 - Better sanity-checks, etc.
#   2019/07/23: v1.5 - objDevice class moved to its own module.
#   2019/06/07: v1.4 - Simplified code.
#   2019/05/02: v1.3 - Added handshaking and some simplification.
#   2019/03/21: v1.2 - Implemented packet usage.
#   2019/02/09: v1.1 - Comments/Documentation standardization.
#   2019/01/14: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import time, vkos_globals
from COMMS_Packet import *
from objDevice import objDevice
from VKOS_Constants import *



# Class Object
#--------------
class objCOMMSManager(object):
	# Initialization
	#----------------
	def __init__(self):
		# Create device instances.
		#--------------------------
		self.devices = {}
		for dev in vkos_globals.POTENTIAL_DEVICES.keys():
			self.devices.update({dev : objDevice(dev, vkos_globals.POTENTIAL_DEVICES[dev][0])})
		return


	# Functions
	#-------------------------------------------------------
	# connectDevice(strDevice)
	# Desc: If not already connected, initiates a connection
	#       to the indicated device.
	#-------------------------------------------------------
	# - strDevice: Device to initiate connection with.
	#-------------------------------------------------------
	def connectDevice(self, strDevice):
		vkos_globals.addStatusMsg("Initiating serial connection to '" + strDevice + "'...", "*", bConsole=True)
		if (self.isDevice(strDevice)):
			# Show popup while connecting.
			#------------------------------
			if (vkos_globals.SM):
				vkos_globals.SM.showPopup(strTitle = ("COMMS Initialization (" + strDevice + ")"),
							  strIntroMsg = "Establishing serial connections...",
							  colorBG = [192, 192, 192],
							  colorText = [255, 255, 255],
							  dispRect = [vkos_globals.SCREEN_CENTER, vkos_globals.SCREEN_CENTER,
								      480, 500])

			if not (vkos_globals.emulateDevices):
				# Attempt connection.
				#---------------------
				try:
					if (self.devices[strDevice].connect()):				# Open port and handshake.
						self.devices[strDevice].state = STATE_CONNECTED		# Flag device as ready.
						strStatus = "'" + strDevice + "' successfully connected."
						if ((vkos_globals.SM) and (vkos_globals.SM.desktop)):	# Enable UI Section.
							vkos_globals.SM.desktop.sections[strDevice].enabled = True
					else:
						self.devices[strDevice].state = STATE_DISCONNECTED	# Handshake error.
						strStatus = "Error handshaking with '" + strDevice + "'!"
				except:
					# Error: Could not open serial port.
					strStatus = "Error initializing port for '" + strDevice + "'!"
			else:
				# If we're emulating devices, just report them as ready.
				#--------------------------------------------------------
				self.devices[strDevice].state = STATE_CONNECTED				# Set device state.
				strStatus = "(Emulated) '" + strDevice + "' successfully connected."	# Return msg.
				if ((vkos_globals.SM) and (vkos_globals.SM.desktop)):			# Enable UI Section, if active.
					vkos_globals.SM.desktop.sections[strDevice].enabled = True
					self.devices[strDevice].muted = False		# UI bug fix.

			# Terminate popup and update status console.
			#--------------------------------------------
			if (vkos_globals.SM): vkos_globals.SM.closePopup()	# Delete popup object.
			strPrefix = "!" if (strStatus[-1:] == "!") else "*"	# Determine message prefix via the ending char.
			vkos_globals.addStatusMsg(strStatus, strPrefix, iIndent=1, bConsole=True)
		else:
			# Device not found.
			vkos_globals.addStatusMsg("Cannot connect unknown device '" + strDevice + "'.", "!", 1, bConsole=True)
		return



	#-------------------------------------------------------
	# disconnectDevice(strDevice)
	# Desc: If connected, disconnect the indicated device.
	#-------------------------------------------------------
	# - strDevice: Device to disconnect from.
	#-------------------------------------------------------
	def disconnectDevice(self, strDevice):
		vkos_globals.addStatusMsg("Disconnecting '" + strDevice + "'...", "*")
		if (self.isDevice(strDevice)):
			curDevice = self.devices[strDevice]
			if (curDevice.comms or vkos_globals.emulateDevices):
				if not vkos_globals.emulateDevices:
					# Turn off all lights before disconnecting, or they
					# will be stuck in their current state (or Arduino-side automation).
					offPacket = COMMS_Packet(COMMS_OP_OFF, ALL_LAMPS)
					curDevice.sendPacket(offPacket); time.sleep(0.5)

				# Close connection and reset object class states.
				#-------------------------------------------------
				curDevice.comms = None		# Close serial connection.
				curDevice.muted = False		# Reset mute status.
				self.devices[strDevice].state = STATE_DISCONNECTED	# Set device state.
				strStatusMsg = "Disconnected " + ("(emulated) " if (vkos_globals.emulateDevices) else "") + "'" + strDevice + "'."
				vkos_globals.addStatusMsg(strStatusMsg, "-", 1)
			else:
				vkos_globals.addStatusMsg("'" + strDevice + "' has no active COMMS instance to disconnect.", "!", 1)
		else: vkos_globals.addStatusMsg("Cannot disconnect unknown device '" + strDevice + "'.", "!", 1)
		return



	#-------------------------------------------------------
	# getNumChannels(strDevice)
	# Desc: Returns the number of lamps a given device uses.
	#-------------------------------------------------------
	# - strDevice: Device to check.
	#-------------------------------------------------------
	def getNumChannels(self, strDevice):
		# Sanity-check; Ensure the device exists.
		if (self.isDevice(strDevice)):
			return self.devices[strDevice].numChannels
		return (-1)	# Device not found; Invalidate.



	#-------------------------------------------------------
	# intToLampConst(i)
	# Desc: Converts an integer to a lamp constant.
	#-------------------------------------------------------
	def intToLampConst(self, i):
		if (i == 0): return LAMP_0
		elif (i == 1): return LAMP_1
		elif (i == 2): return LAMP_2
		elif (i == 3): return LAMP_3
		elif (i == 4): return LAMP_4
		elif (i == 5): return LAMP_5
		elif (i == 6): return LAMP_6
		elif (i == 7): return LAMP_7
		else: return COMMS_NOP
		return



	#-------------------------------------------------------
	# isDevice(strDevice)
	# Desc: Checks if a given string matches any known
	#       "Potential Devices".
	#-------------------------------------------------------
	# - strDevice: String; Case-sensitive device name.
	#-------------------------------------------------------
	def isDevice(self, strDevice): return bool(strDevice in self.devices.keys())



	#-------------------------------------------------------
	# isDeviceConnected(strDevice)
	# Desc: Returns the status of a device's connection.
	#-------------------------------------------------------
	# - strDevice: The device we are checking.
	#-------------------------------------------------------
	def isDeviceConnected(self, strDevice):
		# Check the state of the specified device.
		#------------------------------------------
		if (self.isDevice(strDevice)):
			return bool(self.devices[strDevice].state == STATE_CONNECTED)
		else:
			vkos_globals.addStatusMsg("Cannot check connection state for unknown device '" + strDevice + "'.", "!")
		return False



	#-------------------------------------------------------
	# isDeviceMuted(strDevice)
	# Desc: Returns the status of a device's mute function.
	#-------------------------------------------------------
	# - strDevice: The device we are checking.
	#-------------------------------------------------------
	def isDeviceMuted(self, strDevice):
		# Check the state of the specified device.
		#------------------------------------------
		if (self.isDevice(strDevice)): return self.devices[strDevice].muted
		else: vkos_globals.addStatusMsg("Cannot check mute state for unknown device '" + strDevice + "'.", "!")
		return False



	#-------------------------------------------------------
	# processIO()
	# Desc: Checks the serial IO queue and creates packets.
	#-------------------------------------------------------
	def processIO(self):
		# Check all available devices for serial data.
		#----------------------------------------------
		for i in self.devices.keys():
			if (self.isDeviceConnected(i)):
				ioDev = self.devices[i].comms	        # Set serial device instance.
				bInvalidate = False			# Error flag.

				if (ioDev):
					# Loop over and accumulate any incoming serial data.
					#----------------------------------------------------
					while (ioDev.inWaiting()):
						inByte = ord(ioDev.read());	# Get next byte.
						#vkos_globals.addStatusMsg("['" + str(i) + "'] Got byte: '" + getConstStr(inByte) + "' (" + chr(inByte) + ")", "", 1)

						# String from the device (Debugging, mainly).
						#---------------------------------------------
						if (inByte == COMMS_MSG):
							# Interpret the rest of the packet as a string until a '\n'.
							strMsg = ioDev.read_until("\n");

							# Display message and reset buffer.
							vkos_globals.addStatusMsg("[Message from " + i + "] " + strMsg, "*")
							self.devices[i].resetBuffer()
							continue

						# Packet start
						#--------------
						elif (inByte == CPACK_START):
							# Reset accumulation buffer.
							self.devices[i].resetBuffer()
							self.devices[i].accBuffer.append(CPACK_START)
							self.devices[i].accNumBytes += 1

						# Packet end
						#------------
						elif (inByte == CPACK_END):
							self.devices[i].accBuffer.append(CPACK_END)

							# Process buffer into a packet, then add it to the device's queue.
							tmpPacket = self.devices[i].makePacket(self.devices[i].accBuffer)
							self.devices[i].addPacket(tmpPacket)	# Add packet to queue.
							self.devices[i].resetBuffer()		# Clear buffer.

							if (tmpPacket):
								vkos_globals.addStatusMsg("Created packet!", "*", 1)
							else:
								vkos_globals.addStatusMsg("Failed to create packet.", "!", 1)
								bInvalidate = True

						# All other data
						#----------------
						else:
							# Append to device's accumulation buffer.
							self.devices[i].accBuffer.append(inByte)
							self.devices[i].accNumBytes += 1

					# Handle any queued packets.
					self.devices[i].handlePacket()
		return



	#-------------------------------------------------------
	# sendDevicePacket
	# Desc: Send a packet to the specified COMMS device.
	#-------------------------------------------------------
	# - strDevice: Device to send packet to.
	# -         p: COMMS_Packet() object.
	#-------------------------------------------------------
	def sendDevicePacket(self, strDevice, p):
		if (self.isDevice(strDevice)):
			self.devices[strDevice].sendPacket(p)
			vkos_globals.addStatusMsg(" >>> " + strDevice, "PACKET", 1)
		else:
			vkos_globals.addStatusMsg("Cannot send packet to unknown device '" + strDevice + "'.", "!", 1)
		return



	#-------------------------------------------------------
	# setMute(strDevice, bMute)
	# Desc: Mutes/Unmutes the given device's lighting.
	#-------------------------------------------------------
	# - strDevice: Device to mute/unmute.
	# - bMute: Mute/unmute indicator.
	#-------------------------------------------------------
	def setMute(self, strDevice, bMute=True):
		if (self.isDevice(strDevice)):
			self.devices[strDevice].muted = bMute
			iOp = COMMS_OP_MUTE if (bMute) else COMMS_OP_UNMUTE
			strMsg = "Muted " if (bMute) else "Unmuted "
			mutePack = COMMS_Packet(iOp, ALL_LAMPS)
			self.devices[strDevice].sendPacket(mutePack)
			vkos_globals.addStatusMsg(strMsg + "'" + strDevice + "'.", "*")
		return
