#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objColorChannel.py
#   - Version: 1.6
#   - Author: Kevin Slaughter
#   - Desc: UI Element -> RGB channel configuration control.
#           Used for color selection/display and misc. changes
#           to channel properties (drum input sensitivity, etc).
#
#
#   [Revision History]
#   2019/08/31: v1.6 - Implemented ColorWheel UIE.
#   2019/02/27: v1.5 - Fixed weird bug that prevented the "REVERT"
#                      function from working properly.
#   2019/02/09: v1.4 - Commenting/Documentation standardization.
#   2019/02/04: v1.3 - General/Numerous changes/bugfixes/etc.
#   2019/01/03: v1.2 - My plan to have an object cleanly
#                      wrap around this class, which would
#                      make elements related to drum configuration
#                      on Thor, was just too much obfuscation and
#                      made the work-around code look weird.
#                      This functionality has been merged into
#                      this single class, to keep things tidy.
#   2019/01/01: v1.1 - Changed naming convention to obj.
#   2018/12/23: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, vkos_globals
from UIE_Base import *
from UIE_ColorWheel import UIE_ColorWheel
from UIE_Slider import UIE_Slider
from UIE_Button import UIE_Button



# Class Object
#--------------
class objColorChannel(UIE_Base):
	# Properties
	#------------
	isDrumChannel = False               # Indicates if the channel has a level slider.



	# Initialization
	#----------------
	def __init__(self, dispRect=None, aLoc=None, hasDrumLevel=False, numIndex=None, owner=None):
		UIE_Base.__init__(self, dispRect, aLoc)	# Call standard init for all UIE objects.
		self.locColorBlock = [10, 30, 120, 20]	# Size/Location of the channel's "color preview" block.
		self.brightness = 0			# Brightness level. Default = 0 (OFF).
		self.color = [220, 255, 255]		# Default = Blue.
		self.isDrumChannel = hasDrumLevel	# Determines if a drum trigger is present.
		self.lastColor = self.color		# Last-applied color.
		self.numChannel = numIndex		# RGB Channel index / Lamp #
		self.owner = owner			# UIE host for color channel.
		self.saturation = 255			# Color Saturation level. Default = 255.
		self.sliders = {}			# Slider UIEs present on this channel.

		strOwner = (self.owner + "-c" + str(numIndex))		# All hosted UIEs will have the same owner.
		# Create drum trigger slider (if needed).
		#-----------------------------------------
		if (self.isDrumChannel):	# Default Threshold level is ~1/3 of the max value.
			self.locColorBlock = [10, 30, 92, 20]
			self.locDrum = [(self.locColorBlock[0] + self.locColorBlock[2] - 5), 30]
			aLocDrum = [(self.aLoc[0] + self.locDrum[0]), (self.aLoc[1] + self.locDrum[1])]
			self.sliders.update({"DRUM" : UIE_Slider(numIndex=int(numIndex),
							    dispLoc=self.locDrum, aLoc=aLocDrum, orient="V",
							    maxVal=1024, startVal=500, objName="DRUM",
							    owner=strOwner, prefix="")})

		# Create ColorWheel UIE and color preview block.
		# Adjust color wheel/block ocation, if this channel
		# will use input from a drum trigger.
		#---------------------------------------------------
		if self.isDrumChannel: self.locColorWheel = [4, (self.locColorBlock[1] + self.locColorBlock[3] + 10), 95, 84]
		else: self.locColorWheel = [20, (self.locColorBlock[1] + self.locColorBlock[3] - 12), 100, 140]
		aLocColorWheel = [(self.aLoc[0] + self.locColorWheel[0]), (self.aLoc[1] + self.locColorWheel[1]), self.locColorWheel[2], self.locColorWheel[3]]
		self.colorWheel = UIE_ColorWheel(dispRect=self.locColorWheel, aLoc=aLocColorWheel, numIndex=int(numIndex), owner=strOwner)

		# Calculate positions/dimensions for slider UIEs, set base values and configure value display.
		#----------------------------------------------------------------------------------------------
		locSat = (self.locColorWheel[1] + self.locColorWheel[3] - 12)
		self.locSaturation = [14, (locSat+25)] if (self.isDrumChannel) else [14, locSat]
		aLocSaturation = [(self.aLoc[0] + self.locSaturation[0]), (self.aLoc[1] + self.locSaturation[1])]
		self.sliders.update({"SATURATION" : UIE_Slider(numIndex=int(numIndex),
				     dispLoc=self.locSaturation, aLoc=aLocSaturation, startVal=255,
				     objName="SATURATION", owner=strOwner, prefix="Saturation: ")})

		self.locBrightness = [14, (self.locSaturation[1] + self.sliders["SATURATION"].size[1] + 5)]
		aLocBrightness = [(self.aLoc[0] + self.locBrightness[0]), (self.aLoc[1] + self.locBrightness[1])]
		self.sliders.update({"BRIGHTNESS" : UIE_Slider(numIndex=int(numIndex),
							  dispLoc=self.locBrightness, aLoc=aLocBrightness,
							  objName="BRIGHTNESS", owner=strOwner, prefix="Brightness: ")})

		# Set up dimensions/locations of buttons.
		#-----------------------------------------
		self.rectBtnApply = [74, 262, 50, 25]		#  Apply: Applies changes to the active channel's color.
		self.rectBtnRevert = [12, 262, 55, 25]		# Revert: Returns the channel to the last-known updated color.

		# Create buttons.
		#-----------------
		# Button: 'Apply'
		aRectBtnApply = [(self.aLoc[0] + self.rectBtnApply[0]), (self.aLoc[1] + self.rLoc[1]+self.rectBtnApply[1]),
		                 self.rectBtnApply[2], self.rectBtnApply[3]]

		# Button: 'Revert'
		aRectBtnRevert = [(self.aLoc[0] + self.rectBtnRevert[0]), (self.aLoc[1] + self.rLoc[1]+self.rectBtnRevert[1]),
		                  self.rectBtnRevert[2], self.rectBtnRevert[3]]
		self.buttons = {"APPLY"  : UIE_Button("Apply", dispRect=self.rectBtnApply, aLoc=aRectBtnApply,
						      objName="APPLY", numIndex=str(numIndex), owner=strOwner),
		                "REVERT" : UIE_Button("Revert", dispRect=self.rectBtnRevert, aLoc=aRectBtnRevert,
						      objName="REVERT", numIndex=str(numIndex), owner=strOwner)}
		return



	# Functions
	#-------------------------------------------------------
	# draw()
	# Desc: Draws all elements of the... element.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Blank BG (Only the area this channel occupies) and draw outline frame.
		tmpRect = [0, 0, self.size[0], self.size[1]]
		self.surfMain.fill(self.colors["BG"], tmpRect)
		pygame.draw.rect(self.surfMain, self.colors["Frame"], tmpRect, 4)

		# Draw channel name
		surfColorText = self.fontMain.render("[ Channel  #" + str(self.numChannel) + " ]", True, self.colors["FG"], self.colors["BG"])
		textX = int((self.size[0] / 2) - (surfColorText.get_width() / 2))
		self.surfMain.blit(surfColorText, [textX, 5])

		# Get current color from slider/UIE and draw the color preview block.
		tmpColor = self.colorWheel.getColor()
		pygame.draw.rect(self.surfMain, vkos_globals.HSV2RGB(tmpColor[0], tmpColor[1]), self.locColorBlock)

		# Draw ColorWheel UIE (After drum slider, to avoid overlap).
		self.colorWheel.draw(self.surfMain)

		# Draw Brightness, Saturation (and drum trigger, if needed) sliders.
		# Note: Instead of a loop, draw these in a specific order for
		#       proper layering.
		for s in self.sliders.keys():
			self.sliders[s].draw(self.surfMain)

		# Draw buttons
		for b in self.buttons.keys(): self.buttons[b].draw(self.surfMain)

		# Copy everything back to the host Section.
		destSurf.blit(self.surfMain, self.rLoc)
		self.redraw = False
		return



	#-------------------------------------------------------
	# getColor()
	# Desc: Returns a tuple containing the channel's current
	#       active color. This is to say, the last color
	#       actually set by "APPLY" or automation.
	#-------------------------------------------------------
	def getColor(self): return self.color



	#-------------------------------------------------------
	# getRect()
	# Desc: Returns a tuple of the entire element's drawing
	#       area, relative to its host. We only draw each
	#       channel if it has updated, not all of them.
	#-------------------------------------------------------
	def getRect(self):
		return [self.rLoc[0], self.rLoc[1], self.size[0], self.size[1]]



	#-------------------------------------------------------
	# revert()
	# Desc: Reverts the channel's hue/saturation to the
	#       last applied values.
	#-------------------------------------------------------
	def revert(self):
		# Change color and saturation to their most recent values.
		self.setColor(self.lastColor)			# Set channel/colorWheel colors back to the most recent applied value.
		self.buttons["APPLY"].enabled = False		# Neither button is necessary after a channel revert.
		self.buttons["REVERT"].enabled = False
		self.redraw = True				# Redraw channel to refresh color preview block and buttons.
		return



	#-------------------------------------------------------
	# setButtons(self, bApply, bRevert)
	# Desc: Set the channel's button states. Specify/changes
	#       only the button intended.
	#-------------------------------------------------------
	# - bApply: State of channel's "Apply" button.
	# - bRevert: State of channel's "Revert" button.
	#-------------------------------------------------------
	def setButtons(self, bApply = None, bRevert = None):
		# Apply button states
		if ((bApply) and (isinstance(bApply, bool))): self.buttons["APPLY"].enabled = bApply
		if ((bRevert) and (isinstance(bRevert, bool))): self.buttons["REVERT"].enabled = bRevert
		return



	#-------------------------------------------------------
	# setColor(newColor)
	# Desc: Set the display color of the channel.
	#-------------------------------------------------------
	# - newColor: HSV Tuple.
	#-------------------------------------------------------
	def setColor(self, newColor):
		# Sanity checks; Ensure a valid color is given.
		newColor = [abs(newColor[0]), abs(newColor[1]), abs(newColor[2])]
		if (newColor[0] > 360): newColor = [360, newColor[1], newColor[2]]
		if (newColor[1] > 255): newColor = [newColor[0], 255, newColor[2]]
		if (newColor[2] > 255): newColor = [newColor[0], newColor[1], 255]

		# Assign the new color and save the old one.
		self.colorWheel.curColor = self.lastColor = self.color = newColor
		self.sliders["SATURATION"].setValue(newColor[1], bActualVal = True)
		self.buttons["APPLY"].enabled = self.buttons["REVERT"].enabled = False	# Disable buttons.
		self.redraw = True		# Flag channel for a redraw.
		return
