#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objDesktop.py
#   - Version: 1.5
#   - Author: Kevin Slaughter
#   - Desc: Screen object >> Desktop/UI Container.
#
#
#   [Revision History]
#   2019/09/22: v1.5 - Implemented alpha layering.
#   2019/09/09: v1.4 - Fixed dimming functionality.
#   2019/02/09: v1.3 - Commenting/Documentation standardization.
#                      Also added in StatusBar UIE.
#   2019/02/04: v1.2 - Addition of SideMenu UIE.
#   2019/01/08: v1.1 - Put UI sections into a dictionary,
#                      in order to simplify the main loop.
#   2018/12/16: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, vkos_globals
from objImage import objImage
from UIE_SideMenu import UIE_SideMenu
from objColorChannel import objColorChannel
from objDeviceSection import objDeviceSection
from objStatusBar import objStatusBar



# Class Object
#--------------
class objDesktop(object):
	# Initialization
	#----------------
	def __init__(self):
		self.active = False              # Main state indicator.
		self.redraw = True               # Flag to redraw (if needed).

		# Load BG image and SideMenu.
		#-----------------------------
		self.bg = objImage("_desktop.png")
		self.sMenu = UIE_SideMenu(vkos_globals.SIDEMENU_DIMENSIONS)
		self.sMenu.setOpacity(0)		# Fades in with deviceSection objects.
		self.sMenu.enabled = False		# Prevent a weird bug that causes a huge drawing glitch
							# if any sideMenu buttons are clicked while fading in.

		# Set up StatusBar object.
		#--------------------------
		self.statusBar = objStatusBar(vkos_globals.STATUSBAR_DIMENSIONS)

		# Load all deviceSection objects.
		#---------------------------
		self.sections = {}
		for dev in vkos_globals.POTENTIAL_DEVICES.keys():
			curDev = vkos_globals.POTENTIAL_DEVICES[dev]	# Shorthand, for readability.

			# Thor is the only device that uses drum triggers, so hasDrumInput will
			# always be false for any other deviceSection object.
			self.sections.update({dev : objDeviceSection(strName = dev,
								     strTitle = curDev[1],
								     dispRect = curDev[3],
								     hasDrumInput = bool(dev == "Thor"))})

		# Dimming flags/vars (Dimmer does not cover StatusBar).
		#-------------------------------------------------------
		self.surfDimmer = pygame.Surface([vkos_globals.SCREEN_SIZE[0], vkos_globals.STATUSBAR_DIMENSIONS[1]])
		self.surfDimmer.fill([0, 0, 0])		# Black overlay.
		self.dimmed = False			# Only used for focusing popups.
		self.dimAmt = 0				# Level of dimming.

		# Misc. setup
		#-------------
		self.active = True			# Mark Desktop object as active.
		self.bg.doFade(1000, 1)			# Fade-in BG.

		# Set up deviceSection objects for fading in.
		for f in self.sections.keys(): self.sections[f].setOpacity(0)
		return



	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draw all screen elements.
	#-------------------------------------------------------
	# - destSurf: Main screen object
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Draw desktop/background image first.
		#--------------------------------------
		if (self.bg.redraw):		# If we redraw the BG, we must redraw EVERYTHING.
			self.bg.draw(destSurf)
			self.sMenu.redraw = True
			for s in self.sections.keys(): self.sections[s].redraw = True

		# Force wait through BG fade-in and StatusBar slide-in.
		if (self.bg.animated): return	# Exit if animation is active.
		else:
			# Fade-in device sections immediately after BG fades in.
			for f in self.sections.keys():
				if (self.sections[f].opacity < 255):
					self.sections[f].setOpacity((self.sections[f].opacity + 20))
					self.sections[f].drawBG(destSurf)

			# Fade-in sideMenu UIE with all other elements.
			if (self.sMenu.opacity < 255):
				self.sMenu.setOpacity((self.sMenu.opacity + 15))
				self.sMenu.draw(destSurf)
			else: self.sMenu.enabled = True		# Enable after the fade in is complete.

			# Slide statusBar onto the display.
			if (self.statusBar.slideActive): return
			else:
				if not (self.statusBar.slideCompleted): self.statusBar.doSlide()	# Activate slide.
				else: self.statusBar.draw(destSurf)

		# Draw each section, as needed.
		#-------------------------------
		for s in self.sections:
			if (self.sections[s].redraw):
				self.sections[s].drawBG(destSurf)
				self.sections[s].drawFG(destSurf)
				self.sections[s].redraw = False

		# Draw SideMenu (on top).
		#-------------------------------------
		if (self.sMenu.redraw): self.sMenu.draw(destSurf)

		# Draw StatusBar.
		#------------------------
		if ((self.statusBar.redraw) or (self.redraw)):
			self.statusBar.draw(destSurf)		# Re-draw if the desktop is dimmed.

		# Apply dimming, if active.
		#---------------------------
		if (self.dimmed):
			vkos_globals.SM.screen.blit(self.surfDimmer, [0, 0])
			vkos_globals.addUpdateRgn([0, 0, vkos_globals.SCREEN_SIZE[0], vkos_globals.SCREEN_SIZE[1]])

		self.redraw = False
		return



	#-------------------------------------------------------
	# handleGP(b)
	# Desc: Gamepad event handler
	#-------------------------------------------------------
	# - b: Button pressed
	#-------------------------------------------------------
	def handleGP(self, b):
		# TODO: Handle gamepad events
		return



	#-------------------------------------------------------
	# handleKey(k)
	# Desc: Key event handler.
	#-------------------------------------------------------
	# -   k: Key being handled
	#-------------------------------------------------------
	def handleKey(self, k):
		if (k == pygame.K_q):
			vkos_globals.powerOffAllDevices()	# Turn off all lighting.
			self.active = False			# Initiate exit animation.
		return



	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Mouse event handler.
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		# Check for SideMenu events, as the SideMenu is not
		# handled like a Device Section.
		if (vkos_globals.isInRect(self.sMenu.getAbsRect(), event.pos)):
			self.sMenu.handleMouse(event)

		# Check for Device Section events.
		for i in self.sections.keys():
			if (vkos_globals.isInRect(self.sections[i].getRect(), event.pos)):
				if (self.sections[i].enabled):
					self.sections[i].handleMouse(event)
					return
		return



	#-------------------------------------------------------
	# setDimmer(bDimmed)
	# Desc: Dims the desktop to highlight a popup dialogue.
	#-------------------------------------------------------
	# - iAmt: Dimmer opacity (0 ~ 255).
	#-------------------------------------------------------
	def setDimmer(self, iAmt):
		# Sanity-checks; Amount of dimming must be 0 ~ 100.
		if (iAmt < 0): iAmt = 0
		if (iAmt > 255): iAmt = 255

		# Apply new dimming level.
		self.surfDimmer.set_alpha(iAmt)
		self.surfDimmer.convert_alpha()
		self.dimAmt = iAmt
		self.dimmed = True if (iAmt > 0) else False
		self.redraw = self.bg.redraw = True	# Redraw background.
		return



	#-------------------------------------------------------
	# update()
	# Desc: Updates any timers/etc and checks for redraws.
	#-------------------------------------------------------
	def update(self):
		# Update the BG image and SideMenu.
		#-----------------------------------
		self.bg.update()				# Update BG image.
		self.sMenu.update()				# Update SideMenu.
		if (self.sMenu.redraw):	self.redraw = True	# SideMenu transparency requires desktop redraw.

		# Update/Redraw the status bar object.
		#--------------------------------------
		self.statusBar.update()
		if (self.statusBar.redraw): self.statusBar.draw(vkos_globals.SM.screen)

		# Call update() for each Section object.
		#----------------------------------------
		for s in self.sections.keys():
			self.sections[s].update()	# Update any animated changes to the UI.

			# If we are redrawing the desktop, redraw EVERYTHING.
			if (self.redraw): self.sections[s].redraw = True

		# This class's redraw flag controls whether we draw any of
		# the Section objects it contains, as they are part of the
		# overall Desktop UIE. Therefore, if any of the Sections
		# update whatsoever, we must set this to True.
		if not self.redraw:
			self.redraw = bool((self.bg.redraw) or (self.sMenu.redraw) or
					   (self.sections["FBA"].redraw) or
					   (self.sections["Thor"].redraw))
		return
