#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objDevice.py
#   - Version: 1.0
#   - Author: Kevin Slaughter
#   - Desc: COMMS device object >> A representation of the
#           Arduino devices controlled by this system.
#
#
#   [Revision History]
#   2019/07/23: v1.0 - File created with initial framework.
#                      Contents were originally located in
#                      objCOMMSManager, but grew too large
#                      to not be in its own module.
#---------------------------------------------------------------



# Imports
#---------
import copy, os, serial, time, vkos_globals
from COMMS_Packet import *


# Constants
#-----------
COMMS_BAUD_RATE = 115200



# Object Definitions (Devices)
#------------------------------
class objDevice(object):
	# Initialization
	#----------------
	def __init__(self, strDeviceName, strDevicePath):
		self.comms = None				# Serial connection instance.
		self.muted = False				# Lighting mute state.
		self.name = strDeviceName			# Device name.
		self.packets = {}				# Packet queue.
		self.path = strDevicePath			# Path to the device.
		self.state = STATE_DISCONNECTED			# Connection indicator.
		self.numChannels = vkos_globals.POTENTIAL_DEVICES[strDeviceName][2]	# Number of operable lamps.

		# Device-independent serial accumulation buffer.
		self.accBuffer = []
		self.accNumBytes = 0
		return



	# Functions
	#-------------------------------------------------------
	# addPacket(p):
	# Desc: Add a packet to the device's processing queue.
	#-------------------------------------------------------
	# - p: COMMS_Packet() object.
	#-------------------------------------------------------
	def addPacket(self, p):
		if (p): self.packets.update({p.relativeTime : copy.deepcopy(p)})



	#-------------------------------------------------------
	# connect():
	# Desc: Attempt to establish a connection with the device.
	#-------------------------------------------------------
	def connect(self):
		vkos_globals.addStatusMsg("Initiating connection to '" + self.name + "'...", "*", 1)
		self.comms = serial.Serial(self.path, COMMS_BAUD_RATE)

		# Ensure we have a valid serial instance.
		#-----------------------------------------
		if (self.comms):
			vkos_globals.addStatusMsg("Found device '" + self.name + "' at '" + self.path + "'; Attempting handshake...", "*", 1)
		else:
			vkos_globals.addStatusMsg("Device '" + self.name + "' not found.")

		# Attempt Handshake.
		#--------------------
		if (self.doHandshake()):
			# Set default light states (Blue and OFF).
			self.sendPacket(COMMS_Packet(iCmd=COMMS_OP_SET_COLOR, iLamp=ALL_LAMPS, iParam=[CPACK_ARG_COLOR, [220, 255, 0]]))
			vkos_globals.addStatusMsg("Sent default light state [220, 255, 0].", "*", 1)
			return True
		else:
			return False



	# doHandshake():
	# Desc: Setup and sync device with main system.
	#-----------------------------------------------
	def doHandshake(self):
		if (self.comms):
			self.comms.flushInput()		# Clear serial I/O buffers.
			self.comms.flushOutput()
			bActive = True          	# Flag for staying within the handshake loop.
			bGotID = False          	# Flag to indicate a device ID was received.
			bHandshook = False      	# Flag for handshake success; Returned by function, at end.
			iSyncAcks = 0           	# Number of sync ACKs received.
			iCount = 0

			cBuff = []; iNumBytes = 0
			while(bActive):
				if(self.comms.inWaiting()):
					# Add next byte to accumulation buffer.
					cBuff.append(ord(self.comms.read())); iNumBytes += 1

					# Process buffer on packet end.
					if (cBuff[iNumBytes-1] == CPACK_END):
						activePacket = self.makePacket(cBuff)

						if (activePacket):
							if (activePacket.cmd == COMMS_OP_CONNECT):
								# Stage #1: Device ID and ACK response.
								#---------------------------------------
								vkos_globals.addStatusMsg("Stage #1: Device ID", "-")
								deviceID = activePacket.params[chr(COMMS_DEVICE_ID)]
								if (deviceID == ID_FBA): self.name = "FBA"
								if (deviceID == ID_THOR): self.name = "Thor"

								# Send response ACK packet (No parameters).
								vkos_globals.addStatusMsg("Device ID: " + self.name, "*", 1)
								ackPacket = COMMS_Packet(COMMS_ACK)
								self.sendPacket(ackPacket)

							elif (activePacket.cmd == COMMS_ACK):
								# Stage #2: Mars attacks... Expecting ACK-ACK packet.
								#-------------------------------------------------
								vkos_globals.addStatusMsg("Stage #2: Wait for double-ACK packet.", "-")
								if (activePacket.params[chr(CPACK_ARG_CONST)] == COMMS_ACK):
									vkos_globals.addStatusMsg("Got ACK-ACK; Mars attacks!", "*", 1)
									return True
								else:
									vkos_globals.addStatusMsg("Did not get ACK-ACK packet.", "!", 1)
						else:
							# Something went wrong...
							vkos_globals.addStatusMsg("Packet creation failed.", "ERROR", 1)

						# Reset accumulation buffer after processing.
						#---------------------------------------------
						del cBuff; cBuff = []; iNumBytes = 0
		return bHandshook



	#-------------------------------------------------------
	# handlePacket()
	# Desc: Process the next packet in the queue.
	#-------------------------------------------------------
	def handlePacket(self):
		if (len(self.packets) > 0):
			for p in self.packets.keys():
				print "['" + self.name + "' :: handlePacket()]"
				print str(self.packets[p])

				self.packets.pop(p)	# Delete handled packet.
				return			# Only handle one packet at a time.
		return



	#-------------------------------------------------------
	# makePacket(cBuffer)
	# Desc: Constructs a COMMS_Packet() object from an
	#       accumulation buffer. Returns None on error.
	#-------------------------------------------------------
	# - cBuffer: Tuple of buffered Serial input.
	#-------------------------------------------------------
	def makePacket(self, cBuffer):
		# Sanity-check; Ensure valid packet boundaries.
		iNumBytes = len(cBuffer)
		if (iNumBytes > 3):
			if ((cBuffer[0] == CPACK_START) and
			    (cBuffer[(iNumBytes-1)] == CPACK_END)):
				# Create packet and assign command.
				tmpPacket = COMMS_Packet(cBuffer[1])

				# Handshake packets are handled differently.
				if (cBuffer[1] == COMMS_OP_CONNECT):
					if (iNumBytes == 5):
						# Expected: [COMMS_DEVICE_ID][ID_*]
						tmpPacket.addParam(cBuffer[2], cBuffer[3])
						return tmpPacket


				# Check for parameters.
				#-----------------------
				if (iNumBytes < 5): return None
				iType = cBuffer[2]

				# 1-Byte parameters (bytes, constants, color constants).
				if ((iType == CPACK_ARG_BYTE) or
				        (iType == CPACK_ARG_COLOR_CONST) or
				        (iType == CPACK_ARG_CONST)):
					# Make sure there is a next byte.
					if (iNumBytes >= 5):
						tmpPacket.addParam(iType, cBuffer[3])
					else: return None


				# HSV color.
				elif (iType == CPACK_ARG_COLOR):
					if (iNumBytes >= 5):
						curVal = 0
						iH = ""; iS = ""; iV = ""
						for p in range(4, (iNumBytes-1)):
							# Parse string to form the full value.
							if ((not (cBuffer[p] == CPACK_ARG_DELIMETER)) and
						            (not (cBuffer[p] == CPACK_ARG_STOP))):
								# Buffer receives bytes, ensure we handle
								# the ASCII value, not the ordinal/decimal value.
								if (curVal == 0): iH += chr(cBuffer[p])
								if (curVal == 1): iS += chr(cBuffer[p])
								if (curVal == 2): iV += chr(cBuffer[p])
							else:
								# End of value... Convert to integers.
								if (curVal == 0): iH = int(iH)
								if (curVal == 1): iS = int(iS)
								if (curVal == 2): iV = int(iV)
								curVal += 1
						# Save color.
						tmpPacket.setColor([iH, iS, iV])


				# Integer.
				elif (iType == CPACK_ARG_INT):
					iVal = ""
					for p in range(4, (iNumBytes-1)):
						# Parse string to form the full value.
						if ((not (cBuffer[p] == CPACK_ARG_DELIMETER)) and
					            (not (cBuffer[p] == CPACK_ARG_STOP))):
							# Use ASCII value, not the ordinal/decimal value.
							iVal += chr(cBuffer[p])
						else: iVal = int(iVal)	# Convert to integer.

					# Save value.
					tmpPacket.addParam(CPACK_ARG_INT, iVal)

				# Return packet.
				return tmpPacket

		else: return None	# Error... Not enough bytes to make a valid packet.
		return None



	#-------------------------------------------------------
	# resetBuffer()
	# Desc: Resets the device's serial accumulation buffer.
	#-------------------------------------------------------
	def resetBuffer(self):
		del self.accBuffer; self.accBuffer = []
		self.accNumBytes = 0
		return



	#-------------------------------------------------------
	# sendPacket(p)
	# Desc: Sends a packet to the device.
	#-------------------------------------------------------
	# - p: COMMS Packet object.
	#-------------------------------------------------------
	def sendPacket(self, p):
		# Debug
		#vkos_globals.addStatusMsg("sendPacket() >>> ['" + self.name + "']", "*")
		print str(p)

		# If in emulation mode, just show that a packet was sent.
		if (vkos_globals.emulateDevices): return

		if ((self.comms) and (p)):
			# Start packet & set command.
			#-----------------------------
			self.comms.write(chr(CPACK_START))		# Start packet
			self.comms.write(chr(p.cmd))			# Send command
			print "   - Sent cmd: '" + str(p.cmd) + "' (" + getConstStr(p.cmd) + ")"


			# Send parameters, if needed.
			#-----------------------------
			if (len(p.params.keys()) > 0):
				for iParamType in p.params.keys():
					iParamType = int(ord(iParamType))               # Convert to int
					self.comms.write(str(chr(iParamType)))          # Declare type of next byte.
					print "   - Sent param (" + str(iParamType) + ") type: " + getConstStr(iParamType)

					# Byte/Constants/Color Constants: Send as-is.
					#---------------------------------------------
					if ((iParamType == CPACK_ARG_BYTE) or
					    (iParamType == CPACK_ARG_CONST) or
					    (iParamType == CPACK_ARG_COLOR_CONST)):
						outByte = int(p.params[str(chr(iParamType))])
						self.comms.write(str(chr(outByte)))          # Send const.
						print "   - Sent const/byte: '" + getConstStr(outByte) + "'"

					# Color: Sent as '(int)[delimeter](int)[delimiter](int)[STOP][END]'
					#-------------------------------------------------------------------
					if (iParamType == CPACK_ARG_COLOR):
						print "   - Sending color data..."
						n = int(p.color[0])

						self.comms.write(str(p.color[0]))		# H
						self.comms.write(str(chr(CPACK_ARG_DELIMETER)))
						self.comms.write(str(p.color[1]))		# S
						self.comms.write(str(chr(CPACK_ARG_DELIMETER)))
						self.comms.write(str(p.color[2]))		# V
						self.comms.write(str(chr(CPACK_ARG_STOP)))		# Flag end of arg.
						print "   - Sent [" + str(int(p.color[0])) + ", " + str(int(p.color[1])) + ", " + str(int(p.color[2])) + "]"

					# ColorEnvironment: Sent as a lamp count followed by a line of
					#                   normally-delimited color values, with each
					#                   group of values separated by a [CPACK_ARG_STOP].
					#--------------------------------------------------------------------
					if (iParamType == CPACK_ARG_COLOR_ENVIRONMENT):
						self.comms.write(data)

					# Int: Sent as string, instead of 4 bytes.
					#------------------------------------------
					if (iParamType == CPACK_ARG_INT):
						iData = str(p.params[str(iParamType)])
						self.comms.write(iData)				# Send string.

			# End packet
			#------------
			self.comms.write(str(chr(CPACK_END)))
		return



	#-------------------------------------------------------
	# waitACK(iNumAcks)
	# Desc: Waits for an acknowledgment (ACK) packet.
	#-------------------------------------------------------
	def waitACK(self, iNumAcks = 1):
		bGotAck = False; iRetryCount = 0
		while (not bGotAck):
			if (self.comms.inWaiting() >= 3):
				ackPacket = self.comms.read(3)
				print "ackPacket[0] = " + getConstStr(ord(ackPacket[0]))
				if (ord(ackPacket[iNumAcks]) == CPACK_END):
						print "   - CPACK_END: OK"
						if (iNumAcks == 1):
							print "Waiting on ACK"
							if (ord(ackPacket[0]) == COMMS_ACK):
								print "Got 1 ACK"
								return True
						elif (iNumAcks == 2):
							if ((ord(ackPacket[0]) == COMMS_ACK) and
							    (ord(ackPacket[1]) == COMMS_ACK)):
								print "Got 2 ACKs"
								return True
							else: return False
			else:
				time.sleep(0.1)
				iRetryCount += 1
		return False



	#-------------------------------------------------------
	# waitColor()
	# Desc: Processes/returns an anticipated color arg.
	#-------------------------------------------------------
	def waitColor(self, serInst):
		# Create buffers
		iCurColor = 0; strRcvColor = ["", "", ""]

		# Loop during next bytes to get color data.
		for iCurColor in range(0, 3):
			if (self.comms.inWaiting() > 0):
				tmpColor = self.comms.read(1)			# Get next byte.
				if (tmpColor == CPACK_ARG_DELIMETER):
					iCurColor += 1; continue		# Got delimeter, go to next slot.
				elif (tmpColor == CPACK_ARG_STOP):		# Got end of arg.
					vkos_globals.printLog("Color = " + str(strRcvColor), "*", 1)
					return strRcvColor
				else:
					# Store color value (Sent from device as string).
					strRcvColor[iCurColor] = str(tmpColor)
					print "   - Got color arg #" + str(iCurColor) + ": '" + str(strRcvColor) + "'"
			else:
				return None	# Something went wrong, no more bytes waiting.
		return None



	#-------------------------------------------------------
	# waitConst()
	# Desc: Processes/returns an anticipated constant arg.
	#-------------------------------------------------------
	def waitConst(self):
		if (self.comms.inWaiting() > 0): return ord(self.comms.read(1))
		else: return COMMS_NOP



	#-------------------------------------------------------
	# waitDeviceID()
	# Desc: Processes/returns an anticipated constant arg.
	#-------------------------------------------------------
	def waitDeviceID(self):
		bGotID = False; iRetryCount = 0
		while (not bGotID):
			print "   - waitDeviceID()-> In loop."
			if (self.comms.inWaiting() >= 4):
				print "   - waitDeviceID()-> inWaiting(): OK"
				# Check all packet properties.
				#------------------------------
				idPacketBytes = self.comms.read(4)
				if ((ord(idPacketBytes[0]) == COMMS_OP_CONNECT) and
				    (ord(idPacketBytes[1]) == COMMS_DEVICE_ID) and
				    (ord(idPacketBytes[3]) == CPACK_END)):
					iID = ord(idPacketBytes[2])
					if (iID == ID_FBA):    self.name = "FBA"
					elif (iID == ID_THOR): self.name = "Thor"
					return True
			else:
				time.sleep(0.1)		# Delay and retry.
				iRetryCount += 1
				if (iRetryCount > 10): return False
		return False
