#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objDeviceSection.py
#   - Version: 1.3
#   - Author: Kevin Slaughter
#   - Desc: Screen object >> Device Section (aka: Window).
#
#
#   [Revision History]
#   2019/09/23: v1.3 - Hardcoded modules for each device have
#                      been abstracted and absorbed into this
#                      single module for overall simplification
#                      and to reduce code redundancy.
#   2019/02/09: v1.2 - Commenting/Documentation standardization.
#   2019/02/04: v1.1 - General housekeeping.
#   2018/12/21: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, vkos_globals
from COMMS_Packet import COMMS_Packet
from objColorChannel import objColorChannel
from VKOS_Constants import *


# Class Object
#--------------
class objDeviceSection(object):
	# Initialization
	#----------------
	def __init__(self, dispRect=[10,10,10,10], strName="", strTitle="<No Title>", hasDrumInput=False):
		# Set up element colors and section fonts.
		#------------------------------------------
		self.aLoc = [dispRect[0], dispRect[1]]			# Absolute X/Y coordinates.
		self.channels = []					# Array/Tuple of lamp channels under this section.
		self.clickable = {}					# Dictionary of clickable regions within this section.
		self.deviceName = strName				# Device name (COMMS Manager reference).
		self.enabled = False					# Disable until device is connected.
		self.fontMain = vkos_globals.SM.fonts["Main"]		# Main font.
		self.fontTitleBar = vkos_globals.SM.fonts["TitleBar"]	# Titlebar font.
		self.hasDrumInput = hasDrumInput			# Determines if a channel responds to drum triggers.
		self.numChannels = vkos_globals.POTENTIAL_DEVICES[self.deviceName][2]	# Number of channels to display.
		self.opacity = 255					# Alpha layering.
		self.redraw = True					# Flag to indicate if the Section needs to be redrawn.
		self.size = [dispRect[2], dispRect[3]]			# Section dimensions.
		self.colors = {"BG"           : [12, 12, 12],
			       "Frame"        : [64, 64, 64],
			       "FG"           : [255, 255, 255],
			       "Disconnected" : vkos_globals.TC_TITLEBAR["Disconnected"],
			       "Muted"        : vkos_globals.TC_TITLEBAR["Muted"],
			       "TitleBar"     : vkos_globals.TC_TITLEBAR[self.deviceName]}

		# Ensure device is/isn't connected.
		#-----------------------------------
		if (vkos_globals.CM): self.enabled = vkos_globals.CM.isDeviceConnected(self.deviceName)

		# Set up background, client and TitleBar dimensions.
		#----------------------------------------------------
		self.rectBG = [0, 0, dispRect[2], dispRect[3]]
		self.rectTitleBar = [self.rectBG[0], self.rectBG[1], self.rectBG[2], 32]
		self.rectClient = [3, (self.rectTitleBar[3] + 3), (self.rectBG[2] - 3), (self.rectBG[3] - 3)]

		# Create surfaces.
		#------------------
		self.surfBG = pygame.Surface([dispRect[2], dispRect[3]])    				# BG surface.
		self.surfClient = pygame.Surface([self.rectClient[2], self.rectClient[3]])		# Client surface.
		self.surfTitleBar = pygame.Surface([self.rectTitleBar[2], self.rectTitleBar[3]])	# Titlebar surface.
		self.title = strTitle									# Titlebar text.

		# Set up all UIE controls (150x300).
		#------------------------------------
		# Relative working-area rect.
		tempRect = (self.rectClient[0], self.rectClient[1], vkos_globals.CHANNEL_SIZE[0], vkos_globals.CHANNEL_SIZE[1])

		# Create an objColorChannel for each lamp/channel on the device.
		for i in range(0, self.numChannels):
			# Create channel and allow for 2px of space between each.
			#---------------------------------------------------------
			tempRect = [(self.rectClient[0] + 2 + (i*vkos_globals.CHANNEL_SIZE[0]) + (i*vkos_globals.CHANNEL_SPACER)), tempRect[1], tempRect[2], tempRect[3]]
			aRect = [(self.aLoc[0] + tempRect[0]), (self.aLoc[1] + tempRect[1]), tempRect[2], tempRect[3]]
			self.channels.append(objColorChannel(numIndex = int(i), dispRect = tempRect, aLoc = aRect,
							     hasDrumLevel = hasDrumInput, owner = self.deviceName))

			# Set new object ID info.
			#-------------------------
			cc = self.channels[i]						# Use "shorthand", for code readbility.
			for b in cc.buttons.keys(): cc.buttons[b].enabled = False	# Channel buttons disabled on init.


			# Define clickable areas.
			#-------------------------
			actualX = (self.aLoc[0] + cc.rLoc[0])	# Base X coordinate
			actualY = (self.aLoc[1] + cc.rLoc[1])	# Base Y coordinate

			# Button: Apply
			cc.buttons["APPLY"].aLoc = [(actualX + cc.rectBtnApply[0]), (actualY + cc.rectBtnApply[1]),
						    cc.rectBtnApply[2], cc.rectBtnApply[3]]
			self.clickable.update({cc.buttons["APPLY"].name : cc.buttons["APPLY"].getAbsRect()})

			# Button: Revert
			cc.buttons["REVERT"].aLoc = [(actualX + cc.rectBtnRevert[0]), (actualY + cc.rectBtnRevert[1]),
								     cc.rectBtnRevert[2], cc.rectBtnRevert[3]]	# Set absolute location.
			self.clickable.update({cc.buttons["REVERT"].name : cc.buttons["REVERT"].getAbsRect()})

			# ColorWheel UIE
			#----------------
			self.clickable.update({cc.colorWheel.name : cc.colorWheel.getAbsRect()})

			# Sliders
			#---------
			for s in cc.sliders.keys():
				sPos = cc.sliders[s].getLoc()
				sPos = [(actualX + sPos[0]), (actualY + sPos[1]), sPos[2], sPos[3]]
				self.clickable.update({cc.sliders[s].name: sPos})
		return



	# Functions
	#-------------------------------------------------------
	# drawBG(destSurf)
	# Desc: Standardized frame/section drawing.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def drawBG(self, destSurf):
		# Blank Section's BG surface.
		#-----------------------------
		self.surfBG.fill(self.colors["BG"])

		# Draw the Title Bar first. Color gradients
		# are dependent on the state of the device.
		#-------------------------------------------
		colorTitleBar = self.colors["TitleBar"]
		if (vkos_globals.CM.isDeviceMuted(self.deviceName) and vkos_globals.CM.isDeviceConnected(self.deviceName)):
			colorTitleBar = self.colors["Muted"]
		elif not vkos_globals.CM.isDeviceConnected(self.deviceName):
			colorTitleBar = self.colors["Disconnected"]
		vkos_globals.drawGradient(self.surfTitleBar, colorTitleBar, self.rectBG, False, True)

		# Title Bar Text. Since TileBars use gradients,
		# allow alpha antialiasing (no BG color specification).
		#-------------------------------------------------------
		textTitle = self.title
		if (vkos_globals.CM.isDeviceMuted(self.deviceName)): textTitle = "[MUTED]   " + textTitle + "   [MUTED]"
		surfTitleText = self.fontTitleBar.render(textTitle, True, vkos_globals.TC_TITLEBAR_TEXT)
		tempRectT = surfTitleText.get_rect()        # Get dimensions of rendered text.
		tempRectB = self.surfTitleBar.get_rect()    # Get dimensions of Title Bar.
		tempRectB[0] = ((tempRectB[2]/2)-(tempRectT[2]/2))      # Center text within Title Bar.
		tempRectB[1] = (tempRectB[3]/5)                         # Approximate vertical centering.
		self.surfTitleBar.blit(surfTitleText, tempRectB)        # Copy the text onto the Title Bar surface.

		# Assemble
		#----------
		self.surfBG.blit(self.surfTitleBar, self.rectTitleBar)                      # Title Bar onto BG...
		pygame.draw.rect(self.surfBG, vkos_globals.TC_FRAME, self.rectBG, 6)        # Draw frames on top of all layers.
		pygame.draw.rect(self.surfBG, vkos_globals.TC_FRAME, self.rectTitleBar, 4)

		# Apply transparency.
		#---------------------
		self.surfBG.set_alpha(self.opacity)
		self.surfBG.convert_alpha()

		# Copy back to the destination surface.
		#---------------------------------------
		destSurf.blit(self.surfBG, self.getRect())		# BG onto destSurf...
		vkos_globals.addUpdateRgn(self.getRect())		# Add display region to update list.
		return



	#-------------------------------------------------------
	# drawFG()
	# Desc: Draws all UI elements.
	#-------------------------------------------------------
	# - destSurf: Surface to draw Client area onto.
	#-------------------------------------------------------
	def drawFG(self, destSurf):
		# Connection status display overrides everything.
		#-------------------------------------------------
		# If device is not connected, overlay status.
		if (self.opacity < 255): return			# No text during a fade, it causes oddities.
		showErrorMsg = True if (not vkos_globals.CM) else False
		if ((vkos_globals.CM)):
			if not vkos_globals.CM.isDeviceConnected(self.deviceName):
				showErrorMsg = True

		# Display error message, if needed.
		if (showErrorMsg):
			self.surfClient.fill(vkos_globals.TC_BG_SECTION)	# Blank client area of section.
			vkos_globals.SM.drawText(self.deviceName + " is not connected!",
						 [((self.size[0] / 2) - 75), (self.aLoc[1] + (self.size[1] / 2))],
						 [128, 0, 128])
			return

		# Draw channels, if a redraw is needed.
		for i in range(0, vkos_globals.POTENTIAL_DEVICES[self.deviceName][2]):
			if (self.channels[i].redraw):
				# Draw channel.
				self.channels[i].draw(self.surfClient)

				# Add each updated element drawing area to the update list,
				# adjusting for actual screen position (rLoc[] is relative).
				# *** Since we're maintaining multiple dynamically-updating components,
				#     we only redraw a channel if some part of it has changed.
				tempRect = self.channels[i].getRect()
				tempRect = [(tempRect[0] + self.aLoc[0]), (tempRect[1] + self.aLoc[1]), tempRect[2], tempRect[3]]
				destSurf.blit(self.surfClient, tempRect, area=self.channels[i].getRect())	# Blit back to host.
				vkos_globals.addUpdateRgn(tempRect)		# Add screen update region.
				self.channels[i].redraw = False			# Mark as done.
			else:
				# No need to redraw, just copy from the previous drawing.
				tempRect = self.channels[i].getRect()
				tempLoc = [(tempRect[0] + self.aLoc[0]), (tempRect[1] + self.aLoc[1]), tempRect[2], tempRect[3]]
				destSurf.blit(self.channels[i].surfMain, tempLoc)	# Blit previously-drawn channel.
				vkos_globals.addUpdateRgn(tempLoc)			# Add screen update region.
		return



	#-------------------------------------------------------
	# getCtl(mpos)
	# Desc: Determines what UI Element is under the cursor.
	#       Returns a list containing the channel index,
	#       object name and a reference to the selected
	#       UI Element. If nothing is found, return type
	#       is None.
	#-------------------------------------------------------
	# - mpos: Tuple containing the coordinates to check.
	#-------------------------------------------------------
	def getCtl(self, mpos):
		# Iterate through the section's clickable regions.
		for ctl in self.clickable.keys():
			if (vkos_globals.isInRect(self.clickable[ctl], mpos)):
				strCtlName = ctl[3:]			# Get control name.
				channelIndex = int(str(ctl)[:2][1])     # Get channel index from dictionary key.

				# Assign object from lookup table.
				cc = self.channels[channelIndex]
				if (strCtlName == "btnAPPLY"):           clickedObj = cc.buttons["APPLY"]
				elif (strCtlName == "btnREVERT"):        clickedObj = cc.buttons["REVERT"]
				elif (strCtlName == "colorWheel"):	 clickedObj = cc.colorWheel
				elif (strCtlName == "sliderBRIGHTNESS"): clickedObj = cc.sliders["BRIGHTNESS"]
				elif (strCtlName == "sliderDRUM"):       clickedObj = cc.sliders["DRUM"]
				elif (strCtlName == "sliderSATURATION"): clickedObj = cc.sliders["SATURATION"]
				else:
					# Error: No object found by the given name.
					channelIndex = -1; clickedObj = None
					strCtlName = "(Object name not set)"

				# Return list containing channel index and selected object reference.
				return [channelIndex, strCtlName, clickedObj]

		# Nothing found.
		return None


	#-------------------------------------------------------
	# getRect()
	# Desc: Returns a list for the entire Section, in
	#       absolute pixels.
	#-------------------------------------------------------
	def getRect(self):
		return ((self.rectBG[0] + self.aLoc[0]),
		        (self.rectBG[1] + self.aLoc[1]),
		        (self.rectBG[2] + self.aLoc[0]),
		        (self.rectBG[3] + self.aLoc[1]))



	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Mouse event handler.
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		# See if a UIE is already selected.
		if (vkos_globals.SM.mouseCurUIE):
			# Populate working variables from selected UIE.
			clickedObj = vkos_globals.SM.mouseCurUIE		# Copy current UIE data.
			channelIndex = int(clickedObj.owner[-1:])		# Get channel index.
			strCtlName = vkos_globals.SM.mouseCurUIE.name		# Copy UIE name.
		else:
			# Nothing already selected... Determine which control will be affected.
			rtnObj = self.getCtl(event.pos)		# Determine UIE via screen coordinates.
			if (rtnObj):				# Populate working variables.
				channelIndex = rtnObj[0]
				clickedObj = rtnObj[2]
				strCtlName = rtnObj[1]
			else: return				# Could not determine object. Exit.
		bUpdateBrightness = False		# Brightness is updated live and may occur in more than one event.
		cc = self.channels[channelIndex]	# Current working channel; More "shorthand" for readability.


		# Process event struct
		#--------------------------------------
		# MouseButton DOWN
		#--------------------------------------
		if ((event.type == pygame.MOUSEBUTTONDOWN) and (clickedObj.enabled)):
			# Set last-active UIE.
			#----------------------
			vkos_globals.SM.setActiveUIE(clickedObj)
			clickedObj.mouseDown = True		# Set UIE's mouse button state.
			clickedObj.mouseButton = event.button	# Set clicked button.

			# Sliders
			if ("slider" in strCtlName):
				# Apply mouse axis locking for sliders.
				if (clickedObj.orientation == "H"): vkos_globals.SM.mouseLock("y")
				if (clickedObj.orientation == "V"): vkos_globals.SM.mouseLock("x")
				bUpdateBrightness = ("BRIGHTNESS" in strCtlName)	# Brightness is handled live.
				clickedObj.handleMouse(event)		# Pass event to object.

			# Buttons
			if (("btnREVERT" in strCtlName) or ("btnAPPLY" in strCtlName)):
				if (event.button == 1):			# Left mouse button.
					clickedObj.setState(True)	# Visually "press" the button.

			# Flag for redraw
			self.redraw = True


		# MouseButton UP [PROCESS CLICKS HERE]
		#--------------------------------------
		if ((event.type == pygame.MOUSEBUTTONUP) and (clickedObj.enabled)):
			# Sliders/ColorWheel
			#---------------------
			if (("sliderSATURATION" in strCtlName) or ("sliderBRIGHTNESS" in strCtlName) or
			    ("colorWheel" in strCtlName)):
				clickedObj.mouseDown = False			# Set UIE's mouse button state.
				clickedObj.mouseButton = -1			# Invalidate.

				# Update channel's preview color (Brightness or Saturation changed).
				cc.color = [cc.colorWheel.getColor()[0], cc.sliders["SATURATION"].getDisplayVal(), cc.sliders["BRIGHTNESS"].getDisplayVal()]
				cc.colorWheel.setColor(cc.color)		# Mirror channel's color info.

				# Changing the selected color enables the
				# channel's "APPLY" & "REVERT" buttons.
				# NOTE: Brightness is updated live, so it does not enable buttons.
				if not ("BRIGHTNESS" in strCtlName):
					cc.setButtons(bApply = True, bRevert = True)

			# Drum slider
			#-------------
			if ("sliderDRUM" in strCtlName):
				vkos_globals.addStatusMsg("Drum threshold for channel #" + str(channelIndex) + " is now " + str(clickedObj.getDisplayVal()), "!", 1)
				threshPacket = COMMS_Packet(iCmd = COMMS_OP_SET_DRUM_LEVEL,
							    iLamp = vkos_globals.CM.intToLampConst(channelIndex),
							    iParam = [CPACK_ARG_INT, clickedObj.getDisplayVal()])
				vkos_globals.CM.sendDevicePacket(self.deviceName, threshPacket)

			# Buttons
			#---------
			if (("btnREVERT" in strCtlName) or ("btnAPPLY" in strCtlName)):
				if (event.button == 1):			# Left mousebutton
					clickedObj.setState(False)	# Reset button's visual state.

					# Button: "APPLY"
					#-----------------
					if ("btnAPPLY" in strCtlName):
						# Get lamp ID and set new channel color.
						#----------------------------------------
						lamp = vkos_globals.CM.intToLampConst(channelIndex)
						cc.setColor([cc.colorWheel.getColor()[0],
							     cc.sliders["SATURATION"].getDisplayVal(),
							     cc.sliders["BRIGHTNESS"].getDisplayVal()])

						# Create lamp update packet.
						#----------------------------
						packColor = COMMS_Packet(COMMS_OP_SET_COLOR, lamp, [CPACK_ARG_COLOR, cc.getColor()])
						vkos_globals.CM.sendDevicePacket(self.deviceName, packColor)	# Send color update.

						# Fix button visual states.
						#---------------------------
						cc.setButtons(bApply = False, bRevert = False)
						strStatus = self.deviceName + ": Channel #" + str(channelIndex) + " is now set to " + str(cc.getColor())
						vkos_globals.addStatusMsg(strStatus, "*")
						return


					# Button: "REVERT"
					#------------------
					if ("btnREVERT" in strCtlName):
						sMsg = "Reverting [" + self.deviceName + " ch#" + str(channelIndex) + "] to " + str(cc.lastColor)
						vkos_globals.addStatusMsg(sMsg, "*")
						cc.revert()		# Restore previously-applied hue/saturation.
						return

			# ColorWheel UIE
			#-----------------
			if ("colorWheel" in strCtlName):
				cc.colorWheel.handleMouse(event)	# Color selected.
				cc.setButtons(bApply = True)		# Enable "Apply" button.
				cc.color = [cc.colorWheel.getColor()[0], cc.color[1], cc.color[2]]	# Update channel color.
				cc.colorWheel.setColor(cc.color)	# Mirror channel's color info.

			# Flag for redraw, for all UIEs.
			cc.redraw = True
			self.redraw = True


		# MouseMotion
		#-------------
		if ((event.type == pygame.MOUSEMOTION) and (clickedObj.enabled)):
			# Buttons do not (currently) process mouse motion.
			if (("btnREVERT" in strCtlName) or ("btnAPPLY" in strCtlName)): return
			else:
				# Update any values that may have changed.
				clickedObj.handleMouse(event)
				self.redraw = clickedObj.redraw		# Redraw, if needed.

				# If saturation was changed, update the color preview block.
				if (("SATURATION" in clickedObj.name) and (clickedObj.mouseDown)):
					cc.color = [cc.color[0], cc.sliders["SATURATION"].getDisplayVal(), cc.color[2]]
					cc.colorWheel.curColor = cc.color		# UIE mirrors channel's preview.
					cc.redraw = True				# Redraw channel.
					cc.draw(self.surfClient)

				# Brightness is live, updating as we scroll.
				if (("BRIGHTNESS" in strCtlName) and (clickedObj.mouseDown)): bUpdateBrightness = True


		# Brightness updates can occur from MOUSEDOWN/MOUSEMOTION events,
		# so handle it here to reduce redundancy.
		if (bUpdateBrightness):
			# Right mouse button affects brightness for ALL channels.
			if (clickedObj.mouseButton == 3):
				uniBrightness = clickedObj.getDisplayVal()
				uniPacket = COMMS_Packet(iCmd=COMMS_OP_SET_BRIGHTNESS, iLamp=ALL_LAMPS, iParam=[CPACK_ARG_BYTE, uniBrightness])
				vkos_globals.CM.sendDevicePacket(self.deviceName, uniPacket)	# Send universal update packet.

				# Update UI only
				for c in self.channels:
					self.setChannelBrightness(c.numChannel, uniBrightness, sendPacket=False, updateSlider=True)
				vkos_globals.addStatusMsg("Setting brightness to " + str(uniBrightness) + " for ALL channels on '" + self.deviceName + "'", "!", 1)
			else:
				# Set the current channel's brightness (No need to update slider).
				self.setChannelBrightness(channelIndex, clickedObj.getDisplayVal(), updateSlider=False)
				vkos_globals.addStatusMsg("Brightness for channel #" + str(channelIndex) + " is now " + str(clickedObj.getDisplayVal()), "*", 1)
		return



	#-------------------------------------------------------
	# setChannelBrightness(iChannel, iBrightness)
	# Desc: Set the brightness of a channel.
	#-------------------------------------------------------
	# -     iChannel: Channel to alter.
	# -  iBrightness: New brightness level (0~255).
	# - updateSlider: Indicates whether to set the position
	#                 of the channel's brightness slider.
	#-------------------------------------------------------
	def setChannelBrightness(self, iChannel, iBrightness, sendPacket=True, updateSlider=False):
		# Sanity-check; Brightness is 0~255.
		iBrightness = abs(iBrightness)			# Ensure positive numbers.
		if (iBrightness > 255): iBrightness = 255

		cc = self.channels[iChannel]				# Shorthand.
		lampID = vkos_globals.CM.intToLampConst(iChannel)	# Affected lamp ID.

		# If the lamp was previously off, we must actually enable it.
		#if ((cc.color[2] <= 10) and (iBrightness >= 1)):
		#	vkos_globals.CM.sendDevicePacket(self.deviceName, COMMS_Packet(COMMS_OP_ON, lampID))

		# Update channel's brightness level.
		cc.color = [cc.color[0], cc.color[1], iBrightness]
		if (updateSlider):
			cc.sliders["BRIGHTNESS"].setValue(iBrightness, bActualVal=True)
		if (sendPacket):
			brightnessPacket = COMMS_Packet(COMMS_OP_SET_BRIGHTNESS, lampID, [CPACK_ARG_BYTE, iBrightness])
			vkos_globals.CM.sendDevicePacket(self.deviceName, brightnessPacket)

		# If brightness is zero, turn the lamp completely off.
		if (iBrightness == 0):
			vkos_globals.CM.sendDevicePacket(self.deviceName, COMMS_Packet(COMMS_OP_OFF, lampID))
		return



	#-------------------------------------------------------
	# setOpacity(newOpacity)
	# Desc: Adjust opacity for alpha layering.
	#-------------------------------------------------------
	# - newOpacity: Transparency level (0~255)
	#-------------------------------------------------------
	def setOpacity(self, newOpacity):
		# Sanity checks; Only positive numbers, 0~255.
		if (newOpacity < 0): newOpacity = 0
		if (newOpacity > 255): newOpacity = 255
		self.opacity = newOpacity
		self.redraw = True
		return



	#-------------------------------------------------------
	# update()
	# Desc: Update and check for necessary redraws.
	#-------------------------------------------------------
	def update(self):
		# Check all UI elements for redraws.
		for c in range(0, self.numChannels):
			# If the channel is already flagged to redraw, skip checking its UIEs.
			if not self.channels[c].redraw:
				cc = self.channels[c]   	# "Shorthand" for code readbility.

				# Sliders/Buttons needing redrawn will redraw the entire channel.
				for s in cc.sliders.keys():
					if (cc.sliders[s].redraw): cc.redraw = True
				for b in cc.buttons.keys():
					if (cc.buttons[b].redraw): cc.redraw = True
		return
