#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objImage.py
#   - Version: 1.2
#   - Author: Kevin Slaughter
#   - Desc: Screen object >> Image Object
#        - This object is basically a wrapper to make
#          displaying images less of a headache.
#
#
#   [Revision History]
#   2019/09/26: v1.2 - Fixed bug that wouldn't let 2 instances of
#                      this class occupy the same screen (due to
#                      full-screen blanking).
#   2019/02/09: v1.1 - Commenting/Documentation standardization.
#                      General bugfixes, as well.
#   2018/12/16: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import time, os, pygame, vkos_globals



# Class Object
#--------------
class objImage(object):
	# Initialization
	#----------------
	def __init__(self, strFile=""):
		# General attributes
		#--------------------
		self.file = strFile	# Relative path to source image (Optionally passed to init).
		self.opacity = 255	# Alpha blending level (255 = none).
		self.owner = ""		# General host UIE/Section.
		self.redraw = True	# Flag to update the screen, if we changed it.
		self.rLoc = (0,0)	# Object location in absolute pixels.
		self.size = (0,0)	# Object dimensions.

		# Animation-related attributes
		#------------------------------
		self.animated = False	# Only True when object is in motion.
		self.animDuration = 0	# Full duration of current animation (in MS)
		self.animStep = 0	# Current step of animation.
		self.animStepFade = 0	# Amount to adjust fade by on each step.
		self.animStepX = 0	# Only applies to movement. Amount in pixels to move per step.
		self.animStepY = 0	# ^^^

		# Load image, if given.
		if (not (strFile == "")): self.loadImage(strFile)
		return


	# Functions
	#-------------------------------------------------------
	# doFade(fadeTime, fadeType)
	# Desc: Perform a fade out sequence over x-time (in MS).
	#-------------------------------------------------------
	# - fadeTime: Duration of fade out (in MS).
	# - fadeType: Type of fade. (-1: Fade out. | 1: Fade in)
	#-------------------------------------------------------
	def doFade(self, fadeTime, fadeType):
		self.animated = True					# Set active state.
		self.animDuration = fadeTime				# Animation duration.
		self.animStartTime = pygame.time.get_ticks()		# Mark start time for deltas.
		self.animStep = 0					# Reset the step/frame counter.
		self.animStepFade = float((255.0/fadeTime) * fadeType)	# Amount to adjust fade on each step.
		self.redraw = True					# Flag for a redraw.
		return


	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the offscreen image to another surface.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	# -      rgn: Rect; Draw only a specific region of the surface.
	#-------------------------------------------------------
	def draw(self, destSurf, rgn=None):
		# Determine if we're only drawing a specific region
		# of the image, or the entire thing. This is basically
		# only used for drawing underneath the SideMenu's
		# transparency, but may be handy later elsewhere.
		#--------------------------------------------
		if (not (rgn == None)):
			# Ensure a Rect-like array is provided.
			if (len(rgn) == 4):
				destRect = list(rgn)
				absDest = ((self.rLoc[0] + rgn[0]),
				           (self.rLoc[1] + rgn[1]))
			else:
				vkos_globals.printLog("Invalid region given for image hosted by '" + self.owner + "'", "!")
				vkos_globals.printLog("Region given: " + str(rgn), "*", 1)
				return		# Invalid parameter; Exit.
		else:
			# Default to entire image at it's current coordinates.
			destRect = self.getRelRect()		# Absolute coordinates/size (Rect, basically).
			absDest = (destRect[0], destRect[1])	# X/Y only.

		# Blank, then blit.
		#-------------------
		if not (rgn): rgn = self.getRelRect()
		destSurf.fill((0, 0, 0), rect=rgn)
		destSurf.blit(self.img, absDest)
		vkos_globals.addUpdateRgn(destRect)
		self.redraw = False
		return


	#-------------------------------------------------------
	# getAbsRect()
	# Desc: Returns a list for the image.
	#-------------------------------------------------------
	def getRect(self):
		return (0, 0, self.size[0], self.size[1])



	#-------------------------------------------------------
	# getRelRect()
	# Desc: Returns a list for the image.
	#-------------------------------------------------------
	def getRelRect(self):
		return (self.rLoc[0], self.rLoc[1], self.size[0], self.size[1])



	#-------------------------------------------------------
	# handleMouse(b, mpos)
	# Desc: Mouse event handler.
	#-------------------------------------------------------
	# -    b: Button pressed.
	# - mpos: Mouse position.
	#-------------------------------------------------------
	def handleMouse(self, b, mpos):
		# For now, images don't need to handle mouse events.
		return


	#-------------------------------------------------------
	# loadImage(filePath)
	# Desc: Loads an image from disk and initializes the
	#       surface it will be drawn upon.
	#-------------------------------------------------------
	# - filePath: Relative path to the image file, which
	#             starts in "img/" within the script's folder.
	#-------------------------------------------------------
	def loadImage(self, filePath):
		# Load file and set basic properties.
		#-------------------------------------
		self.file = ("img/" + filePath)		# All images are stored in the "img/" dir.
		self.img = None     			# Destroy any previous surface data
		self.img = pygame.image.load(self.file).convert()	# Prepare for use on the surface.
		self.img.set_colorkey((255, 0, 255))	# Magenta color key, for transparent region blitting.
		self.rLoc = (0,0)                    	# Default position
		self.size = self.img.get_size()     	# Update dimensions
		self.opacity = 255                  	# Reset alpha level
		self.redraw = True

		# Make sure animation is reset.
		#-------------------------------
		self.animated = False				# General flag for animation.
		self.animDuration = 0				# Duration (in MS) of the animation.
		self.animStep = 0				# Current animation step.
		self.animStepFade = 0.0				# Amount to adjust fade by on each step.
		self.animStepX = 0.0				# Amount to adjust X position on each step.
		self.animStepY = 0.0				# Amount to adjust Y position on each step.
		self.animClock = pygame.time.Clock()		# Initialize timer.
		self.animStartTime = pygame.time.get_ticks()	# Mark the start time, to calculate deltas/etc.
		return


	#-------------------------------------------------------
	# update()
	# Desc: Updates any active animation timing/position/alpha.
	#-------------------------------------------------------
	def update(self):
		if (self.animated):
			# Set current frame/step.
			#-------------------------
			self.animStep = (pygame.time.get_ticks() - self.animStartTime)		# Get current MS.
			if (self.animStep >= self.animDuration):
				self.animStep = self.animDuration	# We went past the mark... Jump back.
				self.animated = False               	# Animation completed, process last frame.

			# Adjust X/Y positioning (if needed).
			#-------------------------------------
			if ((self.animStepX <> 0) or (self.animStepY <> 0)):
				self.rLoc = ((self.rLoc[0] + self.animStepX), (self.rLoc[1] + self.animStepY))

				# Sanity check; Ensure coordinates do not exceed the screen boundaries.
				if (self.rLoc[0] > vkos_globals.SCREEN_SIZE[0]): self.rLoc[0] = vkos_globals.SCREEN_SIZE[0]
				if (self.rLoc[1] > vkos_globals.SCREEN_SIZE[1]): self.rLoc[1] = vkos_globals.SCREEN_SIZE[1]
				self.redraw = True	# Flag for redraw.

			# Adjust alpha level (if needed).
			#-----------------------------------------------
			# This looks funky, but a fade out just
			# uses a negative multiplier and only counts
			# in one direction, to (ironically, apparently)
			# simplify the logic in the update() function.
			#-----------------------------------------------
			if (not (self.animStepFade == 0)):
				self.opacity = int(self.animStepFade * self.animStep)
				if (self.animStepFade < 0):	# Negative step (Fading out)
					self.opacity = int((255 + self.opacity))	# Uses a negative value.

				# Sanity check, then update, since alpha is not handled elsewhere.
				if (self.opacity < 0):   self.opacity = 0
				if (self.opacity > 255): self.opacity = 255

				# Apply modified opacity to the surface and flag for redraw.
				self.img.set_alpha(self.opacity)
				self.img.convert_alpha()
				self.redraw = True
		return
