#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objPopup.py
#   - Version: 1.1
#   - Author: Kevin Slaughter
#   - Desc: Pop-up display (ex: tool-tips/modal dialogue boxes).
#
#
#   [Revision History]
#   2019/09/23: v1.1 - Implemented alpha layering, amid generally
#                      simlification/bug fixes.
#   2019/03/06: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import os, pygame, pygame.gfxdraw, sys, vkos_globals

# Object Declarations
#---------------------
#from vkos_globals import *



# Class Object
#--------------
class objPopup(object):
	# Initialization
	#----------------
	def __init__(self, title, displayText, colorBG, colorFG, dispRect):
		# The "Active" flag determines when the popup is to be
		# deleted. Since these are generally intended for
		# COMMS notifications, they aren't treated like regular
		# Sections/UIEs and thus do not have a generally-fixed conditional
		# stop indicator. To avoid more hackish code, setting this to
		# False anywhere in the program will trigger these actions.
		self.active = True

		# Set BG/FG colors.
		self.colors = {"BG": colorBG,
		               "FG": colorFG}

		# Adjust for centering, if needed.
		if (vkos_globals.SCREEN_CENTER in dispRect):
			dispRect = vkos_globals.autoRelativeCentering(dispRect)

		# Set dimension/size info.
		self.loc = (dispRect[0], dispRect[1])	# Store coordinates.
		self.size = (dispRect[2], dispRect[3])	# Store dimensions.
		self.fonts = {"Title"       : pygame.font.SysFont("Arial", 32, bold=True),
		              "displayText" : pygame.font.SysFont("Arial", 16)}
		self.opacity = 255			# Alpha layering (Fully opaque, by default).
		self.redraw = True			# Redraw flag.
		self.text = displayText			# Text to display.
		self.title = title			# Title of pop-up.
		self.surfMain = pygame.Surface((dispRect[2], dispRect[3]))	# Create display surface.
		return



	#-------------------------------------------------------
	# Functions
	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the popup display.
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Draw popup background.
		vkos_globals.AAfilledRoundedRect(self.surfMain, [0, 0, self.size[0], self.size[1]], self.colors["BG"], radius=0.09)
		tmpRect = [2, 2, (self.size[0]-4), (self.size[1]-4)]
		vkos_globals.AAfilledRoundedRect(self.surfMain, tmpRect, [96, 96, 96], radius=0.09)

		# Draw title.
		textSurf = self.fonts["Title"].render(self.title, True, self.colors["FG"])
		self.surfMain.blit(textSurf, [24, 9])
		textSurf = self.fonts["Title"].render(self.title, True, vkos_globals.changeColor(self.colors["FG"], 64))
		self.surfMain.blit(textSurf, [26, 11])

		# Draw message text (Multiple lines must be rendered separately).
		strLines = self.text.split("\n")
		for t in range(0, len(strLines)):
			textSurf = self.fonts["displayText"].render(strLines[t], True, (self.colors["FG"]))
			self.surfMain.blit(textSurf, (16, (64 + (24*t))))

		# Use alpha layering, if enabled.
		self.surfMain.set_alpha(self.opacity)
		self.surfMain.convert_alpha()

		# Blit to main screen.
		destSurf.blit(self.surfMain, (self.loc[0], self.loc[1]))
		self.redraw = False
		return



	#-------------------------------------------------------
	# getRect():
	# Desc: Returns a Rect-like tuple (x, y, w, h).
	#-------------------------------------------------------
	def getRect(self):
		return (self.loc[0], self.loc[1], self.size[0], self.size[1])



	#-------------------------------------------------------
	# setOpacity(newOpacity)
	# Desc: Adjust opacity for alpha layering.
	#-------------------------------------------------------
	# - newOpacity: Transparency level (0~255).
	#-------------------------------------------------------
	def setOpacity(self, newOpacity):
		# Sanity checks; Only positive numbers, 0~255.
		if (newOpacity < 0): newOpacity = 0
		if (newOpacity > 255): newOpacity = 255
		self.opacity = newOpacity
		self.redraw = True
		return



	#-------------------------------------------------------
	# setText(newText):
	# Desc: Changes the popup's display text.
	#-------------------------------------------------------
	# -    newText: New display text.
	# - appendText: If True, append text as a new line.
	#-------------------------------------------------------
	def setText(self, newText, appendText = False):
		if (appendText): self.text = (self.text + "\n" + newText)
		else: self.text = newText
		self.redraw = True
		return
