#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objScreenManager.py
#   - Version: 1.4
#   - Author: Kevin Slaughter
#   - Desc: Framebuffer graphic driver for VisualK-OS.
#
#
#   [Revision History]
#   2019/09/23: v1.4 - Numerous simplifications and fixes.
#                      Alpha layering (Transparency) is now
#                      utilized during GUI initialization,
#                      popup creation/dismissal and dimming
#                      of the desktop to focus said popups.
#   2019/03/12: v1.3 - Implemented status bar input indicators.
#   2019/02/09: v1.2 - Commenting/Documentation standardization.
#   2019/02/04: v1.1 - General/Numerous changes/bugfixes/etc.
#   2018/11/16: v1.0 - File created with initial framework.
#--------------------------------------------------------------



# Imports
#---------
import time, os, pygame, objDesktop, vkos_globals
from objImage import objImage
from objPopup import objPopup



# Class Object
#--------------
class objScreenManager(object):
	# Initialization
	#----------------
	def __init__(self):
		# Check if X is active.
		#-----------------------
		disp_no = os.getenv("DISPLAY")

		# Attempt to use FB video driver.
		#---------------------------------
		found = False; drivers = ["fbcon", "directfb", "svgalib"]
		for driver in drivers:
			if not os.getenv("SDL_VIDEODRIVER"):
				os.putenv("SDL_VIDEODRIVER", driver)
			try:
				pygame.display.init()
			except pygame.error:
				strMsg = "Driver: {0} failed."
				strMsg = strMsg.format(driver)
				vkos_globals.printLog(strMsg,"!",1)
				continue
			found = True
			break

		# Set up the display.
		#---------------------
		if not found:
			# No usable driver found... Exit.
			vkos_globals.printLog("No usable video driver found... Exiting. (Make sure you are using sudo).","!")
			quit()
		else:
			# Set screen mode and display resolution.
			#-----------------------------------------
			screenSize = [vkos_globals.SCREEN_SIZE[0], vkos_globals.SCREEN_SIZE[1]]
			vkos_globals.printLog("Setting display to " + str(screenSize[0]) + "x" + str(screenSize[1]) + "(F).", "*", 1)
			self.screen = pygame.display.set_mode(screenSize, pygame.FULLSCREEN or pygame.DOUBLEBUF)
			self.mouseLocked = False			# Ensure mouse axis locking is not active.

			# Set up cursor graphic and center.
			#-----------------------------------

			self.cursorPos = [(vkos_globals.SCREEN_SIZE[0] / 2), (vkos_globals.SCREEN_SIZE[1] / 2)]
			pygame.mouse.set_pos(self.cursorPos)

			# Initialize font subsystem and main objects.
			#---------------------------------------------
			vkos_globals.printLog("Initializing fonts.", "*", 1)
			pygame.font.init()
			self.fonts = {"Main"            : pygame.font.SysFont("Arial", 16),
			              "SideMenu-Header" : pygame.font.SysFont("Courier", 32, bold=True),
			              "Slider"          : pygame.font.SysFont("Arial", 16, bold=True),
			              "StatusBar"       : pygame.font.SysFont("Courier", 16, bold=True),
			              "TitleBar"        : pygame.font.SysFont("Arial", 18, bold=True, italic=True)}

			# Ready to begin!
			#-----------------
			self.clock = pygame.time.Clock()		# Timer for animations
			pygame.event.set_grab(True)			# Grab active input focus.
			self.blankScreen(); pygame.display.flip()	# Blank and refresh screen.
			self.mouseCurUIE = None				# Set UI mouse focus to None.
			self.activePopup = None				# No popups onscreen yet.
			self.desktop = None				# Desktop object placeholder.
		return


	# Functions
	#-------------------------------------------------------
	# blankScreen():
	# Desc: Blanks the screen.
	#-------------------------------------------------------
	def blankScreen(self): self.screen.fill([0,0,0]); return



	#-------------------------------------------------------
	# closePopup()
	# Desc: Safely closes out an existing popup.
	#-------------------------------------------------------
	def closePopup(self):
		if (self.activePopup):
			if (self.desktop):
				# Fade out popup.
				self.activePopup.setOpacity(255)	# Set to full opacity.
				for i in range(0, 6):
					pygame.event.pump()
					self.desktop.redraw = True
					self.activePopup.setOpacity((self.activePopup.opacity - 50))
					self.updateDisplay(bRepaintOnly=True)

				# Fade out dimmer.
				for i in range(0, 6):
					pygame.event.pump()
					self.desktop.setDimmer((self.desktop.dimAmt - 50))
					self.updateDisplay(bRepaintOnly=True)

			# Delete popup instance.
			self.activePopup = None
		return



	#-------------------------------------------------------
	# doPowerOut():
	# Desc: Crunches the main surface down vertically to several lines,
	#       then horizontally into nothing, like an old CRT
	#       display suddenly losing power (hence the name).
	#-------------------------------------------------------
	def doPowerOut(self):
		screenShot = pygame.Surface((vkos_globals.SCREEN_SIZE[0], vkos_globals.SCREEN_SIZE[1]))
		screenShot.blit(vkos_globals.SM.screen, vkos_globals.SM.screen.get_rect())
		scrHeight = vkos_globals.SCREEN_SIZE[1]; scrWidth = vkos_globals.SCREEN_SIZE[0]

		# Vertical shrinking
		#--------------------
		for v in range(1, ((vkos_globals.SCREEN_SIZE[1] / 2) - 5), 5):
			self.screen.fill([0, 0, 0])	# Blank before each redraw (effect is full-screen)
			tempSurf = pygame.transform.scale(screenShot, (scrWidth, (scrHeight - (v*2))))
			self.screen.blit(tempSurf, (0, v))
			pygame.display.update()

		# Horizontal shrinking
		#----------------------
		scrHeight = ((vkos_globals.SCREEN_SIZE[1] / 2) - 5)
		for v in range(1, (vkos_globals.SCREEN_SIZE[0] / 2), 5):
			self.screen.fill([0, 0, 0])	# Blank before each redraw (effect is full-screen)
			tempSurf = pygame.transform.scale(screenShot, ((scrWidth - (v*2)), 12))
			self.screen.blit(tempSurf, (v, scrHeight))
			pygame.display.update();

		# This is meant to be an "Exit" visual...
		# When the Desktop exits, this function is called,
		# so it should be generally safe to simply exit.
		pygame.quit()
		quit()
		return



	#-------------------------------------------------------
	# drawText(strText, (x,y), (r,g,b))
	# Desc: Draws text at the given coordinates in x-color,
	#       on the top-most layer, but does NOT set an
	#       update region.
	#-------------------------------------------------------
	# -  strText: Text to be drawn to the screen.
	# -    (x,y): (Tuple) Coordinates to draw text at.
	# -  (r,g,b): (Tuple) RGB color of text.
	# - destSurf: (Optional) Destination surface to
	#             render onto. Default is the main screen.
	#-------------------------------------------------------
	def drawText(self, strText, destCoords, arColor, destSurf=None):
		# Sanity-check: Colors must be in a 3-part tuple and follow [0 <= X <= 255].
		for c in range(0,3):
			bArgsOK = bool(((arColor[c] >= 0) and (arColor[c] <= 255)))

			# Invalid color; Change to something garrish to stand out.
			if not (bArgsOK): arColor = [128, 255, 70]; continue
		if (bArgsOK): textColor = arColor	# If all parts cleared, we can use the specified color.

		# Create a surface to render text onto, then blit for next main update.
		textSurf = self.fonts["StatusBar"].render(strText, True, textColor)
		if (destSurf): destSurf.blit(textSurf, destCoords)
		else:          self.screen.blit(textSurf, destCoords)	# Draw to main screen surface, by default.



	#-------------------------------------------------------
	# handleInputEvents(event):
	# Desc: Parses and distributes input events.
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleInputEvents(self, event):
		# Gamepad button pressed
		#------------------------
		if ((event.type == pygame.JOYBUTTONDOWN) or (event.type == pygame.JOYBUTTONUP)):
			# Determine status marker state.
			if (event.type == pygame.JOYBUTTONDOWN): statusGP = True
			if (event.type == pygame.JOYBUTTONUP):   statusGP = False

			# Mark input indicator and pass the event to the Desktop.
			self.desktop.statusBar.updateInputStatus("GP", statusGP)
			self.desktop.handleGP(event)


		# Key pressed
		#-------------
		if ((event.type == pygame.KEYDOWN) or (event.type == pygame.KEYUP)):
			# Determine status marker state.
			if (event.type == pygame.KEYDOWN): statusKB = True
			if (event.type == pygame.KEYUP):   statusKB = False

			# On "Esc", dismiss any active popups.
			if (self.activePopup):
				if ((event.type == pygame.KEYUP) and (event.key == pygame.K_ESCAPE)):
					self.desktop.setDimmer(dimmed=False)
					self.activePopup = None


			# Pass to desktop and mark input indicator.
			self.desktop.handleKey(event.key)
			self.desktop.statusBar.updateInputStatus("KB", statusKB)


		# Mouse event (movement/click/release)
		#--------------------------------------
		if ((event.type == pygame.MOUSEBUTTONUP) or
		    (event.type == pygame.MOUSEBUTTONDOWN) or
		    (event.type == pygame.MOUSEMOTION)):
			# If any popups are active, clicking will dismiss them.
			# Any messages that would have been displayed in the
			# popup will then be redirected to the statusBar console.
			#---------------------------------------------------------
			if ((self.activePopup) and
			    ((event.type == pygame.MOUSEBUTTONDOWN) or (event.type == pygame.MOUSEBUTTONUP))):
				self.closePopup()

			# Update input indicator
			#----------------------
			# Instead of keeping track of a timer after MOUSEMOTION, just don't react to it.
			if not (event.type == pygame.MOUSEMOTION):
				if (event.type == pygame.MOUSEBUTTONDOWN): statusMouse = True
				if (event.type == pygame.MOUSEBUTTONUP):   statusMouse = False
				self.desktop.statusBar.updateInputStatus("Mouse", statusMouse)


			# Apply mouse axis locking (if needed).
			#---------------------------------------
			if ((self.mouseCurUIE and self.mouseLocked) and (event.type == pygame.MOUSEMOTION)):
				curPos = pygame.mouse.get_pos()		# Get current mouse position.
				iOffset = 0				# Accounts for text labels/drawing-related spacing.

				# Restrict mouse movement to the Y axis (Vertical).
				#----------------------------------------------------
				if (self.mouseLockAxis == "x"):
					if (self.mouseCurUIE.type == "SLIDER"):
						# Vertical sliders have their text below the track.
						# Make sure to restrict to the usable track area
						iOffset = self.mouseCurUIE.rectText[3] + 10

					mousePos = (self.mouseLockPoint, curPos[1])
					if (curPos[1] < self.mouseCurUIE.aLoc[1]):
						# Above UIE bounds
						mousePos = [self.mouseLockPoint, self.mouseCurUIE.aLoc[1]]
					if (curPos[1] > ((self.mouseCurUIE.aLoc[1] + self.mouseCurUIE.size[1]) - iOffset)):
						# Below UIE bounds
						mousePos = [self.mouseLockPoint, ((self.mouseCurUIE.aLoc[1] + self.mouseCurUIE.size[1]) - iOffset)]

				# Restrict mouse movement to the X axis (Horizontal).
				#------------------------------------------------------
				if (self.mouseLockAxis == "y"):
					if (self.mouseCurUIE.type == "SLIDER"):
						# Horizonal sliders have their text above the track.
						# Make sure to restrict to the usable track area.
						iOffset = 10

					mousePos = [curPos[0], self.mouseLockPoint]
					if (curPos[0] < (self.mouseCurUIE.aLoc[0] + iOffset)):
						# Stop at left of UIE bounds.
						mousePos = [(self.mouseCurUIE.aLoc[0] + iOffset), self.mouseLockPoint]
					if (curPos[0] > (self.mouseCurUIE.aLoc[0] + self.mouseCurUIE.size[0] - iOffset)):
						# Stop at right of UIE bounds.
						mousePos = [(self.mouseCurUIE.aLoc[0] + self.mouseCurUIE.size[0] - iOffset), self.mouseLockPoint]

				# Apply cursor position modification.
				pygame.mouse.set_pos(mousePos)


			# If a mouse button is clicked and held,
			# all events should go to the originally-clicked UIE.
			# Trying to handle this by rects alone gets buggy/messy.
			if (self.mouseCurUIE and ((event.type == pygame.MOUSEMOTION) or (event.type == pygame.MOUSEBUTTONUP))):
				curPos = pygame.mouse.get_pos()		# Get current mouse position.

				# Check if we left the bounds of the UIE.
				if (not vkos_globals.isInRect(self.mouseCurUIE.getAbsRect(), curPos)):
					# If we go outside the bounds of a button, cancel any reaction.
					if (self.mouseCurUIE.type == "BUTTON"):
						self.mouseCurUIE.setState(False)	# Reset button state to normal.

					# Reset all mouse-related states/flags.
					if (self.mouseLocked): self.mouseUnlock()	# Remove any active axis locking.
					self.setActiveUIE(None)				# Release UIE reference.
					return

				# Process mousebutton release.
				if (event.type == pygame.MOUSEBUTTONUP):
					if (self.mouseCurUIE.type == "BUTTON"):
						self.mouseCurUIE.setState(False)	# Reset visual state for buttons.

					# Redraw host (if it's a color channel)
					if ("-c" in self.mouseCurUIE.owner):
						# Make sure we only redraw the proper channel.
						iLoc = self.mouseCurUIE.owner.find("-c")	# Extract channel number.
						sSection = self.mouseCurUIE.owner[:iLoc]	# Get Section name.
						iChannel = int(self.mouseCurUIE.owner[-1:])	# Get channel index from owner name.

						# Flag channel/section for redraw.
						self.desktop.sections[sSection].redraw = True
						self.desktop.sections[sSection].channels[iChannel].redraw = True


				# Process mousemotion (hovering)
				#if (event.type == pygame.MOUSEMOTION):


			# Pass all events on to normal handling, making
			# sure we are only affecting an active UIE.
			#-----------------------------------------------
			if ((event.type == pygame.MOUSEBUTTONUP) and not self.mouseCurUIE): return
			else: self.desktop.handleMouse(event)

			# Release mouse hooks on button release AFTER event processing.
			#---------------------------------------------------------------
			if ((event.type == pygame.MOUSEBUTTONUP) and self.mouseCurUIE):
				vkos_globals.SM.setActiveUIE(None)		# Release UIE reference.
		return



	#-------------------------------------------------------
	# mouseLock(axis)
	# Desc: Lock the mouse to the specified axis (X/Y)
	#-------------------------------------------------------
	# - axis: "x" = Vertical movement only.
	#         "y" = Horizontal movement only.
	#-------------------------------------------------------
	def mouseLock(self, axis):
		# Sanity check: Ensure we get a valid axis.
		if ((axis == "x") or (axis == "y")):
			self.mouseLocked = True
			self.mouseLockAxis = axis	# Define the locked axis
			i = 0 if (axis == "x") else 1
			self.mouseLockPoint = pygame.mouse.get_pos()[i]	# Set locked point.
		return



	#-------------------------------------------------------
	# mouseUnlock()
	# Desc: Unlock anything from a call to mouseLock().
	#-------------------------------------------------------
	def mouseUnlock(self): self.mouseLocked = False; return



	#-------------------------------------------------------
	# setActiveUIE(newUIE)
	# Desc: Set the current UIE that mouse events will
	#       work on.
	#-------------------------------------------------------
	# - newUIE: UIE object reference. Setting this to None
	#           will clear the active object.
	#-------------------------------------------------------
	def setActiveUIE(self, newUIE):
		# Sanity-check; Make sure we don't keep re-assigning the same UIE.
		if (self.mouseCurUIE and newUIE):
			if (self.mouseCurUIE.name == newUIE.name): return
			#else: vkos_globals.addStatusMsg("setActiveUIE: CURRENT = " + str(self.mouseCurUIE.name), "*")

		# Assign UIE reference, or clear, as necessary.
		#-----------------------------------------------
		if (newUIE):
			self.mouseCurUIE = newUIE		# Set UIE reference.
			#vkos_globals.addStatusMsg("setActiveUIE(): " + str(newUIE.name), "*", 1)
		else:
			self.mouseLocked = False		# Release any active axis locking.
			if (self.mouseCurUIE):
				if (self.mouseCurUIE.type == "BUTTON"): self.mouseCurUIE.setState(False)
				self.mouseCurUIE.mouseDown = False
				self.mouseCurUIE = None		# Clear UIE reference.
			#vkos_globals.addStatusMsg("ActiveUIE cleared.", "*", 1)
		return



	#-------------------------------------------------------
	# showDesktop()
	# Desc: Load the desktop object, which will host our main interface.
	#-------------------------------------------------------
	def showDesktop(self):
		self.desktop = objDesktop.objDesktop()				# Create Desktop object.
		vkos_globals.printLog("Desktop is now active.", "*", 1)		# Tidy-up terminal verbosity.
		vkos_globals.addStatusMsg("Initialized GUI", "*")		# First message to display in console.
		return



	#-------------------------------------------------------
	# showPopup(title, msg, colorBG, colorFG, dispRect):
	# Desc: Display a modal popup dialogue box.
	#-------------------------------------------------------
	# -    strTitle: Titlebar text.
	# - strIntroMsg: Display text.
	# -     colorBG: (Tuple) Background color [R, G, B].
	# -   colorText: (Tuple) Text color [R, G, B].
	# -    dispRect: (Tuple) Rect-like struct (x, y, w, h).
	#-------------------------------------------------------
	def showPopup(self, strTitle="", strIntroMsg="", colorBG=[64, 64, 64], colorText=[255, 255, 255], dispRect=[0, 0, 320, 240]):
		# Dim desktop and fade in popup.
		#--------------------------------
		if (self.desktop):
			# Fade in dimmer.
			for i in range(0, 3):
				pygame.event.pump()
				self.desktop.setDimmer(self.desktop.dimAmt + 80)
				self.updateDisplay(bRepaintOnly=True)

				# Max out at ~75% dimming.
				if (self.desktop.dimAmt >= 220):
					self.desktop.setDimmer(220)
					break

			# Create popup UIE
			#------------------
			self.activePopup = objPopup(strTitle, strIntroMsg, colorBG, colorText, dispRect)

			# Fade in popup
			self.activePopup.setOpacity(0)		# Set to fully transparent.
			for i in range(0, 6):
				pygame.event.pump()		# Allow system mouse events while looping.
				self.desktop.redraw = True	# Redraw the background each time.
				self.activePopup.setOpacity((self.activePopup.opacity + 50))	# Increase opacity.
				self.updateDisplay(bRepaintOnly=True)				# Update screen.
		return



	#-------------------------------------------------------
	# showSplashScreen()
	# Desc: Displays the Splash Screen graphic. ("logo.png")
	#       This is basically a miniature version of the
	#       general update() function used in the Desktop
	#       object, which is far less convoluted than using
	#       direct screen calls.
	#-------------------------------------------------------
	def showSplashScreen(self):
		# Load the banner image with approximate centering.
		#---------------------------------------------------
		imgLogo = objImage("logo.png")
		imgLogo.rLoc = [int((vkos_globals.SCREEN_SIZE[0] / 2) - (imgLogo.size[0] / 2)),
		                int((vkos_globals.SCREEN_SIZE[1] / 2) - (imgLogo.size[1] / 1.6))]

		# Load GPL logo
		#---------------
		imgGPL = objImage("gplv3.png")
		imgGPL.rLoc = [int((vkos_globals.SCREEN_SIZE[0]) - 164), int((vkos_globals.SCREEN_SIZE[1]) - 70)]

		# Fade in, pause, then fade out.
		#--------------------------------
		for f in range(1, -2, -2):
			if (f == 0): continue		# Sanity check; Skip 0, it will make the fade operation hang.
			imgLogo.doFade(1000, f)		# Initiate fade type (1 = fade in; -1 = fade out).
			imgGPL.doFade(1000, f)
			while(imgLogo.animated):
				pygame.event.pump()				# Pass on input events (or we look frozen).
				imgLogo.update()				# Advance animation frame.
				imgGPL.update()
				if (imgLogo.redraw): imgLogo.draw(self.screen)	# Redraw, if needed.
				if (imgGPL.redraw): imgGPL.draw(self.screen)
				pygame.display.update(vkos_globals.updateRgns)	# draw() adds to the main update stack...
				vkos_globals.clearUpdateRgns()			# Clean up our "mess".

			if (f == 1):
				for i in range(0, 10):
					pygame.event.pump()
					self.clock.tick(60)
			self.clock.tick(60)		# Limit to 60FPS.
		return



	#-------------------------------------------------------
	# updateDisplay()
	# Desc: Refresh any aspect of the screen, as needed.
	#-------------------------------------------------------
	# - bRepaintOnly: Skips event processing. (Kinda hacky)
	#-------------------------------------------------------
	def updateDisplay(self, bRepaintOnly=False):
		# For Popup overlays to redraw properly, the
		# bRepaintOnly argument bypasses event handling,
		# which was causing a recursion bug.
		if not bRepaintOnly:
			# Process input event queue before anything else.
			#-------------------------------------------------
			for event in pygame.event.get():
				self.handleInputEvents(event)
			pygame.event.pump()	# Clear/pass any unhandled events.

		# Update/Draw desktop.
		#----------------------
		if (self.desktop):
			# Update all elements and redraw anything needed.
			#-------------------------------------------------
			self.desktop.update()
			if (self.desktop.redraw): self.desktop.draw(self.screen)


		# Display mouse coordinates (if enabled).
		#----------------------------------------
		if (vkos_globals.showMouseInfo):
			mLoc = pygame.mouse.get_pos(); infoBacker = [0, 0, 200, 30]
			pygame.draw.rect(self.screen, [0, 0, 0], infoBacker)	# Blank behind text.
			self.drawText(("[X: " + str(mLoc[0]) + " - Y: " + str(mLoc[1]) + "]"), [5, 5], [255, 255, 255])
			vkos_globals.addUpdateRgn(infoBacker)

			# Draw bounding boxes for clickable areas.
			#------------------------------------------
			# deviceSection objects.
			iCnt = 0		# Iteration count for changing the colors for each deviceSection.
			iColor = [128, 0, 255]	# Device #0: Purple
			for b in self.desktop.sections.keys():
				if (iCnt == 1):   iColor = [255, 255, 0]	# Yellow
				elif (iCnt == 2): iColor = [0, 255, 255]	# Cyan

				if (len(self.desktop.sections[b].clickable) > 0):
					for c in self.desktop.sections[b].clickable:
						pygame.draw.rect(self.screen, iColor, self.desktop.sections[b].clickable[c], 1)
				iCnt += 1	# Increase for the next device.

			# SideMenu
			if (len(self.desktop.sMenu.clickable) > 0):
				iColor = [0, 255, 0]		# Green
				for b in self.desktop.sMenu.clickable:
					pygame.draw.rect(self.screen, iColor, self.desktop.sMenu.clickable[b], 1)



		# Popups are drawn over everything.
		#-----------------------------------
		if (self.activePopup):
			if (self.activePopup.active):
				if (self.activePopup.redraw):	# Redraw as needed.
					self.activePopup.draw(self.screen)
					vkos_globals.addUpdateRgn(self.activePopup.getRect())
			else:					# Delete popup when active is set to False.
				self.desktop.setDimmer(dimmed=False)		# Remove dimming.
				self.activePopup = None				# Delete popup.
				self.desktop.redraw = self.desktop.bg.redraw = True
				self.desktop.update()

			# Keep updating everything underneath the popup.
			self.desktop.redraw = True

		# Update any necessary screen regions.
		#--------------------------------------
		if (len(vkos_globals.updateRgns)>0):
			pygame.display.update(vkos_globals.updateRgns)	# Update any indicated regions.
			vkos_globals.clearUpdateRgns()			# Clear the list of update regions.
		return
