#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: objStatusBar.py
#   - Version: 1.5
#   - Author: Kevin Slaughter
#   - Desc: Screen object >> Status Bar
#           Displays a console and input indicator icons.
#
#
#   [Revision History]
#   2019/09/25: v1.5 - I realized that the Status Bar is hardly
#                      a "User Interface Element", as much as it
#                      is a notification area that may contain
#                      buttons at some point in the future, so
#                      the naming convention was changed to use
#                      "objStatusBar" instead of "UIE_StatusBar".
#   2019/09/23: v1.4 - Changed console display width.
#   2019/07/??: v1.3 - Added input status icons (statusIcon class).
#   2019/02/13: v1.2 - Console display fully implemented.
#   2019/02/09: v1.1 - Commenting/Documentation standardization.
#   2019/02/08: v1.0 - Initial framework.
#---------------------------------------------------------------



# Imports
#---------
import pygame, random, time, vkos_globals
from objImage import objImage
from UIE_Button import UIE_Button
from UIE_Base import *



# Constants
#-----------
INPUT_ICON_SIZE = 32
NUM_INPUT_ICONS = 3


# Class Objects
#---------------
class statusIcon(object):
	# Initialization
	#----------------
	def __init__(self, sType, locIcon):
		# Ensure Keyboard (KB), Gamepad (GP) or Mouse input.
		if ((sType == "KB") or (sType == "GP") or (sType == "Mouse")):
			self.loc = tuple(locIcon)	# X/Y tuple of icon's relative location.
			self.redraw = True		# Redraw flag.
			self.state = 0			# Default (0 = Inactive --- 1 = Active)
			self.type = sType

			# Load icons (32x32 PNG)
			#------------------------
			iconNormal = str("input-" + self.type + ".png")
			iconActive = str("input-" + self.type + "-active.png")
			self.iconNormal = objImage(iconNormal)		# Load normal icon.
			self.iconActive = objImage(iconActive)		# Load active icon.
		return



	# Functions
	#-------------------------------------------------------
	# draw(destSurf):
	# Desc: Draws the icon onto a given surface.
	#-------------------------------------------------------
	def draw(self, destSurf):
		curImg = self.getCurIcon()					# Get the current icon.
		if not (curImg == None): destSurf.blit(curImg, self.loc)	# Blit to screen.
		else:	# Error: Unknown icon.
			vkos_globals.printLog(("Error drawing status input icon: " + self.type), "!", 1)

		self.redraw = False		# Reset redraw flag.
		return



	#-------------------------------------------------------
	# getCurIcon():
	# Desc: Retrieve the current icon (determined by state).
	#-------------------------------------------------------
	def getCurIcon(self):
		if (self.state == 0): return self.iconNormal.img
		if (self.state == 1): return self.iconActive.img
		return None	# Failsafe. This should not happen.



	#-------------------------------------------------------
	# setState(state):
	# Desc: Set the current state of the icon (True/False).
	#-------------------------------------------------------
	# - stateNew: Active/Inactive icon (True/False).
	#-------------------------------------------------------
	def setState(self, stateNew):
		# Sanity-check; Ensure the new state is a boolean.
		if ((stateNew == True) or (stateNew == False)):
			self.state = stateNew
			self.redraw = True
		return



class objStatusBar(UIE_Base):
	# Initialization
	#----------------
	def __init__(self, dispRect):
		UIE_Base.__init__(self, dispRect, dispRect)     # Call standard init for all UIE objects.
		self.aLoc = [dispRect[0], dispRect[1]]		# StatusBar uses absolute locations.
		self.size = [dispRect[2], dispRect[3]]		# Store UIE dimensions.
		self.type = "MENU"				# Define UIE type.

		# Set up StatusBar components.
		#------------------------------
		self.slideActive = False			# Animation flag.
		self.slideCompleted = False			# Keeps up from re-displaying.
		self.console_lines = ["", "", "", ""]		# Console box contents (4 items max).
		self.rectConsole = [7, 7, (self.size[0] / 1.65), (self.size[1] / 1.16)]		# Define console area.
		self.surfConsole = pygame.Surface((self.rectConsole[2], self.rectConsole[3]))	# Create console surface.
		self.surfConsole.fill(vkos_globals.TC_STATUSBAR_CONSOLE)			# Fill console BG.

		# Set up buttons.
		#---------------------------------
		self.buttons = {}

		# Input Status indicators
		#-------------------------
		self.statusInput = {}
		self.statusLoc = [(vkos_globals.SCREEN_SIZE[0] - (INPUT_ICON_SIZE * NUM_INPUT_ICONS)), (self.size[1] - 45)]	# 3 (32x32) icons, 4px apart.
		self.statusInput.update({"GP" : statusIcon("GP", [(vkos_globals.SCREEN_SIZE[0] - 110), 62])})
		self.statusInput.update({"KB" : statusIcon("KB", [(vkos_globals.SCREEN_SIZE[0] - 74), 62])})
		self.statusInput.update({"Mouse" : statusIcon("Mouse", [(vkos_globals.SCREEN_SIZE[0] - 38), 62])})

		# Define clickable areas
		#------------------------
		self.clickable = {}
		return



	# Functions
	#-------------------------------------------------------
	# console_addLine(strMsg, strPrefix, iIndent)
	# Desc: Adds a line of text to the console box.
	#-------------------------------------------------------
	# -    strMsg: Message to add to console box.
	# - strPrefix: Message prefix, in brackets.
	# -   iIndent: Amount of message indentation.
	#-------------------------------------------------------
	def console_addLine(self, strMsg="", strPrefix="", iIndent=0):
		# Remove brackets, if no prefix is specified.
		if not (strPrefix == ""): strPrefix = "[" + strPrefix + "] "

		# Prefix timestamp (Date/time) to each message.
		strTimestamp = "(" + str(time.asctime())[:-5][11:] + ") "

		# Apply indentation, if needed.
		if (iIndent > 0):
			if (iIndent > 5): iIndent = 5		# Sanity-check; Max indent level = 5.
			strMsg = ("    " * iIndent) + strMsg

		# Determine which part of the console_lines tuple we are modifying.
		for c in range(0, 3):
			self.console_lines[c] = self.console_lines[(c+1)]	# Move each entry up by one.
		self.console_lines[3] = strTimestamp + strPrefix + strMsg	# Newest message at the bottom.

		# Redraw console text. Pygame does not handle multiline text,
		# so we must take care of it ourselves.
		self.surfConsole.fill(vkos_globals.TC_STATUSBAR_CONSOLE)	# Blank console area
		for t in range(0, 4):
			consoleText = self.console_lines[t]
			textSurf = vkos_globals.SM.fonts["StatusBar"].render(consoleText, True, [255, 255, 255], vkos_globals.TC_STATUSBAR_CONSOLE)
			self.surfConsole.blit(textSurf, [4, (3 + (20*t))])	# Margins: 4px from left, 3px from top.

		# Flag for redraw.
		self.redraw = True
		return


	#-------------------------------------------------------
	# doSlide()
	# Desc: Brings the display in with a flourish!
	#-------------------------------------------------------
	def doSlide(self):
		self.slideStep = 1				# Set up animation delta.
		self.slideStartTime = pygame.time.get_ticks()	# Set animation start time.
		self.slideActive = True
		self.aLoc = [self.aLoc[0], vkos_globals.SCREEN_SIZE[1]]	# Position off-screen, just past the bottom edge.
		return


	#-------------------------------------------------------
	# draw(destSurf)
	# Desc: Draws the element itself on a given surface.
	#-------------------------------------------------------
	# - destSurf: Destination surface to draw onto.
	#-------------------------------------------------------
	def draw(self, destSurf):
		# Blank BG and draw frame.
		self.surfMain.fill(vkos_globals.TC_STATUSBAR_BG)
		vkos_globals.drawContouredRect(self.surfMain, self.getRect(), vkos_globals.TC_BG_SECTION, invert=True)

		# Display version
		vkos_globals.SM.drawText(("v" + vkos_globals.PROGVERSION), [(self.size[0] - 50), (self.aLoc[1] + 5)], [255,255,255])

		# Draw text area on main surface, then overlay the text.
		pygame.draw.rect(self.surfMain, vkos_globals.TC_STATUSBAR_CONSOLE, self.rectConsole)
		vkos_globals.drawContouredRect(self.surfMain, self.rectConsole, vkos_globals.TC_STATUSBAR_CONSOLE, invert=True)
		self.surfMain.blit(self.surfConsole, [9, 9])	# Blit pre-rendered text display.


		# Draw buttons
		#--------------
		#for b in self.buttons.keys():
		#	self.buttons[b].draw(self.surfOpaque)

		# Draw input status icons
		#-------------------------
		for inpIcon in self.statusInput:
			self.statusInput[inpIcon].draw(self.surfMain)

		# Copy everything back to the main surface.
		#-------------------------------------------
		destSurf.blit(self.surfMain, self.getAbsRect())
		vkos_globals.addUpdateRgn(self.getAbsRect())
		self.redraw = False
		return


	#-------------------------------------------------------
	# handleMouse(event)
	# Desc: Handles mouse events
	#-------------------------------------------------------
	# - event: Event struct.
	#-------------------------------------------------------
	def handleMouse(self, event):
		# Determine affected UIE
		curObj = None
		for c in self.clickable.keys():
			if (vkos_globals.isInRect(self.clickable[c], event.pos)):
				objName = str(c)[3:]
				curObj = self.buttons[objName]
				print "* StatusBar Clicked \"" + str(c) + "\""
				vkos_globals.SM.setActiveUIE(curObj)
		if (curObj == None): return

		# Process event struct
		#----------------------
		# Mousebutton DOWN
		if (event.type == pygame.MOUSEBUTTONDOWN):
			if (curObj.enabled):
				if (curObj.type == "BUTTON"):
					curObj.setState(True)

		# Mousebutton UP
		#----------------
		if (event.type == pygame.MOUSEBUTTONUP):
			if (curObj.enabled):
				if (curObj.type == "BUTTON"):
					curObj.setState(False)		# Reset button's visual state.

					# EXIT
					if (curObj.text == "Exit"):
						vkos_globals.SM.doPowerOut()
						quit()
		return


	#-------------------------------------------------------
	# update()
	# Desc: Updates any timers/deltas in the UIE.
	#-------------------------------------------------------
	def update(self):
		# Update buttons (if needed/added).
		#for b in self.buttons.keys():
		#	if self.buttons[b].redraw == True:
		#		self.redraw = True

		# Check and/or update animations.
		if (self.slideActive):
			curStep = (pygame.time.get_ticks() - self.slideStartTime)	# Get the current step.
			if (curStep < 500):			# .5sec slide
				self.slideStep = curStep	# Update delta.
				if (self.slideStep < 500):
					vkos_globals.addUpdateRgn(self.getAbsRect())	# Add screen update region.

					# Update position.
					self.aLoc = [self.aLoc[0], int(self.aLoc[1] - ((self.slideStep / 250)*10))]

					# If we somehow moved beyond the screen height minus the statusbar, fix it.
					if (self.aLoc[1] < (vkos_globals.SCREEN_SIZE[1] - self.size[1])):
						self.aLoc = [self.aLoc[0], (vkos_globals.SCREEN_SIZE[1] - self.size[1])]
						self.slideActive = False
						self.slideCompleted = True
					self.redraw = True
				else:
					self.slideActive = False	# Deactivate slide.
					self.slideCompleted = True	# Mark as completed.
		return


	#-------------------------------------------------------
	# updateInputStatus()
	# Desc: Updates input-related status info.
	#-------------------------------------------------------
	# - device: Keyboard ("KB")
	#           Gamepad ("GP")
	#           Mouse ("Mouse")
	# - state: Input currently occurring (True/False)
	#-------------------------------------------------------
	def updateInputStatus(self, device, state):
		if ((device == "KB") or (device == "GP") or (device == "Mouse")):
			self.statusInput[device].setState(state)	# Update icon state.
			self.redraw = True
		else:
			# Unknown device... How the fuck..?
			vkos_globals.printLog(("Unknown input device: " + device), "!", 1)
		return
