#---------------------------------------------------------------
# * VisualK-OS (v5.0)
#   - File: vkos_globals.py
#   - Version: 1.4
#   - Author: Kevin Slaughter
#   - Desc: Include file of globaly-used constants/etc.
#           This module can be included anywhere in the
#           program, and as such, has no class module.
#
#
#   [Revision History]
#   2019/09/23: v1.4 - Further device abstraction via usage of
#                      POTENTIAL_DEVICES dictionary, in order
#                      to make adding future devices easier.
#   2019/06/07: v1.3 - General simplification and abstraction.
#   2019/02/09: v1.2 - Commenting/Documentation standardization.
#   2019/02/04: v1.1 - General housekeeping. Hardcoded resolution
#                      changed from 1280x720 --> 1366x768.
#   2018/12/10: v1.0 - File created with initial framework.
#---------------------------------------------------------------



# Imports
#---------
import colorsys, pygame



# Global Constants
#------------------
# Devices the system will look for/display. eg: {dev_name : [dev_path, UIE_title, num_channels, [x,y,w,h]]}
POTENTIAL_DEVICES = {"FBA"  : ["/dev/vkos/FBA",  "Flood Bank Array", 8, [50, 0, 1132, 333]],
		     "Thor" : ["/dev/vkos/Thor", "Thor",             8, [20, 338, 1132, 333]]}

# Program info and UIE dimensions.
PROGTITLE = "VisualK-OS"		# Program title displayed in console.
PROGVERSION = "5.0"			# Current working software version. If this is changed, update the splash logo, too.
SCREEN_CENTER = -9			# Avoid re-calculating relative centers.
SCREEN_SIZE = [1366, 768]		# Screen resolution is NOT variable. Seriously, leave this the fuck alone.
CHANNEL_SIZE = [138, 294]		# Used in objColorChannel, for each lamp channel's dimensions.
CHANNEL_SPACER = 3			# Distance (in pixels) between objColorChannel instances.
STATUSBAR_DIMENSIONS = [0, (SCREEN_SIZE[1] - 95), SCREEN_SIZE[0], 95]	# Dimensions of the StatusBar UIE object.
SIDEMENU_DIMENSIONS = [(SCREEN_SIZE[0] - 156), 3, 156, (SCREEN_SIZE[1] - STATUSBAR_DIMENSIONS[3])]
SIDEMENU_TEXT = "COMMS"



# Global objects/variables
#--------------------------
CM = None			# Globally-accessible COMMS reference.
SM = None			# Globally-accessible ScreenManager reference.
emulateDevices = False		# Emulates ALL slave devices for UI debugging; COMMS is completely disabled.
showMouseInfo = False		# Displays mouse coordinates on-screen.
updateRgns = []			# List of Rects for more-efficient screen updates.



# Theme-related values, to more easily customize the
# UI without implementing a scripting interpretation module.
#------------------------------------------------------------
TC_FRAME = [128, 128, 128]			# General UIE frame color.
TC_BG_SECTION = [12, 12, 12]			# BG color of all Section objects.
TC_FG_BUTTON = [255, 255, 255]			# Button text
TC_FG_BUTTON_ACTIVE = [0, 255, 0]		# Button text, when clicked/active.
TC_FG_BUTTON_DISABLED = [40, 40, 40]		# Button text, when disabled.
TC_FG_SECTION = [255, 255, 255]			# Main text color within all Sections.
TC_SIDEMENU_BG = [80, 80, 80]			# SideMenu BG color
TC_SIDEMENU_FG = [255, 255, 255]		# SideMenu text color.
TC_STATUSBAR_BG = [70, 70, 70]			# StatusBar BG color.
TC_STATUSBAR_COLOR = [40, 40, 40]		# Status bar color.
TC_STATUSBAR_TEXT = [0, 255, 0]			# Status bar text color.
TC_STATUSBAR_CONSOLE = [0, 0, 200]		# BG color of the StatusBar console.
TC_STATUSBAR_CONSOLE_TEXT = [255, 255, 255]	# StatusBar console text color.

# TitleBar gradients & font color
TC_TITLEBAR = {"FBA"          : [[0, 0, 50], [0, 128, 100]],
	       "Thor"         : [[0, 0, 0], [0, 192, 0]],
	       "Disconnected" : [[16, 16, 16], [80, 80, 80]],
	       "Muted"        : [[64, 0, 0], [255, 0, 0]]}
TC_TITLEBAR_TEXT = [255, 255, 255]              # Color of all Sections' TitleBar text.



# Global Functions
#-------------------------------------------------------
# AAfilledRoundedRect(surface, rect, color, radius)
# Desc: Draws an anti-aliased rounded rectangle.
#       Code from: https://www.pygame.org/project-AAfilledRoundedRect-2349-.html
#-------------------------------------------------------
def AAfilledRoundedRect(surface, rect, color, radius=0.4):
	rect = pygame.Rect(rect); color = pygame.Color(*color)
	alpha = color.a; color.a = 0; pos = rect.topleft; rect.topleft = 0,0
	rectangle = pygame.Surface(rect.size,pygame.SRCALPHA)
	circle = pygame.Surface([min(rect.size)*3]*2,pygame.SRCALPHA)
	pygame.draw.ellipse(circle,(0,0,0),circle.get_rect(),0)
	circle = pygame.transform.smoothscale(circle,[int(min(rect.size)*radius)]*2)
	radius = rectangle.blit(circle,(0,0))
	radius.bottomright = rect.bottomright; rectangle.blit(circle,radius)
	radius.topright = rect.topright; rectangle.blit(circle,radius)
	radius.bottomleft = rect.bottomleft; rectangle.blit(circle,radius)
	rectangle.fill((0,0,0),rect.inflate(-radius.w,0))
	rectangle.fill((0,0,0),rect.inflate(0,-radius.h))
	rectangle.fill(color,special_flags=pygame.BLEND_RGBA_MAX)
	rectangle.fill((255,255,255,alpha),special_flags=pygame.BLEND_RGBA_MIN)
	surface.blit(rectangle,pos)
	return



#--------------------------------------------------------
# addStatusMsg(strMsg, strPrefix, iIndent)
# Desc: Shortcut to output info to either the statusBar,
#       an active popup object or (if all else fails)
#       the terminal.
#-------------------------------------------------------
# -    strMsg: Message to display.
# - strPrefix: Prefix to display (Usually "*" or "!").
# -   iIndent: Level of tab indentation.
# -  bConsole: Specifically direct output to the StatusBar console.
#-------------------------------------------------------
def addStatusMsg(strMsg, strPrefix, iIndent=0, bConsole=False):
	if (SM):
		# Append to an active popup, if active.
		#---------------------------------------
		if ((SM.activePopup) and (not bConsole)):
			SM.activePopup.setText(strMsg, True)	# Append to existing text in the popup.
			SM.updateDisplay(bRepaintOnly = True)	# Force repaint to update popup overlay.
			return


		# No active popup, output to either the statusBar or terminal.
		#--------------------------------------------------------------
		if (SM.desktop):
			if (SM.desktop.statusBar):
				# Output to StatusBar.
				SM.desktop.statusBar.console_addLine(strMsg, strPrefix, iIndent)
			else:
				# Failsafe; StatusBar not loaded... Output to terminal.
				printLog(strMsg, strPrefix, iIndent)

	else:
		# ScreenManager not loaded, output to terminal.
		#-----------------------------------------------
		printLog(strMsg, strPrefix, iIndent)
	return



#-------------------------------------------------------
# addUpdateRgn(uRect):
# Desc: Adds a region to the screen update stack.
#-------------------------------------------------------
# - uRect: Rect region to update.
#-------------------------------------------------------
def addUpdateRgn(uRect):
	# Sanity check; If (for some reason) we end up with an oversized rect,
	#               we just assume it was intended for the whole screen.
	global updateRgns
	if ((uRect[2] == SCREEN_SIZE[0]) and (uRect[3] == SCREEN_SIZE[1])):
		clearUpdateRgns(); updateRgns.append(pygame.Rect(0, 0, SCREEN_SIZE[0], SCREEN_SIZE[1]))
	else:
		# Add the rect structure to the update list.
		updateRgns.append(pygame.Rect(uRect[0], uRect[1], uRect[2], uRect[3]))
	return



#-------------------------------------------------------
# autoRelativeCentering(uRect)
# Desc: Calculates relative centering with respect to
#       the object we are creating. NOTE: This function
#       expects the size to be in uRect[2]/uRect[3].
#-------------------------------------------------------
# - uRect: Rect structure to modify.
#-------------------------------------------------------
def autoRelativeCentering(uRect):
	if (uRect[0] == SCREEN_CENTER):
		uRect = [((SCREEN_SIZE[0] / 2) - (uRect[2] / 2)), uRect[1], uRect[2], uRect[3]]
	if (uRect[1] == SCREEN_CENTER):
		uRect = [uRect[0], ((SCREEN_SIZE[1] / 2) - (uRect[3] / 2)), uRect[2], uRect[3]]
	return uRect



#-------------------------------------------------------
# changeColor(colorSrc, iAmt)
# Desc: Alter each channel of a color tuple by the
#       given amount. Returns a modified color tuple.
#-------------------------------------------------------
# - colorSrc: Source color struct.
# -     iAmt: Amount (Neg/Pos) to alter each channel by.
#-------------------------------------------------------
def changeColor(colorSrc, iAmt):
	# Apply indicated changes.
	tempColor = [(colorSrc[0] + iAmt),
	             (colorSrc[1] + iAmt),
	             (colorSrc[2] + iAmt)]

	# Sanity check; Ensure we return a valid color.
	for e in range(0, 3):
		# Check boundaries
		if (tempColor[e] < 0):     tempColor[e] = 0
		elif (tempColor[e] > 255): tempColor[e] = 255
	return [tempColor[0], tempColor[1], tempColor[2]]



#-------------------------------------------------------
# clearUpdateRgns()
# Desc: Clears the update region list.
#-------------------------------------------------------
def clearUpdateRgns():
	global updateRgns
	updateRgns = None; updateRgns = []
	return



#-------------------------------------------------------
# drawContouredRect(destSurf, colorBG)
# Desc: Draws a contoured box from a list.
#-------------------------------------------------------
# - destSurf: Destination surface to draw onto.
# -  rectSrc: Rect area to utilize.
# -  colorBG: Main rect color.
# -   invert: Draw contouring in reverse, for a
#             "pressed-down" effect).
#-------------------------------------------------------
def drawContouredRect(destSurf, rectSrc, colorBG, invert=False):
	ps = [];
	if (invert): chgColor = (-20)
	else: chgColor = 20
	tempColor = changeColor(colorBG, chgColor)        	# Change color
	ps.append([rectSrc[0], (rectSrc[1] + rectSrc[3])])  	# Bottom-left
	ps.append([rectSrc[0], rectSrc[1]])			# Top-left
	ps.append([(rectSrc[0] + rectSrc[2]), rectSrc[1]])    	# Top-right
	pygame.draw.lines(destSurf, tempColor, False, ps, 2)

	ps.pop(1);	# Drop top-left point and change color...
	if (invert): chgColor = 30
	else: chgColor = (-30)
	tempColor = changeColor(colorBG, chgColor)
	ps.insert(1, [(rectSrc[0] + rectSrc[2]), (rectSrc[1] + rectSrc[3])])  	# Bottom-right
	pygame.draw.lines(destSurf, tempColor, False, ps, 2)			# Draw frame lines.
	return



#-------------------------------------------------------
# drawGradient(destSurf, fromColor, toColor, rectDim, isVertical, isL2R)
# Desc: Draws a gradient. (Adapted from pyGame Wiki)
#-------------------------------------------------------
# -   destSurf: Destination surface to draw onto.
# -     colors: Tuple of color tuples.
#               (eg: ((0,0,0),(0,255,0)) is black->green)
# -    rectDim: Rect gradient will be painted within.
# - isVertical: Determines if gradient progression is
#               vertically-oriented.
# -      isL2R: Determines if gradient progression
#               goes from left to right, or in reverse.
#-------------------------------------------------------
def drawGradient(destSurf, colors, rectDim, isVertical=False, isL2R=True):
	x1,x2 = rectDim[0], rectDim[2]
	y1,y2 = rectDim[1], rectDim[3]
	if isVertical: h = (y2 - y1)
	else:          h = (x2 - x1)
	if (h==0): h = 1
	if isL2R: a, b = colors[0], colors[1]
	else:     b, a = colors[0], colors[1]
	rate = (float(b[0] - a[0]) / h,
		float(b[1] - a[1]) / h,
	float(b[2] - a[2]) / h)
	if isVertical:
		for line in range(y1,y2):
			color = (min(max(a[0] + (rate[0]*(line-y1)), 0), 255),
				 min(max(a[1] + (rate[1]*(line-y1)), 0), 255),
		 min(max(a[2] + (rate[2]*(line-y1)), 0), 255))
			pygame.draw.line(destSurf, color, (x1,line), (x2,line))
	else:
		for col in range(x1,x2):
			color = (min(max(a[0] + (rate[0]*(col-x1)),0),255),
				 min(max(a[1] + (rate[1]*(col-x1)),0),255),
		 min(max(a[2] + (rate[2]*(col-x1)),0),255))
			pygame.draw.line(destSurf, color, (col, y1), (col, y2))
	return



#-------------------------------------------------------
# HSV2RGB(h)
# Desc: Returns an RGB tuple of an HSV color value.
#       Only H is needed, S/V are assummed to be 255
#       since we control brightness separately.
#       Colorsys requires a value from 0~1.0, which is
#       why the function looks a little odd.
#-------------------------------------------------------
# - h: Color Hue (H); 0 ~ 360
# - s: Color Saturation (S): 255 = Color   -   0 = White
#-------------------------------------------------------
def HSV2RGB(h, s):
	# Sanity checks; A valid color value is between 0~1.
	if (h > 360): h = 360
	elif (h < 0): h = 0
	if (s > 255): s = 255
	elif (s < 0): s = 0

	# Return an RGB tuple, rounded and converted to integers.
	return tuple(int(round(i * 255)) for i in colorsys.hsv_to_rgb(((1.0 / 360.0) * h), ((1.0 / 255.0) * s), 1.0))



#-------------------------------------------------------
# isInRect(uRect, pos)
# Desc: Checks if a point is within a given Rect,
#       returns True/False.
#-------------------------------------------------------
# - uRect: list to check.
# -   pos: (X,Y) of point to check.
#-------------------------------------------------------
def isInRect(uRect, pos): return bool(pygame.Rect(uRect[:]).collidepoint(pos))



#-------------------------------------------------------
# powerOffAllDevices()
# Desc: Properly closes all active devices.
#-------------------------------------------------------
def powerOffAllDevices():
	if (CM):
		printLog("Turning off all lighting devices...", "*")
		for dev in CM.devices.keys(): CM.disconnectDevice(dev)
	return



#-------------------------------------------------------
# printLog(sMsg, sPrefix, iIndent)
# Desc: Output a standardized message to the terminal,
#       with a given prefix and level of indentation.
#-------------------------------------------------------
# -    sMsg: Message to be printed, following bracketed
#            prefix.
# - sPrefix: Prefix which appears in brackets, before
#            the specified text. (ie: "*" shows "[*]")
#            Set this to None to just output directly.
# - iIndent: Level of indentation.
#-------------------------------------------------------
def printLog(sMsg, sPrefix = "", iIndent = 0):
	if (abs(int(iIndent)) > 5): iIndent = 5				# No more than 5 indentations.
	if not (sPrefix == ""): sPrefix = "[" + sPrefix + "] "	# Prepend prefix, if needed.
	print str(("    " * iIndent) + sPrefix + sMsg)
	return
